!> \file
!! Contains subroutines def_geom() and cal_geom()
!!
!! This file contains the routines which define the
!! geometry of the grid (cartesian, spherical or
!! cylindrical). It computes the cell coordinates, sizes,
!! ares and volumes.
!<

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine DEF_GEOM
!
!> Definition of the grid geometry by computing the cell
!! interface coordinates.
!<
subroutine def_geom

  use parameters, only : nx,ndim,nx_glob,nbuf
  use cl        , only : icl
  use geom      , only : x,x_glob

  implicit none
  
  integer :: ii,idim,ibuf

  do idim = 1,ndim

     do ibuf=1,nbuf
        ii = 2*idim-1
        if(icl(ii).eq.1 .or. icl(ii).eq.4 .or. icl(ii).eq.5) then ! flux or imposed values
           x(1-ibuf,idim) = 2.*x(1-ibuf+1,idim)-x(1-ibuf+2,idim)
        else if (icl(ii).eq.2) then ! periodic
           x(1-ibuf,idim) = x(1,idim)+ x_glob(nx_glob(idim)+1-ibuf,idim) - x_glob(nx_glob(idim)+1,idim)
        else if (icl(ii).eq.3) then ! reflexive
           x(1-ibuf,idim) =  2.*x(1,idim)-x(2-(1-ibuf),idim)
        endif
     
        ii = 2*idim
        if(icl(ii).eq.1 .or. icl(ii).eq.4 .or. icl(ii).eq.5) then
           x(nx(idim)+1+ibuf,idim) = 2.*x(nx(idim)+ibuf,idim)-x(nx(idim)-1+ibuf  ,idim)
        else if (icl(ii).eq.2) then
           x(nx(idim)+1+ibuf,idim) = x(nx(idim)+1,idim)+ x_glob(1+ibuf,idim)-x_glob(1,idim)
        else if (icl(ii).eq.3) then
           x(nx(idim)+1+ibuf,idim) = 2.*x(nx(idim)+1,idim) -x(nx(idim)+1-ibuf  ,idim)
        endif
     enddo
  enddo
  
  return
  
end subroutine def_geom

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine CAL_GEOM
!
!> Computes the cell sizes, face areas and volumes.
!<
subroutine cal_geom(nnx,nny,nnz)

  use parameters, only : ndim
  use var_loc   , only : xloc
  use geom

  implicit none
  
  integer               :: nnx,nny,nnz
  integer               :: i,j,k,idim,nmax
  integer, dimension(3) :: nn
  real   , dimension(3) :: dd
  real                  :: dcost, sint

  nmax = max(nnx,nny,nnz)
  nn(1) = nnx ; nn(2) = nny ; nn(3) = nnz

  allocate(dx(-1:nmax+2,3),dxc(0:nmax+2,3),xcloc(-1:nmax+2,3))
  allocate(ds( 0:nnx+2,0:nny+2,0:nnz+2,1:ndim))
  allocate(dv( 0:nnx+1,0:nny+1,0:nnz+1))

  dx = 1. ; xcloc = 0.5 ; dxc = 1.
  do idim = 1,ndim

     ! Beware of the calculation of the cell surfaces and volumes if the definition of xc is changed
     do i=-1,nn(idim)+2
        xcloc(i,idim) = 0.50*(xloc(i,idim) + xloc(i+1,idim))
     enddo

     do i=-1,nn(idim)+2
        dx(i,idim) = xloc(i+1,idim) - xloc(i,idim)
     enddo

     Do i=0,nn(idim)+2
        dxc(i,idim) = xcloc(i,idim) - xcloc(i-1,idim)  ! distance centre to centre
     Enddo
     
  enddo


  ds = 1.
  dv = 1.
  
  !    ***********
  !
  !    CARTESIAN
  !
  !    ***********

  if (Cartesien) then

     do k=0,nn(3)+1
     do j=0,nn(2)+1
     do i=0,nn(1)+1

         dv(i,j,k) = dx(i,1)*dx(j,2)*dx(k,3)
     enddo
     enddo
     enddo

     do idim=1,ndim

        do k=0,nn(3)+2
        do j=0,nn(2)+2
        do i=0,nn(1)+2

           dd(1) = dx(i,1)
           dd(2) = dx(j,2)
           dd(3) = dx(k,3)
           dd(idim) = 1.
           
           ds(i,j,k,idim) = dd(1)*dd(2)*dd(3)
           
        enddo
        enddo
        enddo

     enddo
     
  endif

  !    ***********
  !
  !    CYLINDRICAL
  !
  !    ***********

  if (Cylindrique) then

     if (geometrie =='rc') then
        do k=0,nn(3)+2
        do j=0,nn(2)+2
        do i=0,nn(1)+2

           ds(i,j,k,1) = xloc(i,1)*        dx(j,2)*dx(k,3)            

        enddo
        enddo
        enddo

        do k=0,nn(3)+1
        do j=0,nn(2)+1
        do i=0,nn(1)+1

           dv(i,j,k  ) = xcloc(i,1)*dx(i,1)*dx(j,2)*dx(k,3)

        enddo
        enddo
        enddo
        
     endif

     if (geometrie =='rp') then
        do k=0,nn(3)+2
        do j=0,nn(2)+2
        do i=0,nn(1)+2

           ds(i,j,k,1) = xloc(i,1)*        dx(j,2)*dx(k,3)            
           ds(i,j,k,2) =           dx(i,1)*        dx(k,3)            

        enddo
        enddo
        enddo

        do k=0,nn(3)+1
        do j=0,nn(2)+1
        do i=0,nn(1)+1

           dv(i,j,k  ) = xcloc(i,1)*dx(i,1)*dx(j,2)*dx(k,3)

        enddo
        enddo
        enddo

     endif

     if (geometrie =='rz') then
        do k=0,nn(3)+2
        do j=0,nn(2)+2
        do i=0,nn(1)+2

           ds(i,j,k,1) = xloc (i,1)        *dx(j,2)*dx(k,3)            
           ds(i,j,k,2) = xcloc(i,1)*dx(i,1)        *dx(k,3)            

        enddo
        enddo
        enddo

        do k=0,nn(3)+1
        do j=0,nn(2)+1
        do i=0,nn(1)+1

           dv(i,j,k  ) = xcloc(i,1)*dx(i,1)*dx(j,2)*dx(k,3)

        enddo
        enddo
        enddo

     endif

     if (geometrie =='rpz') then
        do k=0,nn(3)+2
        do j=0,nn(2)+2
        do i=0,nn(1)+2

           ds(i,j,k,1) = xloc (i,1)        *dx(j,2)*dx(k,3)            
           ds(i,j,k,2) =            dx(i,1)        *dx(k,3)            
           ds(i,j,k,3) = xcloc(i,1)*dx(i,1)*dx(j,2)            

        enddo
        enddo
        enddo

        do k=0,nn(3)+1
        do j=0,nn(2)+1
        do i=0,nn(1)+1

           dv(i,j,k  ) = xcloc(i,1)*dx(i,1)*dx(j,2)*dx(k,3)

        enddo
        enddo
        enddo

     endif

  endif

  !    *********
  !
  !    SPHERICAL
  !
  !    *********

  if (Spherique) then

     if (geometrie =='rs') then
        do k=0,nn(3)+2
        do j=0,nn(2)+2
        do i=0,nn(1)+2

           ds(i,j,k,1) = xloc(i,1)**2   *dx(k,3)            

        enddo
        enddo
        enddo

        do k=0,nn(3)+1
        do j=0,nn(2)+1
        do i=0,nn(1)+1

           dv(i,j,k  ) = (xloc(i+1,1)**3 - xloc(i,1)**3)/3. *dx(k,3)  

        enddo
        enddo
        enddo
        
     endif

     if (geometrie =='rthe') then

        do k=0,nn(3)+2
        do j=0,nn(2)+2
        do i=0,nn(1)+2

           dcost = cos(xloc(j,2))-cos(xloc(j+1,2))
            sint = sin(xcloc(j,2))

           ds(i,j,k,1) =  xloc(i,1)**2 * dcost *           dx(k,3)            
           ds(i,j,k,2) = xcloc(i,1)    *  sint * dx(i,1)*  dx(k,3)            

        enddo
        enddo
        enddo

        do k=0,nn(3)+1
        do j=0,nn(2)+1
        do i=0,nn(1)+1

           dcost = cos(xloc(j,2))-cos(xloc(j+1,2))

           dv(i,j,k) = (xloc(i+1,1)**3 - xloc(i,1)**3)/3. * dcost * dx(k,3)  

        enddo
        enddo
        enddo

     endif

     if (geometrie =='rtp') then

        do k=0,nn(3)+2
        do j=0,nn(2)+2
        do i=0,nn(1)+2

           dcost = cos(xloc(j,2))-cos(xloc(j+1,2))
            sint = sin(xcloc(j,2))

           ds(i,j,k,1) =  xloc(i,1)**2 * dcost *                 dx(k,3)            
           ds(i,j,k,2) = xcloc(i,1)    *  sint * dx(i,1)*        dx(k,3)            
           ds(i,j,k,3) = xcloc(i,1)            * dx(i,1)*dx(j,2)            

        enddo
        enddo
        enddo

        do k=0,nn(3)+1
        do j=0,nn(2)+1
        do i=0,nn(1)+1

           dcost = cos(xloc(j,2))-cos(xloc(j+1,2))

           dv(i,j,k) = (xloc(i+1,1)**3 - xloc(i,1)**3)/3. * dcost * dx(k,3)  

        enddo
        enddo
        enddo

     endif

  endif
 
  return

end subroutine cal_geom
