!> \file
!! Contains subroutines init_phys() and physical_constants()
!<

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine INIT_PHYS
!
!> Initialises the physical modules required by the
!! simulation.
!<
subroutine init_phys

  use para   , only : mype
  use divers , only : ray,conduc,diffu,grav

  implicit none

  ! Init radiative transfer
#if RAYO==1
  if (RAY) then
     call init_ray
#if PARA==1
     call allocate_array_com_ray
#endif
  endif
#endif

  ! Init conduction
#if CON==1
  if (CONDUC) call init_conduction
#endif

  ! Init diffusion
#if DIF==1
  if (DIFFU) then
     call init_diffusion
     ! Allocate Er_dif
     call allocate_array_dif
  endif
#endif

  ! Init conduction
#if GRA==1
  if (GRAV) call init_gravity
#endif

  return
  
end subroutine init_phys

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine PHYSICAL_CONSTANTS
!
!> Initialises the physical constants required by the
!! simulation.
!<
subroutine physical_constants

  use unites
  use gammas
  use parameters, only : pi

  implicit none
  
  ! cgs basis
  gramme      = 1.0
  centimetre  = 1.0
  seconde     = 1.0
  degre       = 1.0
  gauss       = 1.0
  
  call dimensions
  
  ! cgs in the new basis
  gramme      = 1.0/unitM
  centimetre  = 1.0/unitL
  seconde     = 1.0/unitT
  degre       = 1.0
  gauss       = 1.0
  
  ! additional units
  dyne    =  1.000000e+00 * gramme*centimetre/seconde**2
  erg     =  1.000000e+00 * gramme*centimetre**2/seconde**2
  
  ! physical constants
  c       =  2.997925e+10 * centimetre/seconde     
  c2      =  c*c
  eV      =  1.602200e-12 * erg
  G       =  6.673000e-08 * dyne*centimetre**2/gramme**2
  hplanck =  6.626200e-27 * erg*seconde
  kb      =  1.380620e-16 * erg/degre
  uma     =  1.660531e-24 * gramme 
  a_R     =  7.564640e-15 * erg/centimetre**3/degre**4 
  Lsol    =  3.826800e+33 * erg/seconde
  Msol    =  1.989100e+33 * gramme
  Rsol    =  6.959900e+10 * centimetre
  pc      =  3.085678e+18 * centimetre
  mu0     =  4.000000e+00 * pi * gauss**2 * centimetre * seconde**2 / gramme
  
  ! useful units
  cm2     =  1.000000e+00 * centimetre**2
  cm3     =  1.000000e+00 * centimetre**3
  ms      =  1.000000e+02 * centimetre/seconde
  kms     =  1.000000e+05 * centimetre/seconde
  an      =  3.155760e+07 * seconde
  Mans    =  3.155760e+13 * seconde
  Gans    =  3.155760e+16 * seconde
  kpc     =  3.085678e+21 * centimetre
  Mpc     =  3.085678e+24 * centimetre
  Kelvin  =  1.000000e+00 * degre
  kg      =  1.000000e+03 * gramme
  metre   =  1.000000e+02 * centimetre
  Pascal  =  1.000000e+00 * kg/metre   /seconde**2
  joule   =  1.000000e+00 * kg*metre**2/seconde**2
  H0      =  5.000000e+01 * kms/Mpc
  
  ! useful gamma relations
  g1      = (gamma - 1.0)/(2.0*gamma)
  g2      = (gamma + 1.0)/(2.0*gamma)
  g3      = 2.0*gamma/(gamma - 1.0)
  g4      = 2.0      /(gamma - 1.0)
  g5      = 2.0      /(gamma + 1.0)
  g6      = (gamma - 1.0)/(gamma + 1.0)
  g7      = (gamma - 1.0)/2.0
  g8      = (gamma - 1.0)

  return
  
end subroutine physical_constants
