!> \file
!! Contains subroutine limites()
!<

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine LIMITES
!
!> This routine fills the ghost cells at the domain
!! boundaries with the appropriate values according
!! to the type of boudary condition.
!!
!! Type of boundaries:
!!   - 1 : free-flow
!!   - 2 : periodic
!!   - 3 : reflexive
!!   - 4 : imposed values
!!   - 5 : analytical
!<
subroutine limites

  use divers    , only : verbose
  use var       , only : rho,E,rhou,B,fx
  use parameters, only : ndim,nx,nxmin,nxmax,N_vit, Nbuf,nfx
  use cl

  implicit none
  
  integer :: i,j,iface
  integer, dimension(1:6) :: face
  integer :: nx_min,nx_max,ny_min,ny_max,nz_min,nz_max
  integer :: nnx,nny,nnz

  ! ICL = 1 : FREE-FLOW
  ! ICL = 2 : PERIODIC       (already done in the communications)
  ! ICL = 3 : REFLEXIVE
  ! ICL = 4 : IMPOSED VALUES (inside the ghost cells)
  ! ICL = 5 : ANALYTICAL
  
  if(verbose) write(*,*) 'Start of limites'

  ! Order in which the faces are treated
  if (ndim == 3) then
     face(1) = 1
     face(2) = 2
     face(3) = 5
     face(4) = 6
     face(5) = 4
     face(6) = 3
  else if (ndim ==2) then
     face(1) = 3
     face(2) = 4
     face(3) = 1
     face(4) = 2
     face(5) = 5
     face(6) = 6
  else if ( ndim == 1)  then
     face(1) = 1
     face(2) = 2
  endif

  nx_min = nxmin(1) ; nx_max = nxmax(1)  ; nnx = nx(1)
  ny_min = nxmin(2) ; ny_max = nxmax(2)  ; nny = nx(2)
  nz_min = nxmin(3) ; nz_max = nxmax(3)  ; nnz = nx(3)

  do iface = 1,2*ndim


     ! 1)----------------------------------------
    
     if (face(iface) == 1) then
 
        if (ICL(1) == 1) then ! free-flow
           do i=1-Nbuf,0
              rho (i,ny_min:ny_max,nz_min:nz_max        ) =  rho(1,ny_min:ny_max,nz_min:nz_max        )
              E   (i,ny_min:ny_max,nz_min:nz_max        ) =    E(1,ny_min:ny_max,nz_min:nz_max        )
              rhou(i,ny_min:ny_max,nz_min:nz_max,1:N_vit) = rhou(1,ny_min:ny_max,nz_min:nz_max,1:N_vit)
#if MHD==1
                 B(i,ny_min:ny_max,nz_min:nz_max,1:N_vit) =    B(1,ny_min:ny_max,nz_min:nz_max,1:N_vit)
#endif
#if SCALP==1
                fx(i,ny_min:ny_max,nz_min:nz_max,1:nfx  ) =   fx(1,ny_min:ny_max,nz_min:nz_max,1:nfx )
#endif
           enddo
   
        else if (ICL(1) == 2) then ! periodic (already done in the communications)
   
#if PARA==0  
           do i=1-Nbuf,0
              rho (i,ny_min:ny_max,nz_min:nz_max        ) =  rho(nnx+i,ny_min:ny_max,nz_min:nz_max       )
              E   (i,ny_min:ny_max,nz_min:nz_max        ) =    E(nnx+i,ny_min:ny_max,nz_min:nz_max       )
              rhou(i,ny_min:ny_max,nz_min:nz_max,1:N_vit) = rhou(nnx+i,ny_min:ny_max,nz_min:nz_max,1:N_vit)
#if MHD==1
                 B(i,ny_min:ny_max,nz_min:nz_max,1:N_vit) =    B(nnx+i,ny_min:ny_max,nz_min:nz_max,1:N_vit)
#endif
#if SCALP==1
                fx(i,ny_min:ny_max,nz_min:nz_max,1:nfx ) =    fx(nnx+i,ny_min:ny_max,nz_min:nz_max,1:nfx )
#endif
           enddo
#endif
   
        else if (ICL(1) == 3) then ! reflexive
           do i=1-Nbuf,0
              rho (i,ny_min:ny_max,nz_min:nz_max        ) =   rho(1-i,ny_min:ny_max,nz_min:nz_max        )
              E   (i,ny_min:ny_max,nz_min:nz_max        ) =     E(1-i,ny_min:ny_max,nz_min:nz_max        )
              rhou(i,ny_min:ny_max,nz_min:nz_max,1:N_vit) =  rhou(1-i,ny_min:ny_max,nz_min:nz_max,1:N_vit)
              rhou(i,ny_min:ny_max,nz_min:nz_max,1      ) = -rhou(1-i,ny_min:ny_max,nz_min:nz_max,1      )
#if MHD==1
                 B(i,ny_min:ny_max,nz_min:nz_max,1:N_vit) =     B(1-i,ny_min:ny_max,nz_min:nz_max,1:N_vit)
                 B(i,ny_min:ny_max,nz_min:nz_max,1      ) =    -B(1-i,ny_min:ny_max,nz_min:nz_max,1      )  ! is this true?....
#endif
#if SCALP==1
                fx(i,ny_min:ny_max,nz_min:nz_max,1:nfx  ) =   fx (1-i,ny_min:ny_max,nz_min:nz_max,1:nfx )
#endif
           enddo
        else if (ICL(1) == 4) then ! imposed value
           do i=1-Nbuf,0
              rho (i,ny_min:ny_max,nz_min:nz_max  ) =  rho_l(  1)
              E   (i,ny_min:ny_max,nz_min:nz_max  ) =    E_l(  1)
              do j=1,N_vit
              rhou(i,ny_min:ny_max,nz_min:nz_max,j) = rhou_l(j,1)
              enddo
#if MHD==1
              ! Warning: does not ensure divergence free
              do j=1,N_vit
                 B(i,ny_min:ny_max,nz_min:nz_max,j) =    B_l(j,1)
              enddo
#endif
#if SCALP==1
              do j=1,nfx
                fx(i,ny_min:ny_max,nz_min:nz_max,j) =   fx_l(j,1)
             enddo
#endif
           enddo
   
        else if (ICL(1) == 5) then
           call cl_ana(1)
        endif

     endif

     ! 2)----------------------------------------

     if (face(iface) == 2) then
     
        if (ICL(2) == 1) then ! free-flow
           do i=1,Nbuf
              rho (nnx+i,ny_min:ny_max,nz_min:nz_max        ) =  rho(nnx,ny_min:ny_max,nz_min:nz_max       )
              E   (nnx+i,ny_min:ny_max,nz_min:nz_max        ) =    E(nnx,ny_min:ny_max,nz_min:nz_max       )
              rhou(nnx+i,ny_min:ny_max,nz_min:nz_max,1:N_vit) = rhou(nnx,ny_min:ny_max,nz_min:nz_max,1:N_vit)
#if MHD==1
                 B(nnx+i,ny_min:ny_max,nz_min:nz_max,1:N_vit) =    B(nnx,ny_min:ny_max,nz_min:nz_max,1:N_vit)
#endif
#if SCALP==1
                fx(nnx+i,ny_min:ny_max,nz_min:nz_max,1:nfx  ) =   fx(nnx,ny_min:ny_max,nz_min:nz_max,1:nfx )
#endif
           enddo
   
        else if (ICL(2) == 2) then ! periodic (already done in the communications)
   
#if PARA==0
           do i=1,Nbuf
              rho (nnx+i,ny_min:ny_max,nz_min:nz_max        ) =  rho(i,ny_min:ny_max,nz_min:nz_max        )
              E   (nnx+i,ny_min:ny_max,nz_min:nz_max        ) =    E(i,ny_min:ny_max,nz_min:nz_max        )
              rhou(nnx+i,ny_min:ny_max,nz_min:nz_max,1:N_vit) = rhou(i,ny_min:ny_max,nz_min:nz_max,1:N_vit)
#if MHD==1
                 B(nnx+i,ny_min:ny_max,nz_min:nz_max,1:N_vit) =    B(i,ny_min:ny_max,nz_min:nz_max,1:N_vit)
#endif
#if SCALP==1
                fx(nnx+i,ny_min:ny_max,nz_min:nz_max,1:nfx  ) =   fx(i,ny_min:ny_max,nz_min:nz_max,1:nfx )
#endif
           enddo
#endif
   
        else if (ICL(2) == 3) then ! reflexive
           do i=1,Nbuf
              rho (nnx+i,ny_min:ny_max,nz_min:nz_max        ) =   rho(nnx+1-i,ny_min:ny_max,nz_min:nz_max        )
              E   (nnx+i,ny_min:ny_max,nz_min:nz_max        ) =     E(nnx+1-i,ny_min:ny_max,nz_min:nz_max        )
              rhou(nnx+i,ny_min:ny_max,nz_min:nz_max,1:N_vit) =  rhou(nnx+1-i,ny_min:ny_max,nz_min:nz_max,1:N_vit)       
              rhou(nnx+i,ny_min:ny_max,nz_min:nz_max,1      ) = -rhou(nnx+1-i,ny_min:ny_max,nz_min:nz_max,1      )
#if MHD==1
                 B(nnx+i,ny_min:ny_max,nz_min:nz_max,1:N_vit) =     B(nnx+1-i,ny_min:ny_max,nz_min:nz_max,1:N_vit)
                 B(nnx+i,ny_min:ny_max,nz_min:nz_max,1      ) =    -B(nnx+1-i,ny_min:ny_max,nz_min:nz_max,1      )
#endif
#if SCALP==1
                fx(nnx+i,ny_min:ny_max,nz_min:nz_max,1:nfx  ) =    fx(nnx+1-i,ny_min:ny_max,nz_min:nz_max,1:nfx )
#endif
           enddo
        else if (ICL(2) == 4) then ! imposed value
           do i=1,Nbuf
              rho (nnx+i,ny_min:ny_max,nz_min:nz_max  ) =  rho_l(  2)
              E   (nnx+i,ny_min:ny_max,nz_min:nz_max  ) =    E_l(  2)
              do j=1,N_vit
              rhou(nnx+i,ny_min:ny_max,nz_min:nz_max,j) = rhou_l(j,2)
              enddo
#if MHD==1
              ! Warning: does not ensure divergence free
              do j=1,N_vit
                 B(nnx+i,ny_min:ny_max,nz_min:nz_max,j) =    B_l(j,2)
              enddo
#endif
#if SCALP==1
              do j=1,nfx
                fx(nnx+i,ny_min:ny_max,nz_min:nz_max,j) =   fx_l(j,2)
              enddo
#endif
           enddo
   
        else if (ICL(2) == 5) then
           call cl_ana(2)
        endif
    
     endif
    
!    if (ndim == 1) return

     ! 3)----------------------------------------
     if (face(iface) == 3) then

        if (ICL(3) == 1) then ! free-flow
           do i=1-Nbuf,0
              rho (nx_min:nx_max,i,nz_min:nz_max        ) =  rho(nx_min:nx_max,1,nz_min:nz_max        )
              E   (nx_min:nx_max,i,nz_min:nz_max        ) =    E(nx_min:nx_max,1,nz_min:nz_max        )
              rhou(nx_min:nx_max,i,nz_min:nz_max,1:N_vit) = rhou(nx_min:nx_max,1,nz_min:nz_max,1:N_vit)
#if MHD==1
                 B(nx_min:nx_max,i,nz_min:nz_max,1:N_vit) =    B(nx_min:nx_max,1,nz_min:nz_max,1:N_vit)
#endif
#if SCALP==1
                fx(nx_min:nx_max,i,nz_min:nz_max,1:nfx  ) =   fx(nx_min:nx_max,1,nz_min:nz_max,1:nfx )
#endif
           enddo
   
        else if (ICL(3) == 2) then ! periodic (already done in the communications)
   
#if PARA==0
           do i=1-Nbuf,0
              rho (nx_min:nx_max,i,nz_min:nz_max        ) =  rho(nx_min:nx_max,nny+i,nz_min:nz_max        )
              E   (nx_min:nx_max,i,nz_min:nz_max        ) =    E(nx_min:nx_max,nny+i,nz_min:nz_max        )
              rhou(nx_min:nx_max,i,nz_min:nz_max,1:N_vit) = rhou(nx_min:nx_max,nny+i,nz_min:nz_max,1:N_vit)
#if MHD==1
                 B(nx_min:nx_max,i,nz_min:nz_max,1:N_vit) =    B(nx_min:nx_max,nny+i,nz_min:nz_max,1:N_vit)
#endif
#if SCALP==1
                fx(nx_min:nx_max,i,nz_min:nz_max,1:nfx  ) =   fx(nx_min:nx_max,nny+i,nz_min:nz_max,1:nfx )
#endif
           enddo
#endif
   
        else if (ICL(3) == 3) then ! refexive
           do i=1-Nbuf,0
              rho (nx_min:nx_max,i,nz_min:nz_max        ) =   rho(nx_min:nx_max,1-i,nz_min:nz_max        ) 
              E   (nx_min:nx_max,i,nz_min:nz_max        ) =     E(nx_min:nx_max,1-i,nz_min:nz_max        )
              rhou(nx_min:nx_max,i,nz_min:nz_max,1:N_vit) =  rhou(nx_min:nx_max,1-i,nz_min:nz_max,1:N_vit)
              rhou(nx_min:nx_max,i,nz_min:nz_max,2      ) = -rhou(nx_min:nx_max,1-i,nz_min:nz_max,2      )
#if MHD==1
                 B(nx_min:nx_max,i,nz_min:nz_max,1:N_vit) =     B(nx_min:nx_max,1-i,nz_min:nz_max,1:N_vit)
                 B(nx_min:nx_max,i,nz_min:nz_max,2      ) =    -B(nx_min:nx_max,1-i,nz_min:nz_max,2      )
#endif
#if SCALP==1
                fx(nx_min:nx_max,i,nz_min:nz_max,1:nfx  ) =    fx(nx_min:nx_max,1-i,nz_min:nz_max,1:nfx )
#endif
           enddo
        else if (ICL(3) ==4) then ! imposed value
           do i=1-Nbuf,0
              rho (nx_min:nx_max,i,nz_min:nz_max  ) =  rho_l(  3)
              E   (nx_min:nx_max,i,nz_min:nz_max  ) =    E_l(  3)
              do j=1,N_vit
              rhou(nx_min:nx_max,i,nz_min:nz_max,j) = rhou_l(j,3)
              enddo
#if MHD==1
              ! Warning: does not ensure divergence free
              do j=1,N_vit
                 B(nx_min:nx_max,i,nz_min:nz_max,j) =    B_l(j,3)
              enddo
#endif
#if SCALP==1
              do j=1,nfx
                fx(nx_min:nx_max,i,nz_min:nz_max,j) =   fx_l(j,3)
              enddo
#endif
           enddo
        else if (ICL(3) == 5) then
           call cl_ana(3)
        endif
    
     endif

     !4)----------------------------------------

     if (face(iface) == 4) then

        if (ICL(4) == 1) then ! free-flow
           do i=1,Nbuf
              rho (nx_min:nx_max,nny+i,nz_min:nz_max        ) =  rho(nx_min:nx_max,nny,nz_min:nz_max        )
              E   (nx_min:nx_max,nny+i,nz_min:nz_max        ) =    E(nx_min:nx_max,nny,nz_min:nz_max        )
              rhou(nx_min:nx_max,nny+i,nz_min:nz_max,1:N_vit) = rhou(nx_min:nx_max,nny,nz_min:nz_max,1:N_vit)
#if MHD==1
                 B(nx_min:nx_max,nny+i,nz_min:nz_max,1:N_vit) =    B(nx_min:nx_max,nny,nz_min:nz_max,1:N_vit)
#endif
#if SCALP==1
                fx(nx_min:nx_max,nny+i,nz_min:nz_max,1:nfx  ) =   fx(nx_min:nx_max,nny,nz_min:nz_max,1:nfx )
#endif
           enddo
   
        else if (ICL(4) == 2) then ! periodic (already done in the communications)
   
#if PARA==0
           do i=1,Nbuf
              rho (nx_min:nx_max,nny+i,nz_min:nz_max        ) =  rho(nx_min:nx_max,i,nz_min:nz_max        )
              E   (nx_min:nx_max,nny+i,nz_min:nz_max        ) =    E(nx_min:nx_max,i,nz_min:nz_max        )
              rhou(nx_min:nx_max,nny+i,nz_min:nz_max,1:N_vit) = rhou(nx_min:nx_max,i,nz_min:nz_max,1:N_vit)
#if MHD==1
                 B(nx_min:nx_max,nny+i,nz_min:nz_max,1:N_vit) =    B(nx_min:nx_max,i,nz_min:nz_max,1:N_vit)
#endif
#if SCALP==1
                fx(nx_min:nx_max,nny+i,nz_min:nz_max,1:nfx  ) =   fx(nx_min:nx_max,i,nz_min:nz_max,1:nfx )
#endif
           enddo
#endif
   
        else if (ICL(4) == 3) then ! refexive
           do i=1,Nbuf
              rho (nx_min:nx_max,nny+i,nz_min:nz_max        ) =   rho(nx_min:nx_max,nny+1-i,nz_min:nz_max        )
              E   (nx_min:nx_max,nny+i,nz_min:nz_max        ) =     E(nx_min:nx_max,nny+1-i,nz_min:nz_max        )
              rhou(nx_min:nx_max,nny+i,nz_min:nz_max,1:N_vit) =  rhou(nx_min:nx_max,nny+1-i,nz_min:nz_max,1:N_vit)
              rhou(nx_min:nx_max,nny+i,nz_min:nz_max,2      ) = -rhou(nx_min:nx_max,nny+1-i,nz_min:nz_max,2      )
#if MHD==1
                 B(nx_min:nx_max,nny+i,nz_min:nz_max,1:N_vit) =     B(nx_min:nx_max,nny+1-i,nz_min:nz_max,1:N_vit)
                 B(nx_min:nx_max,nny+i,nz_min:nz_max,2      ) =    -B(nx_min:nx_max,nny+1-i,nz_min:nz_max,2      )
#endif
#if SCALP==1
                fx(nx_min:nx_max,nny+i,nz_min:nz_max,1:nfx  ) =    fx(nx_min:nx_max,nny+1-i,nz_min:nz_max,1:nfx )
#endif
           enddo
        else if (ICL(4) == 4) then ! imposed value
           do i=1,Nbuf
              rho (nx_min:nx_max,nny+i,nz_min:nz_max  ) =  rho_l(  4)
              E   (nx_min:nx_max,nny+i,nz_min:nz_max  ) =    E_l(  4)
              do j=1,N_vit
              rhou(nx_min:nx_max,nny+i,nz_min:nz_max,j) = rhou_l(j,4)
              enddo
#if MHD==1
              ! Warning: does not ensure divergence free
              do j=1,N_vit
                 B(nx_min:nx_max,nny+i,nz_min:nz_max,j) =    B_l(j,4)
              enddo
#endif
#if SCALP==1
              do j=1,nfx
                fx(nx_min:nx_max,nny+i,nz_min:nz_max,j) =   fx_l(j,4)
              enddo
#endif
           enddo
        else if (ICL(4) == 5) then
           call cl_ana(4)
        endif

     endif


     !5)----------------------------------------

     if (face(iface) == 5) then

        if (ICL(5) == 1) then ! free-flow
           do i=1-Nbuf,0
              rho (nx_min:nx_max,ny_min:ny_max,i        ) =  rho(nx_min:nx_max,ny_min:ny_max,1        )
              E   (nx_min:nx_max,ny_min:ny_max,i        ) =    E(nx_min:nx_max,ny_min:ny_max,1        )
              rhou(nx_min:nx_max,ny_min:ny_max,i,1:N_vit) = rhou(nx_min:nx_max,ny_min:ny_max,1,1:N_vit)
#if MHD==1
                 B(nx_min:nx_max,ny_min:ny_max,i,1:N_vit) =    B(nx_min:nx_max,ny_min:ny_max,1,1:N_vit)
#endif
#if SCALP==1
                fx(nx_min:nx_max,ny_min:ny_max,i,1:nfx  ) =   fx(nx_min:nx_max,ny_min:ny_max,1,1:nfx )
#endif
           enddo
   
        else if (ICL(5) == 2) then ! periodic (already done in the communications)
   
#if PARA==0
           do i=1-Nbuf,0
              rho (nx_min:nx_max,ny_min:ny_max,i        ) =  rho(nx_min:nx_max,ny_min:ny_max,nnz+i        )
              E   (nx_min:nx_max,ny_min:ny_max,i        ) =    E(nx_min:nx_max,ny_min:ny_max,nnz+i        )
              rhou(nx_min:nx_max,ny_min:ny_max,i,1:N_vit) = rhou(nx_min:nx_max,ny_min:ny_max,nnz+i,1:N_vit)
#if MHD==1
                 B(nx_min:nx_max,ny_min:ny_max,i,1:N_vit) =    B(nx_min:nx_max,ny_min:ny_max,nnz+i,1:N_vit)
#endif
#if SCALP==1
                fx(nx_min:nx_max,ny_min:ny_max,i,1:nfx  ) =   fx(nx_min:nx_max,ny_min:ny_max,nnz+i,1:nfx )
#endif
           enddo
#endif
   
        else if (ICL(5) == 3) then ! refexive
           do i=1-Nbuf,0
              rho (nx_min:nx_max,ny_min:ny_max,i        ) =   rho(nx_min:nx_max,ny_min:ny_max,1-i        )
              E   (nx_min:nx_max,ny_min:ny_max,i        ) =     E(nx_min:nx_max,ny_min:ny_max,1-i        )
              rhou(nx_min:nx_max,ny_min:ny_max,i,1:N_vit) =  rhou(nx_min:nx_max,ny_min:ny_max,1-i,1:N_vit)
              rhou(nx_min:nx_max,ny_min:ny_max,i,3      ) = -rhou(nx_min:nx_max,ny_min:ny_max,1-i,3      )
#if MHD==1
                 B(nx_min:nx_max,ny_min:ny_max,i,1:N_vit) =     B(nx_min:nx_max,ny_min:ny_max,1-i,1:N_vit)
                 B(nx_min:nx_max,ny_min:ny_max,i,3      ) =    -B(nx_min:nx_max,ny_min:ny_max,1-i,3      )
#endif
#if SCALP==1
                fx(nx_min:nx_max,ny_min:ny_max,i,1:nfx  ) =    fx(nx_min:nx_max,ny_min:ny_max,1-i,1:nfx )
#endif
           enddo
        else if (ICL(5) == 4) then ! imposed value
           do i=1-Nbuf,0
              rho (nx_min:nx_max,ny_min:ny_max,i  ) =  rho_l(  5)
              E   (nx_min:nx_max,ny_min:ny_max,i  ) =    E_l(  5)
              do j=1,N_vit
              rhou(nx_min:nx_max,ny_min:ny_max,i,j) = rhou_l(j,5)
              enddo
#if MHD==1
              ! Warning: does not ensure divergence free
              do j=1,N_vit
                 B(nx_min:nx_max,ny_min:ny_max,i,j) =    B_l(j,5)
              enddo
#endif
#if SCALP==1
              do j=1,nfx
                fx(nx_min:nx_max,ny_min:ny_max,i,j) =   fx_l(j,5)
              enddo
#endif
           enddo
        else if (ICL(5) == 5) then
           call cl_ana(5)
        endif

     endif


     !6)----------------------------------------

     if (face(iface) == 6) then

        if (ICL(6) == 1) then ! free-flow
           do i=1,Nbuf
              rho (nx_min:nx_max,ny_min:ny_max,nnz+i        ) =  rho(nx_min:nx_max,ny_min:ny_max,nnz        )
              E   (nx_min:nx_max,ny_min:ny_max,nnz+i        ) =    E(nx_min:nx_max,ny_min:ny_max,nnz        )
              rhou(nx_min:nx_max,ny_min:ny_max,nnz+i,1:N_vit) = rhou(nx_min:nx_max,ny_min:ny_max,nnz,1:N_vit)
#if MHD==1
                 B(nx_min:nx_max,ny_min:ny_max,nnz+i,1:N_vit) =    B(nx_min:nx_max,ny_min:ny_max,nnz,1:N_vit)
#endif
#if SCALP==1
                fx(nx_min:nx_max,ny_min:ny_max,nnz+i,1:nfx  ) =   fx(nx_min:nx_max,ny_min:ny_max,nnz,1:nfx )
#endif
           enddo
   
        else if (ICL(6) == 2) then ! periodic (already done in the communications)
   
#if PARA==0
           do i=1,Nbuf
              rho (nx_min:nx_max,ny_min:ny_max,nnz+i        ) =  rho(nx_min:nx_max,ny_min:ny_max,i        )
              E   (nx_min:nx_max,ny_min:ny_max,nnz+i        ) =    E(nx_min:nx_max,ny_min:ny_max,i        )
              rhou(nx_min:nx_max,ny_min:ny_max,nnz+i,1:N_vit) = rhou(nx_min:nx_max,ny_min:ny_max,i,1:N_vit)
#if MHD==1
                 B(nx_min:nx_max,ny_min:ny_max,nnz+i,1:N_vit) =    B(nx_min:nx_max,ny_min:ny_max,i,1:N_vit)
#endif
#if SCALP==1
                fx(nx_min:nx_max,ny_min:ny_max,nnz+i,1:nfx  ) =   fx(nx_min:nx_max,ny_min:ny_max,i,1:nfx )
#endif
           enddo
#endif
   
        else if (ICL(6) == 3) then ! refexive
           do i=1,Nbuf
              rho (nx_min:nx_max,ny_min:ny_max,nnz+i        ) =   rho(nx_min:nx_max,ny_min:ny_max,nnz+1-i        )
              E   (nx_min:nx_max,ny_min:ny_max,nnz+i        ) =     E(nx_min:nx_max,ny_min:ny_max,nnz+1-i        )
              rhou(nx_min:nx_max,ny_min:ny_max,nnz+i,1:N_vit) =  rhou(nx_min:nx_max,ny_min:ny_max,nnz+1-i,1:N_vit)
              rhou(nx_min:nx_max,ny_min:ny_max,nnz+i,3      ) = -rhou(nx_min:nx_max,ny_min:ny_max,nnz+1-i,3      )
#if MHD==1
                 B(nx_min:nx_max,ny_min:ny_max,nnz+i,1:N_vit) =     B(nx_min:nx_max,ny_min:ny_max,nnz+1-i,1:N_vit)
                 B(nx_min:nx_max,ny_min:ny_max,nnz+i,3      ) =    -B(nx_min:nx_max,ny_min:ny_max,nnz+1-i,3      )
#endif
#if SCALP==1
                fx(nx_min:nx_max,ny_min:ny_max,nnz+i,1:nfx  ) =    fx(nx_min:nx_max,ny_min:ny_max,nnz+1-i,1:nfx )
#endif
           enddo
        else if (ICL(6) == 4) then ! imposed value
           do i=1,Nbuf
              rho (nx_min:nx_max,ny_min:ny_max,nnz+i  ) =  rho_l(  6)
              E   (nx_min:nx_max,ny_min:ny_max,nnz+i  ) =    E_l(  6)
              do j=1,N_vit
              rhou(nx_min:nx_max,ny_min:ny_max,nnz+i,j) = rhou_l(j,6)
              enddo
#if MHD==1
              ! Warning: does not ensure divergence free
              do j=1,N_vit
                 B(nx_min:nx_max,ny_min:ny_max,nnz+i,j) =    B_l(j,6)
              enddo
#endif
#if SCALP==1
              do j=1,nfx
                fx(nx_min:nx_max,ny_min:ny_max,nnz+i,j) =   fx_l(j,6)
              enddo
#endif
           enddo
        else if (ICL(6) == 5) then
           call cl_ana(6)
        endif

     endif

  enddo
  
  if(verbose) write(*,*) 'End of limites'

  return

end subroutine limites
