!> \file
!! Contains subroutine limites_gen()
!<

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine LIMITES_GEN
!
!> This routine is a generic form of the limites()
!! subroutine.
!!
!! Type of boundaries:
!!   - 1 : free-flow
!!   - 2 : periodic
!!   - 3 : reflexive
!!   - 4 : imposed values
!!   - 5 : analytical
!<
subroutine limites_gen(Tab,icl,Tab_l,n)
  
  use parameters , only : ndim,nx,nxmin,nxmax,Nbuf
  use divers     , only : verbose

  implicit none

  integer                        :: n
  integer, dimension(1:6       ) :: icl
  real   , dimension(n,1:2*ndim) :: Tab_l
  Real   , dimension(nxmin(1):nxmax(1),nxmin(2):nxmax(2),nxmin(3):nxmax(3),n) :: Tab

  integer  :: i,l,iface
  integer, dimension(1:6) :: face

  ! ICL = 1 : FREE-FLOW
  ! ICL = 2 : PERIODIC       (already done in the communications)
  ! ICL = 3 : REFLEXIVE
  ! ICL = 4 : IMPOSED VALUES (inside the ghost cells)
  ! ICL = 5 : ANALYTICAL

  if(verbose) write(*,*) 'Start of limites_gen'

  ! Order in which the faces are treated
  face(1) = 1
  face(2) = 2
  face(3) = 3
  face(4) = 4
  face(5) = 5
  face(6) = 6

  do iface = 1,2*ndim

    ! 1) ------------------

    if (face(iface) == 1) then
    
    if   (ICL(1) == 1) then ! free-flow
       do i=1-Nbuf,0
          Tab(i,nxmin(2):nxmax(2),nxmin(3):nxmax(3),1:n) = Tab(1      ,nxmin(2):nxmax(2),nxmin(3):nxmax(3),1:n) 
       enddo
#if PARA==0
    else if (ICL(1) == 2) then ! periodic 
       do i=1-Nbuf,0
          Tab(i,nxmin(2):nxmax(2),nxmin(3):nxmax(3),1:n) = Tab(nx(1)+i,nxmin(2):nxmax(2),nxmin(3):nxmax(3),1:n)
       enddo
#endif
    else if (ICL(1) == 3) then ! refexive
       do i=1-Nbuf,0
          Tab(i,nxmin(2):nxmax(2),nxmin(3):nxmax(3),1:n) = Tab(1-i    ,nxmin(2):nxmax(2),nxmin(3):nxmax(3),1:n)
       enddo

    else if (ICL(1) == 4) then ! imposed values
       do l=1,n
       do i=1-Nbuf,0
          Tab(i,nxmin(2):nxmax(2),nxmin(3):nxmax(3),l) = Tab_l(l,1)
       enddo
       enddo
    else if (ICL(1) == 5) then ! analytical
       call cl_ana_gen(1,Tab,Tab_l,n)
    endif

    endif

    !2)----------------------------------------

    if (face(iface) == 2) then

    if   (ICL(2) == 1) then ! free-flow
       do i=1,Nbuf
          Tab(nx(1)+i,nxmin(2):nxmax(2),nxmin(3):nxmax(3)  ,1:n     ) = Tab(nx(1),nxmin(2):nxmax(2),nxmin(3):nxmax(3) ,1:n      )
       enddo
#if PARA==0
    else if (ICL(2) == 2) then ! periodic 
       do i=1,Nbuf
          Tab(nx(1)+i,nxmin(2):nxmax(2),nxmin(3):nxmax(3)  ,1:n     ) = Tab(i,nxmin(2):nxmax(2),nxmin(3):nxmax(3)   ,1:n    )
       enddo
#endif
    else if (ICL(2) == 3) then ! refexive
       do i=1,Nbuf
          Tab(nx(1)+i,nxmin(2):nxmax(2),nxmin(3):nxmax(3)  ,1:n     ) = Tab(nx(1)+1-i,nxmin(2):nxmax(2),nxmin(3):nxmax(3)    ,1:n   )
       enddo

    else if (ICL(2) == 4) then ! imposed values
       do l=1,n
       do i=1,Nbuf
          Tab(nx(1)+i,nxmin(2):nxmax(2),nxmin(3):nxmax(3)  ,l       ) = Tab_l(l,2)
       enddo
       enddo
    else if (ICL(2) == 5) then ! analytical
       call cl_ana_gen(2,Tab,Tab_l,n)
    endif

    endif

    !3)----------------------------------------

    if (face(iface) == 3) then

    if   (ICL(3) == 1) then ! free-flow
       do i=1-Nbuf,0
          Tab(nxmin(1):nxmax(1),i,nxmin(3):nxmax(3) ,1:n      ) = Tab(nxmin(1):nxmax(1),1,nxmin(3):nxmax(3)   ,1:n    )
       enddo
#if PARA==0
    else if (ICL(3) == 2) then ! periodic 
       do i=1-Nbuf,0
          Tab(nxmin(1):nxmax(1),i,nxmin(3):nxmax(3)    ,1:n   ) = Tab(nxmin(1):nxmax(1),nx(2)+i,nxmin(3):nxmax(3)  ,1:n     )
       enddo
#endif
    else if (ICL(3) == 3) then ! refexive
       do i=1-Nbuf,0
          Tab(nxmin(1):nxmax(1),i,nxmin(3):nxmax(3)    ,1:n   ) =  Tab(nxmin(1):nxmax(1),1-i,nxmin(3):nxmax(3)    ,1:n   )
       enddo

    else if (ICL(3) == 4) then ! imposed values
       do l=1,n
       do i=1-Nbuf,0
          Tab(nxmin(1):nxmax(1),i,nxmin(3):nxmax(3)    ,l     ) =  Tab_l(l,3)
       enddo
       enddo
    else if (ICL(3) == 5) then ! analytical
       call cl_ana_gen(3,Tab,Tab_l,n)
    endif
    
    endif

    !4)----------------------------------------

    if (face(iface) == 4) then

    if   (ICL(4) == 1) then ! free-flow
       do i=1,Nbuf
          Tab(nxmin(1):nxmax(1),nx(2)+i,nxmin(3):nxmax(3)    ,1:n   ) = Tab(nxmin(1):nxmax(1),nx(2),nxmin(3):nxmax(3)  ,1:n     )
       enddo
#if PARA==0
    else if (ICL(4) == 2) then ! periodic 
       do i=1,Nbuf
          Tab(nxmin(1):nxmax(1),nx(2)+i,nxmin(3):nxmax(3)    ,1:n   ) = Tab(nxmin(1):nxmax(1),i,nxmin(3):nxmax(3)   ,1:n    )
       enddo
#endif
    else if (ICL(4) == 3) then ! refexive
       do i=1,Nbuf
          Tab(nxmin(1):nxmax(1),nx(2)+i,nxmin(3):nxmax(3)   ,1:n    ) =  Tab(nxmin(1):nxmax(1),nx(2)+1-i,nxmin(3):nxmax(3)  ,1:n     )
       enddo

    else if (ICL(4) == 4) then ! imposed values
       do l=1,n
       do i=1,Nbuf
          Tab(nxmin(1):nxmax(1),nx(2)+i,nxmin(3):nxmax(3)   ,l      ) = Tab_l(l,4)
       enddo
       enddo
    else if (ICL(4) == 5) then ! analytical
       call cl_ana_gen(4,Tab,Tab_l,n)
    endif

    endif

    !5)----------------------------------------

    if (face(iface) == 5) then

    if   (ICL(5) == 1) then ! free-flow
       do i=1-Nbuf,0
          Tab(nxmin(1):nxmax(1),nxmin(2):nxmax(2),i ,1:n      ) = Tab(nxmin(1):nxmax(1),nxmin(2):nxmax(2),1  ,1:n     )
       enddo
#if PARA==0
    else if (ICL(5) == 2) then ! periodic 
     do i=1-Nbuf,0
        Tab(nxmin(1):nxmax(1),nxmin(2):nxmax(2),i    ,1:n   ) = Tab(nxmin(1):nxmax(1),nxmin(2):nxmax(2),nx(3)+i   ,1:n    )
     enddo
#endif
    else if (ICL(5) == 3) then ! refexive
       do i=1-Nbuf,0
          Tab(nxmin(1):nxmax(1),nxmin(2):nxmax(2),i    ,1:n   ) =  Tab(nxmin(1):nxmax(1),nxmin(2):nxmax(2),1-i  ,1:n     )
       enddo
    else if (ICL(5) == 4) then ! imposed values
       do l=1,n
       do i=1-Nbuf,0
          Tab(nxmin(1):nxmax(1),nxmin(2):nxmax(2),i    ,l     ) =  Tab_l(l,5)
       enddo
       enddo
    else if (ICL(5) == 5) then ! analytical
       call cl_ana_gen(5,Tab,Tab_l,n)
    endif

   endif


    !6)----------------------------------------

    if (face(iface) == 6) then

    if   (ICL(6) == 1) then ! free-flow
       do i=1,Nbuf
          Tab(nxmin(1):nxmax(1),nxmin(2):nxmax(2),nx(3)+i   ,1:n    ) = Tab(nxmin(1):nxmax(1),nxmin(2):nxmax(2),nx(3)  ,1:n     )
       enddo
#if PARA==0
    else if (ICL(6) == 2) then ! periodic 
     do i=1,Nbuf
        Tab(nxmin(1):nxmax(1),nxmin(2):nxmax(2),nx(3)+i   ,1:n    ) = Tab(nxmin(1):nxmax(1),nxmin(2):nxmax(2),i  ,1:n     )
     enddo
#endif
    else if (ICL(6) == 3) then ! refexive
       do i=1,Nbuf
          Tab(nxmin(1):nxmax(1),nxmin(2):nxmax(2),nx(3)+i  ,1:n     ) =  Tab(nxmin(1):nxmax(1),nxmin(2):nxmax(2),nx(3)+1-i   ,1:n    )
       enddo
    else if (ICL(6) == 4) then! imposed values
       do l=1,n
       do i=1,Nbuf
          Tab(nxmin(1):nxmax(1),nxmin(2):nxmax(2),nx(3)+i   ,l      ) = Tab_l(l,6)
       enddo
       enddo
    else if (ICL(6) == 5) then ! analytical
       call cl_ana_gen(6,Tab,Tab_l,n)
    endif

    endif

    enddo

  if(verbose) write(*,*) 'End of limites_gen'

  return

end subroutine limites_gen
