!> \file
!! Contains main program HERACLES and subroutines
!! compute_cputime(), output() and close_program()
!! and functions Arret(), MPI_Wtime()
!!
!! This file contains the main program and the routines used to allocate and deallocate the
!! variable arrays, and open and close the log files.
!<

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!> \mainpage
!! <CENTER> <b>H</b>ydrodynamique <b>E</b>ulerienne <b>RA</b>diative en <b>C</b>oordonn&eacute;es cy<b>L</b>indriques <b>E</b>t <b>S</b>ph&eacute;riques\n\n
!!          Copyright &copy; CEA & Edouard Audit\n
!!          (edouard.audit@cea.fr)\n
!!          30/06/2010\n\n
!!          This software is governed by the CeCILL license\n
!!          (see README and LICENSE files for details)
!! </CENTER>
!!
!! Main contributors to the code:
!!      - Code Architecture  : Edouard Audit
!!      - Parallelization    : Edouard Audit
!!      - Hydrodynamics      : Edouard Audit
!!      - Radiative transfer : Matthias Gonz&aacute;lez, Edouard Audit & Neil Vaytet
!!      - MHD                : Sebastien Fromang, Patrick Hennebelle & Romain Teyssier
!!      - Gravity            : Pascal Tremblin
!!      - HDF5 output        : Bruno Thooris
!<

!  Program HERACLES:
!
!> Main program
!!
!! 3D MPI Eulerian second-order Godunov hydrodynamic simulation code in cartesian, 
!! spherical and cylindrical coordinates, including MHD, radiative transfer and gravity.
!<
program heracles
!
! heracles
!
  use parameters, only : ndim
  use var
  use varold
  use varray
  use divers
  use divers_ray, only : dt_ray,dt_ray_exp,imp_ray,exp_ray,rad_trans_model
  use diffusion , only : dt_dif,dif_imp
  use unites    , only : seconde
  use para      , only : ncpu,mype
  use mfilm     , only : FILM
  use cputime

  implicit none

#if PARA==1
  include 'mpif.h'
  integer :: info,istop,Arret
#else
  real :: MPI_Wtime
#endif

  integer :: ierr

  integer :: kk,kmax

  real    :: tr1,tr2

  logical           :: monit=.false.,calculate_dt=.true.
  character(len=3)  :: format_sortie
  
!********************************************************************************************
!********************************************************************************************

#if PARA==1
  ! Init MPI
  call MPI_INIT(info)
  call MPI_COMM_RANK(MPI_COMM_WORLD,mype,info)
  call MPI_COMM_SIZE(MPI_COMM_WORLD,ncpu,info)
#else
  ncpu = 1
  mype = 0
#endif

  tcpu_ini = MPI_Wtime(ierr)

  ! Greetings on screen
  if (mype == 0) then
     write(*,*)
     write(*,'(a)') '##########################################################################'
     write(*,'(a)') '                                                                          '
     write(*,'(a)') '      _/  _/  _/_/_/  _/_/_/    _/      _/_/  _/      _/_/_/  _/_/_/      '
     write(*,'(a)') '      _/  _/  _/      _/  _/  _/  _/  _/      _/      _/      _/          '
     write(*,'(a)') '      _/_/_/  _/_/    _/_/    _/_/_/  _/      _/      _/_/    _/_/_/      '
     write(*,'(a)') '      _/  _/  _/      _/  _/  _/  _/  _/      _/      _/          _/      '
     write(*,'(a)') '      _/  _/  _/_/_/  _/  _/  _/  _/    _/_/  _/_/_/  _/_/_/  _/_/_/      '
     write(*,'(a)') '                                                                          '
     write(*,'(a)') '              Copyright: E. Audit - CEA/DSM/IRFU/SAp, France              '
     write(*,'(a)') '                         Version: 2.0   (06/2010)                         '
     write(*,'(a)') '                                                                          '
     write(*,'(a)') '##########################################################################'
     write(*,*)
     tr1 = MPI_Wtime(ierr)
  endif

  ! Read Params
  call read_params(kmax,format_sortie)

  ! Checks flags
  call check_flags(format_sortie)

  ! Set dimensions and units
  call physical_constants
  if(.not.reprise) call set_units_out

  ! Calculate block decomposition structure
  call cpu

  ! Domain decomposition
  call init_para

  ! Allocate arrays
  call allocate_array
#if PARA==1
  call allocate_array_com
  call allocate_array_com_gen
#endif

  ! Physics modules initialisation
  call init_phys

  ! User defined initialisation
  call user_init

  ! Init problem
  call init
  kk = 0

  if (FILM) call init_film

  ! Restart
  if (reprise) then
     if (format_sortie=='bin') then
        call restart_new(kk)
     else if (format_sortie=='hdf') then
        call rd_restart_h5(kk)
     endif
  endif
  
  ! Compute output times
  call init_out

  ! Compute buffer cells locations
  call def_geom

  ! First evaluation of gravitational potential
#if GRA==1
  if (GRAV) call gravity_step(.true.,temps_grav,temps_comm)
#endif

  ! Communications and boundary conditions
  call communications_all

  if(mype == 0) call print_configuration(format_sortie)

#if RAYO==1
  if (RAY) call pasdt_ray_exp

  if (reprise .and. RAY) then
     dt_ray = dt
     if(dt.gt.dt_ray_exp) then 
        IMP_RAY=.true. 
        EXP_RAY=.false. 
     else 
        IMP_RAY=.false. 
        EXP_RAY=.true. 
     endif
     calculate_dt = .false.
  endif
#endif

#if DIF==1
  if (reprise .and. DIFFU .and. dif_imp) then
     dt_dif = dt
     calculate_dt = .false.
  endif
#endif
     
  temps1 = MPI_Wtime(ierr)

  if (mype == 0) then
     tr2 = MPI_Wtime(ierr)
     write(*,*)
     write(*,*) 'CPU time for initialization : ',tr2-tr1
     write(*,*)
     write(*,'(a)') '##########################################################################'
  endif

  ! Set max loop index
  kmax = kk + kmax

  ! MAIN LOOP OVER TIME
  do while (temps<tend .and. kk<kmax) 
  
     noutput = 0
     
     if(mype==0) write(*,'(a)') '**************************************************************************'

     ! Increment loop index
     kk = kk + 1

     call pasdt(calculate_dt)

     ! Copy state vector in varold
     rhoold  = rho
     Eold    = E
     rhouold = rhou
#if SCALP==1
     fxold = fx
#endif
#if MHD==1
     Bold  = B
#endif
#if RAYO==1
     if (RAY) then
        Erayold = Eray
        Frayold = Fray
        Tgazold = Tgaz
     endif
#endif       

     ! Solving for base physical modules
     call hydro_step(temps_hydro,temps_comm,monit)

#if RAYO==1
     if (RAY)then
        if(mype==0) write(*,'(a)') '************************* '//rad_trans_model//' radiative transfer **************************'
        call ray_step(temps_ray,temps_comm,monit)
     endif
#endif

#if DIF==1
     if (DIFFU)then
        if(mype==0) write(*,'(a)') '************************* Flux limited diffusion *************************'
        call diffusion_step(temps_dif,temps_comm,monit)
        if(mype==0) write(*,'(a)') '**************************************************************************'
     endif
#endif

#if GRA==1
     if (GRAV) then
        if(mype==0) write(*,'(a)') '******************************** Gravity *********************************'
        call gravity_step(.false.,temps_grav,temps_comm)
        if(mype==0) write(*,'(a)') '**************************************************************************'
     endif
#endif

#if CON==1
     if (CONDUC)then
        if(mype==0) write(*,'(a)') '*************************** Thermal conduction ***************************'
        call conduction_step(temps_conduc,temps_comm)
        if(mype==0) write(*,'(a)') '**************************************************************************'
     endif
#endif

     ! User defined step
     call user_step

     ! Update time
     temps = temps + dt

     if (mype == 0) then
        noutput = noutput + 1
        write(terminal_output(noutput),'(a,I7,2(2x,ES10.3))') 'Ndt, time, dt (s) =',kk, temps/seconde, dt/seconde
        if(ray)then
           noutput = noutput + 1
           write(terminal_output(noutput),'(a, 1(2x,F10.1))') 'CFL factor:    ray/ray_exp=',dt_ray/dt_ray_exp
        endif
        call output_to_terminal
     endif

     if (verbose) write(*,*) 'Performing outputs...'

     ! Fill supplementary array for output if needed
     call fill_new_arrays

     ! Output film
     if (FILM) then
        if (format_sortie=='hdf') then
           call film_hdf
        else        
           call film_
        endif
     endif

     ! Output data
     if (mod(kk,Nimp)==0 .or. temps>tout(iout)) call output(kk,format_sortie,temps_io,mype)

     ! User defined output
     call user_output(kk)

     ! if Tcpu>Tcpu_max, save data and clean stop
#if PARA==1
     istop = Arret()
     if (istop == 1) then
        call output(kk,format_sortie,temps_io,mype)
        call close_program(mype,'Tcpu_max has been reached. Execution stopped to ensure clean exit.')
     endif
#endif

     ! Compute timing
     if (mod(kk,max(100,Nimp)) == 0) call compute_cputime(mype)

     calculate_dt = .true.

     if(mype==0) write(*,'(a)') '**************************************************************************'
     if(mype==0) write(*,*)

     if (verbose) write(*,*) 'End of loop over time.'
     
  enddo
  !END OF MAIN LOOP OVER TIME

  if(mype==0) write(*,*) 'time,tend =',temps/seconde,tend/seconde

  ! Compute timing
  call compute_cputime(mype)

  if(mype==0) call close_program(mype,'End of program reached NORMALLY')

end program heracles

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Function MPI_WTIME
!
!> Computes the MPI World time.
!<
#if PARA==0
function MPI_Wtime(ierr)

  real                  :: MPI_Wtime
  integer               :: ierr
  character(len=8 )     :: date
  character(len=10)     :: time
  character(len=5 )     :: zone
  integer, dimension(8) :: values

!********************************
!            Values
!********************************
!(1): year
!(2): month number
!(3): day number
!(4): gap in minutes with respect to universal time
!(5): hour         (from 0 to 23)
!(6): minutes      (from 0 to 59)
!(7): seconds      (from 0 to 59)
!(8): milliseconds (from 0 to 999)
!********************************

  call date_and_time (date, time, zone, values)
  
  MPI_Wtime = values(5) * 3600. + values(6) * 60. + values(7) + 0.001 * values(8)
  
end function MPI_Wtime
#endif

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Function ARRET
!
!> This routine is used to cleanly exit the program.
!<
function Arret()
  
  use para  , only : mype
  use divers, only : tcpu_max, tcpu_ini, tcpu_last

  implicit none
  
  real    :: tps, tps_max
  integer :: hh1, mm1, hh2, mm2,hh3,mm3
  real    :: tt, ss1,ss2,ss3

#if PARA==1
  include 'mpif.h'
  integer :: info
#else
  real :: MPI_Wtime
#endif

  integer :: arret, ierr

  ierr = 0

  arret = 0

  tps = Mpi_wtime(ierr) - Tcpu_ini

#if PARA==1
  call MPI_ALLREDUCE(tps,tps_max,1,MPI_DOUBLE_PRECISION,MPI_MAX,MPI_COMM_WORLD,info)
#else
  tps_max = tps
#endif


  tt = tps_max
  hh1 = int(tt/3600.) ; tt=tt-hh1*3600.
  mm1 = int(tt/60.  ) ; ss1=(tt -mm1*60.)

  tt = tcpu_max
  hh2 = int(tt/3600.) ; tt=tt-hh2*3600.
  mm2 = int(tt/60.  ) ; ss2=(tt -mm2*60.)

  tt = tps_max - Tcpu_last
  hh3 = int(tt/3600.) ; tt=tt-hh3*3600.
  mm3 = int(tt/60.  ) ; ss3=(tt -mm3*60.)
  
  if (mype == 0) write(*,111) hh1,mm1,ss1,hh2,mm2,ss2
111 format('CPU time passed/max : ',I3,'h ',I2,'m ',F5.2,'s    /  ' ,I2,'h ',I2,'m ',F5.2,'s ') 

  if (mype == 0) write(*,112) hh3,mm3,ss3
112 format('CPU time since last display : ',I2,'h ',I2,'m ',F5.2,'s  ') 

  if (tps_max > tcpu_max) then
     arret = 1
  endif

  tcpu_last = tps_max

end function Arret

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine COMPUTE_CPUTIME
!
!> This routine calculates the fractions of the computation
!! time spent in the different parts of the code.
!<
subroutine compute_cputime(mype)

  use cputime

  implicit none

  integer :: mype,ierr
  real    :: temps2,temps_tot
  real    :: temps_tot_glob,temps_hydro_glob,temps_ref_glob,temps_conduc_glob
  real    :: temps_ray_glob,temps_dif_glob,temps_comm_glob,temps_io_glob

#if PARA==1
  include 'mpif.h'
  integer :: info
  real :: tr1,tr2
#else
  real :: MPI_Wtime
#endif

  temps2 = MPI_Wtime(ierr)
  temps_tot = temps2 - temps1
#if PARA==1
  tr1 = MPI_Wtime(ierr)
  info = 0 ; call MPI_ALLREDUCE(temps_tot   ,temps_tot_glob   ,1,MPI_DOUBLE_PRECISION,MPI_SUM,MPI_COMM_WORLD,info)
  info = 0 ; call MPI_ALLREDUCE(temps_hydro ,temps_hydro_glob ,1,MPI_DOUBLE_PRECISION,MPI_SUM,MPI_COMM_WORLD,info)
  info = 0 ; call MPI_ALLREDUCE(temps_ref   ,temps_ref_glob   ,1,MPI_DOUBLE_PRECISION,MPI_SUM,MPI_COMM_WORLD,info)
  info = 0 ; call MPI_ALLREDUCE(temps_conduc,temps_conduc_glob,1,MPI_DOUBLE_PRECISION,MPI_SUM,MPI_COMM_WORLD,info)
  info = 0 ; call MPI_ALLREDUCE(temps_ray   ,temps_ray_glob   ,1,MPI_DOUBLE_PRECISION,MPI_SUM,MPI_COMM_WORLD,info)
  info = 0 ; call MPI_ALLREDUCE(temps_dif   ,temps_dif_glob   ,1,MPI_DOUBLE_PRECISION,MPI_SUM,MPI_COMM_WORLD,info)
  info = 0 ; call MPI_ALLREDUCE(temps_comm  ,temps_comm_glob  ,1,MPI_DOUBLE_PRECISION,MPI_SUM,MPI_COMM_WORLD,info)
  info = 0 ; call MPI_ALLREDUCE(temps_io    ,temps_io_glob    ,1,MPI_DOUBLE_PRECISION,MPI_SUM,MPI_COMM_WORLD,info)
  tr2 = MPI_Wtime(ierr)
  temps_comm = tr2 - tr1
#else
  temps_tot_glob    = temps_tot
  temps_hydro_glob  = temps_hydro
  temps_ref_glob    = temps_ref
  temps_conduc_glob = temps_conduc
  temps_ray_glob    = temps_ray
  temps_dif_glob    = temps_dif
  temps_comm_glob   = temps_comm
  temps_io_glob     = temps_io
#endif
  if(mype==0) then
     write(*,050) 
     write(*,051) temps_tot_glob   , temps_tot_glob   /temps_tot_glob*100.
     write(*,052) temps_hydro_glob , temps_hydro_glob /temps_tot_glob*100.
     write(*,053) temps_ref_glob   , temps_ref_glob   /temps_tot_glob*100.
     write(*,054) temps_conduc_glob, temps_conduc_glob/temps_tot_glob*100.
     write(*,055) temps_ray_glob   , temps_ray_glob   /temps_tot_glob*100.
     write(*,056) temps_dif_glob   , temps_dif_glob   /temps_tot_glob*100.
     write(*,057) temps_comm_glob  , temps_comm_glob  /temps_tot_glob*100.
     write(*,058) temps_io_glob    , temps_io_glob    /temps_tot_glob*100.
  endif
050     format('CPU time :')
051     format('Total         CPU time : ',E10.4,' = ',F9.3,'% of total time')
052     format('Hydro         CPU time : ',E10.4,' = ',F9.3,'% of total time')
053     format('Ref.          CPU time : ',E10.4,' = ',F9.3,'% of total time')
054     format('Conduc        CPU time : ',E10.4,' = ',F9.3,'% of total time')
055     format('Radiative     CPU time : ',E10.4,' = ',F9.3,'% of total time')
056     format('Diffusion     CPU time : ',E10.4,' = ',F9.3,'% of total time')
057     format('Communication CPU time : ',E10.4,' = ',F9.3,'% of total time')
058     format('I/O           CPU time : ',E10.4,' = ',F9.3,'% of total time')

  return
  
end subroutine compute_cputime

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine OUTPUT
!
!> This routine calls the relevant output routines to
!! write binary or HDF5 files.
!<
subroutine output(kk,output_format,t_io,mype)

  implicit none

  integer :: kk,mype
  real    :: t_io
  character(len=3)  :: output_format

#if PARA==1
  include 'mpif.h'
#else
  real :: MPI_Wtime
#endif

  integer :: ierr
  real :: tr1,tr2

  tr1 = MPI_Wtime(ierr)
  if (output_format=='bin') then
     call aff_new(kk)
  else if (output_format=='hdf') then
     if(mype==0) call wrt_main_h5(kk)
     call wrt_restart_h5(kk)
  endif
  tr2 = MPI_Wtime(ierr)
  t_io = t_io + tr2-tr1

  return

end subroutine output

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine CLOSE_PROGRAM
!
!> This routine closes the HERACLES program cleanly taking
!! into account whether it is using MPI or not. It is also
!! used to close any opened monitoring files if the program
!! exits in the middle of a simulation. This routine should
!! replace every call to 'stop' in the entire source.
!<
subroutine close_program(mype,message)

  implicit none

  integer           :: mype
  character (len=*) :: message 

#if PARA==1
  include 'mpif.h'
  integer :: info
#endif

  write(*,*)
  write(*,'(a   )') '##########################################################################'
  write(*,'(a,i6)') 'Call to the CLOSE_PROGRAM routine by cpu ',mype
  write(*,'(a,a )') 'Message: ',message
  write(*,'(a   )') 'Exiting...'
  write(*,'(a   )') '##########################################################################'

#if PARA==1
  call MPI_FINALIZE(info)
#else
  stop
#endif
  
end subroutine close_program
