!> \file
!! Contains subroutine evol_mhd()
!<

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine EVOL_MHD
!
!> Performs a single timestep evolution of mhd.
!<
subroutine evol_mhd(imin,imax,jmin,jmax,kmin,kmax)

  use var
  use var_loc
  use varold
  use geom
  use divers
  use parameters, only : ndim,nx,nvar,n_vit,nbuf
  use cl
  use gravity   , only : phi
  use Etat_GDS  , only : uS
  use gammas    , only : gamma

  implicit none

  integer, intent(in)   :: imin,imax,jmin,jmax,kmin,kmax
  
  integer, dimension(6) :: icl_loc
  integer, parameter    :: nmhd=2
  integer               :: nnx,nny,nnz,nn
  integer               :: iu1,iu2,ju1,ju2,ku1,ku2,if1,if2,jf1,jf2,kf1,kf2
  integer               :: ilo,ihi,jlo,jhi,klo,khi
  integer               :: i,j,k,ivar,idim
  integer               :: iimin,iimax,jjmin,jjmax,kkmin,kkmax
  
  ! Conservative variables and fluxes
  real, dimension(:,:,:,:  ), allocatable :: uin,gravin,flux_pre_tot
  real, dimension(:,:,:,:,:), allocatable :: flux

  ! Primitive variables
  real, dimension(:,:,:,:  ), allocatable :: qin 
  real, dimension(:,:,:,:  ), allocatable :: bf  

  ! Slopes
  real, dimension(:,:,:,:,:), allocatable :: dq

  ! Slopes of the staggered magnetic field 
  real, dimension(:,:,:,:,:), allocatable :: dbf

  ! Left and right state arrays (defined at face centre)
  real, dimension(:,:,:,:,:), allocatable :: qm,qp
  real, dimension(:,:,:,:,:), allocatable :: qRT,qRB,qLT,qLB
  
  ! Intermediate fluxes
  real, dimension(:,:,:,:  ), allocatable :: flx
  real, dimension(:,:,:    ), allocatable :: flx_pre_tot

  ! emf's
  real, dimension(:,:,:    ), allocatable :: emf,emfx,emfy,emfz

  ! Radiative transfer
#if RAYO==1 || DIF==1
  real, dimension(:,:,:,:), allocatable :: moment_ray
  real :: rayon
#endif

  if (verbose) write(*,*) 'Start of evol_mhd...'

  ! Tests
  call check_size(imin,imax,jmin,jmax,kmin,kmax)

  !  Sub-arrays boundary conditions
  icl_loc = 0
  if(imin.eq.1    ) icl_loc(1)=icl(1)
  if(imax.eq.nx(1)) icl_loc(2)=icl(2)
  if(jmin.eq.1    ) icl_loc(3)=icl(3)
  if(jmax.eq.nx(2)) icl_loc(4)=icl(4)
  if(kmin.eq.1    ) icl_loc(5)=icl(5)
  if(kmax.eq.nx(3)) icl_loc(6)=icl(6)

  ! Allocate arrays
  nnx = imax - imin + 1 ; nny =  jmax - jmin + 1 ; nnz = kmax - kmin + 1
  if(ndim==1) then
     iu1 = 1-nmhd ; iu2 = nnx+nmhd
     ju1 = 1      ; ju2 = 1
     ku1 = 1      ; ku2 = 1
     if1 = 1      ; if2 = nnx+1
     jf1 = 1      ; jf2 = 1
     kf1 = 1      ; kf2 = 1
  else if (ndim==2) then
     iu1 = 1-nmhd ; iu2 = nnx+nmhd
     ju1 = 1-nmhd ; ju2 = nny+nmhd
     ku1 = 1      ; ku2 = 1
     if1 = 1      ; if2 = nnx+1
     jf1 = 1      ; jf2 = nny+1
     kf1 = 1      ; kf2 = 1
  else if (ndim==3) then
     iu1 = 1-nmhd ; iu2 = nnx+nmhd
     ju1 = 1-nmhd ; ju2 = nny+nmhd
     ku1 = 1-nmhd ; ku2 = nnz+nmhd
     if1 = 1      ; if2 = nnx+1
     jf1 = 1      ; jf2 = nny+1
     kf1 = 1      ; kf2 = nnz+1
  endif

  ilo=min(1,iu1+nmhd); ihi=max(1,iu2-nmhd)
  jlo=min(1,ju1+nmhd); jhi=max(1,ju2-nmhd)
  klo=min(1,ku1+nmhd); khi=max(1,ku2-nmhd)

  nn  = max(iu2,ju2,ku2)
  allocate(xloc(1-nmhd:nn+1,ndim))

  allocate(uin         (iu1:iu2,ju1:ju2,ku1:ku2,1:nvar+3       ))
  allocate(gravin      (iu1:iu2,ju1:ju2,ku1:ku2,1:ndim         )) ; gravin       = 0.
  allocate(flux        (if1:if2,jf1:jf2,kf1:kf2,1:nvar  ,1:ndim)) ; flux         = 0.
  allocate(flux_pre_tot(if1:if2,jf1:jf2,kf1:kf2,         1:ndim)) ; flux_pre_tot = 0.
#if GRA==1
  if (GRAV) allocate(philoc(iu1:iu2,ju1:ju2,ku1:ku2)) ; philoc = 0.
#endif

  allocate(qin(iu1:iu2  ,ju1:ju2  ,ku1:ku2  ,1:nvar))
  allocate(bf (iu1:iu2+1,ju1:ju2+1,ku1:ku2+1,1:3   ))

  allocate(dq         (iu1:iu2,ju1:ju2,ku1:ku2,1:nvar,1:ndim))
  allocate(dbf        (iu1:iu2,ju1:ju2,ku1:ku2,1:3   ,1:ndim))

  allocate(qm         (iu1:iu2,ju1:ju2,ku1:ku2,1:nvar,1:ndim))
  allocate(qp         (iu1:iu2,ju1:ju2,ku1:ku2,1:nvar,1:ndim))

  allocate(qRT        (iu1:iu2,ju1:ju2,ku1:ku2,1:nvar,1:3   ))
  allocate(qRB        (iu1:iu2,ju1:ju2,ku1:ku2,1:nvar,1:3   ))
  allocate(qLT        (iu1:iu2,ju1:ju2,ku1:ku2,1:nvar,1:3   ))
  allocate(qLB        (iu1:iu2,ju1:ju2,ku1:ku2,1:nvar,1:3   ))

  allocate(flx        (iu1:iu2,ju1:ju2,ku1:ku2,1:nvar       ))
  allocate(flx_pre_tot(iu1:iu2,ju1:ju2,ku1:ku2              ))

  if (ndim>1) then
     allocate(emf (iu1:iu2,ju1:ju2,ku1:ku2))
     allocate(emfx(iu1:iu2,ju1:ju2,ku1:ku2))
     allocate(emfy(iu1:iu2,ju1:ju2,ku1:ku2))
     allocate(emfz(iu1:iu2,ju1:ju2,ku1:ku2))
  endif

  allocate(uS(iu1:iu2,ju1:ju2,ku1:ku2,1:ndim,1:ndim))
#if RAYO==1 || DIF==1
  allocate(Einloc(iu1:iu2,ju1:ju2,ku1:ku2))
#endif

  !Gather sub-arrays
  if(verbose) write(*,*) 'Gather sub-arrays...'
  if (ndim==1) then
     iimin=imin-nmhd ; iimax=imax+nmhd
     uin(iu1:iu2,ju1:ju2,ku1:ku2,1                ) =  rhoold(iimin:iimax,1:1,1:1        )
     uin(iu1:iu2,ju1:ju2,ku1:ku2,2:n_vit+1        ) = rhouold(iimin:iimax,1:1,1:1,1:n_vit)
     uin(iu1:iu2,ju1:ju2,ku1:ku2,  n_vit+2        ) =    Eold(iimin:iimax,1:1,1:1        )
     uin(iu1:iu2,ju1:ju2,ku1:ku2,n_vit+3:2*n_vit+2) =    Bold(iimin:iimax,1:1,1:1,1:n_vit)
     !right faces B-field
     uin(iu1:iu2,ju1:ju2,ku1:ku2,nvar+1) = Bold(iimin+1:iimax+1,1:1,1:1,1) 
     uin(iu1:iu2,ju1:ju2,ku1:ku2,nvar+2) = Bold(iimin  :iimax  ,1:1,1:1,2) 
     uin(iu1:iu2,ju1:ju2,ku1:ku2,nvar+3) = Bold(iimin  :iimax  ,1:1,1:1,3)
  else if (ndim==2) then
     iimin=imin-nmhd ; iimax=imax+nmhd
     jjmin=jmin-nmhd ; jjmax=jmax+nmhd
     uin(iu1:iu2,ju1:ju2,ku1:ku2,1                ) =  rhoold(iimin:iimax,jjmin:jjmax,1:1        )
     uin(iu1:iu2,ju1:ju2,ku1:ku2,2:n_vit+1        ) = rhouold(iimin:iimax,jjmin:jjmax,1:1,1:n_vit)
     uin(iu1:iu2,ju1:ju2,ku1:ku2,n_vit+2          ) =    Eold(iimin:iimax,jjmin:jjmax,1:1        )
     uin(iu1:iu2,ju1:ju2,ku1:ku2,n_vit+3:2*n_vit+2) =    Bold(iimin:iimax,jjmin:jjmax,1:1,1:n_vit)
     !right faces B-field
     uin(iu1:iu2,ju1:ju2,ku1:ku2,nvar+1) = Bold(iimin+1:iimax+1,jjmin  :jjmax  ,1:1,1)
     uin(iu1:iu2,ju1:ju2,ku1:ku2,nvar+2) = Bold(iimin  :iimax  ,jjmin+1:jjmax+1,1:1,2)
     uin(iu1:iu2,ju1:ju2,ku1:ku2,nvar+3) = Bold(iimin  :iimax  ,jjmin  :jjmax  ,1:1,3)
  else if (ndim==3) then
     iimin=imin-nmhd ; iimax=imax+nmhd
     jjmin=jmin-nmhd ; jjmax=jmax+nmhd
     kkmin=kmin-nmhd ; kkmax=kmax+nmhd
     uin(iu1:iu2,ju1:ju2,ku1:ku2,1                ) =  rhoold(iimin:iimax,jjmin:jjmax,kkmin:kkmax        )
     uin(iu1:iu2,ju1:ju2,ku1:ku2,2:n_vit+1        ) = rhouold(iimin:iimax,jjmin:jjmax,kkmin:kkmax,1:n_vit)
     uin(iu1:iu2,ju1:ju2,ku1:ku2,  n_vit+2        ) =    Eold(iimin:iimax,jjmin:jjmax,kkmin:kkmax        )
     uin(iu1:iu2,ju1:ju2,ku1:ku2,n_vit+3:2*n_vit+2) =    Bold(iimin:iimax,jjmin:jjmax,kkmin:kkmax,1:n_vit)
     !right faces B-field
     uin(iu1:iu2,ju1:ju2,ku1:ku2,nvar+1) = Bold(iimin+1:iimax+1,jjmin  :jjmax  ,kkmin  :kkmax  ,1)
     uin(iu1:iu2,ju1:ju2,ku1:ku2,nvar+2) = Bold(iimin  :iimax  ,jjmin+1:jjmax+1,kkmin  :kkmax  ,2)
     uin(iu1:iu2,ju1:ju2,ku1:ku2,nvar+3) = Bold(iimin  :iimax  ,jjmin  :jjmax  ,kkmin+1:kkmax+1,3)
  endif

  !gravitational potential
#if GRA==1
  if (GRAV) then
     if (ndim==1) then
        philoc(iu1:iu2,ju1:ju2,ku1:ku2) = phi(iimin:iimax,1:1        ,1:1        )
     else if (ndim==2) then
        philoc(iu1:iu2,ju1:ju2,ku1:ku2) = phi(iimin:iimax,jjmin:jjmax,1:1        )
     else if (ndim==3) then
        philoc(iu1:iu2,ju1:ju2,ku1:ku2) = phi(iimin:iimax,jjmin:jjmax,kkmin:kkmax)
     endif
  endif
#endif

  ! Geometry
  if(verbose) write(*,*) 'Compute geometrical term...'

              xloc(1-nmhd:nnx+nmhd+1,1) = x(imin-nmhd:imax+nmhd+1,1)   
  if (ndim>1) xloc(1-nmhd:nny+nmhd+1,2) = x(jmin-nmhd:jmax+nmhd+1,2)   
  if (ndim>2) xloc(1-nmhd:nnz+nmhd+1,3) = x(kmin-nmhd:kmax+nmhd+1,3)   
  call cal_geom(nnx,nny,nnz)

  ! Compute primitive variables
  if(verbose) write(*,*) 'Compute primitive variables...'
  call ctoprim(uin,qin,bf,gravin,dt,iu1,iu2,ju1,ju2,ku1,ku2,nvar,ndim)

  ! Compute TVD slopes
  if(verbose) write(*,*) 'Compute TVD slopes...'
  call uslope(bf,qin,dq,dbf,iu1,iu2,ju1,ju2,ku1,ku2,nvar,ndim)

  ! Compute 3D traced-states in all three directions
  if(verbose) write(*,*) 'Compute predicted states...'
  if (ndim==3) then
     call trace3d(qin,bf,dq,dbf,qm,qp,qRT,qRB,qLT,qLB,dt,iu1,iu2,ju1,ju2,ku1,ku2,nvar,ndim)
  else if (ndim==2) then
     call trace2d(qin,bf,dq,dbf,qm,qp,qRT,qRB,qLT,qLB,dt,iu1,iu2,ju1,ju2,ku1,ku2,nvar,ndim)
  else if (ndim==1) then
     call trace1d(qin,   dq,    qm,qp,                dt,iu1,iu2,ju1,ju2,ku1,ku2,nvar,ndim)
  endif

  !Gravity predictor step
#if GRA==1
  if (GRAV) then
     do idim=1,ndim
        call grav_predictor(qm (:,:,:,2:4,idim),philoc,dt,iu1,iu2,ju1,ju2,ku1,ku2)
        call grav_predictor(qp (:,:,:,2:4,idim),philoc,dt,iu1,iu2,ju1,ju2,ku1,ku2)
        call grav_predictor(qRT(:,:,:,2:4,idim),philoc,dt,iu1,iu2,ju1,ju2,ku1,ku2)
        call grav_predictor(qRB(:,:,:,2:4,idim),philoc,dt,iu1,iu2,ju1,ju2,ku1,ku2)
        call grav_predictor(qLT(:,:,:,2:4,idim),philoc,dt,iu1,iu2,ju1,ju2,ku1,ku2)
        call grav_predictor(qLB(:,:,:,2:4,idim),philoc,dt,iu1,iu2,ju1,ju2,ku1,ku2)
     end do
  endif
#endif


  ! 1D flux in X direction
  if(verbose) write(*,*) 'Compute fluxes (x-direction)...'
  call cmpflxm(qm,iu1+1,iu2+1,ju1  ,ju2  ,ku1  ,ku2  , &
       &       qp,iu1  ,iu2  ,ju1  ,ju2  ,ku1  ,ku2  , &
       &          if1  ,if2  ,jlo  ,jhi  ,klo  ,khi  , 2,3,4,6,7,8, &
       &       flx,flx_pre_tot,uS,nvar,ndim)

  ! SAVE flux in output array
  do ivar=1,nvar
     do k=kf1,kf2
        do j=jf1,jf2
           do i=if1,if2
              flux        (i,j,k,ivar,1)=flx        (i,j,k,ivar)*dt
              flux_pre_tot(i,j,k,     1)=flx_pre_tot(i,j,k     )*dt
           end do
        end do
     end do
  end do

  !2D 
  ! 1D flux in Y direction
  if (ndim>1) then
     if(verbose) write(*,*) 'Compute fluxes (y-direction)...'
     call cmpflxm(qm,iu1  ,iu2  ,ju1+1,ju2+1,ku1  ,ku2  , &
          &       qp,iu1  ,iu2  ,ju1  ,ju2  ,ku1  ,ku2  , &
          &          ilo  ,ihi  ,jf1  ,jf2  ,klo  ,khi  , 3,2,4,7,6,8, &
          &       flx,flx_pre_tot,uS,nvar,ndim)
     ! Save flux in output array
     do ivar=1,nvar
        do k=kf1,kf2
           do j=jf1,jf2
              do i=if1,if2
                 flux        (i,j,k,ivar,2)=flx        (i,j,k,ivar)*dt
                 flux_pre_tot(i,j,k,     2)=flx_pre_tot(i,j,k     )*dt
              end do
           end do
        end do
     end do
     
     call cmp_mag_flx(qRT,iu1+1,iu2+1,ju1+1,ju2+1,ku1  ,ku2  , &
          &           qRB,iu1+1,iu2+1,ju1  ,ju2  ,ku1  ,ku2  , &
          &           qLT,iu1  ,iu2  ,ju1+1,ju2+1,ku1  ,ku2  , &
          &           qLB,iu1  ,iu2  ,ju1  ,ju2  ,ku1  ,ku2  , &
          &               if1  ,if2  ,jf1  ,jf2  ,klo  ,khi  , 2,3,4,6,7,8, &
          &           emf,nvar,ndim)
     
     ! Save vector in output array
     do k=klo,khi
        do j=jf1,jf2
           do i=if1,if2
              emfz(i,j,k)=emf(i,j,k)*dt
           end do
        end do
     end do
  endif
  
  !3D
  !  1D flux in Z direction
  if (ndim>2) then
     if(verbose) write(*,*) 'Compute fluxes (z-direction)...'
     call cmpflxm(qm,iu1  ,iu2  ,ju1  ,ju2  ,ku1+1,ku2+1, &
          &       qp,iu1  ,iu2  ,ju1  ,ju2  ,ku1  ,ku2  , &
          &          ilo  ,ihi  ,jlo  ,jhi  ,kf1  ,kf2  , 4,2,3,8,6,7, &
          &       flx,flx_pre_tot,uS,nvar,ndim)
     ! Save flux in output array
     do ivar=1,nvar
        do k=kf1,kf2
           do j=jf1,jf2
              do i=if1,if2
                 flux        (i,j,k,ivar,3)=flx        (i,j,k,ivar)*dt
                 flux_pre_tot(i,j,k,     3)=flx_pre_tot(i,j,k     )*dt
              end do
           end do
        end do
     end do
     ! Calculate Ey
     call cmp_mag_flx(qRT,iu1+1,iu2+1,ju1,ju2,ku1+1,ku2+1, &
          &           qLT,iu1  ,iu2  ,ju1,ju2,ku1+1,ku2+1, &
          &           qRB,iu1+1,iu2+1,ju1,ju2,ku1  ,ku2  , &
          &           qLB,iu1  ,iu2  ,ju1,ju2,ku1  ,ku2  , &
          &               if1  ,if2  ,jlo,jhi,kf1  ,kf2  , 4,2,3,8,6,7, &
          &           emf,nvar,ndim)
     ! Save vector in output array
     do k=kf1,kf2
        do j=jlo,jhi
           do i=if1,if2
              emfy(i,j,k)=emf(i,j,k)*dt
           end do
        end do
     end do
     ! Calculate Ex
     call cmp_mag_flx(qRT,iu1,iu2,ju1+1,ju2+1,ku1+1,ku2+1, &
          &           qRB,iu1,iu2,ju1+1,ju2+1,ku1  ,ku2  , &
          &           qLT,iu1,iu2,ju1  ,ju2  ,ku1+1,ku2+1, &
          &           qLB,iu1,iu2,ju1  ,ju2  ,ku1  ,ku2  , &
          &               ilo,ihi,jf1  ,jf2  ,kf1  ,kf2  , 3,4,2,7,8,6, &
          &           emf,nvar,ndim)
     
     ! Save vector in output array
     do k=kf1,kf2
        do j=jf1,jf2
           do i=ilo,ihi
              emfx(i,j,k)=emf(i,j,k)*dt
           end do
        end do
     end do
  endif

#if RAYO==1 || DIF==1
  allocate(moment_ray(iu1:iu2,ju1:ju2,ku1:ku2,1:ndim)) ; moment_ray=0.

  if(RAY.or.DIFFU) then
     Einloc=qin(:,:,:,n_vit+2)/(gamma-1.)
     ! Computes the radiative source term (kappa rho+sigma)F/c in moment and energy hydro equations
     call exp_source_ray(uin(:,:,:,1),Einloc,moment_ray,imin,imax,jmin,jmax,kmin,kmax,iu1,iu2,ju1,ju2,ku1,ku2)

     ! Explicit treatment of comoving terms for radiation
     if(RAY  ) call exp_comobile_ray(uin(:,:,:,1),Einloc,uS,imin,imax,jmin,jmax,kmin,kmax,iu1,iu2,ju1,ju2,ku1,ku2,'ray')
     if(DIFFU) call exp_comobile_ray(uin(:,:,:,1),Einloc,uS,imin,imax,jmin,jmax,kmin,kmax,iu1,iu2,ju1,ju2,ku1,ku2,'dif')

     ! Radiative source term update in moment and energy equation
     do k=1,khi
        do j=1,jhi
           do i=1,ihi
              do idim=1,ndim
                 uin(i,j,k,n_vit+2) = uin(i,j,k,n_vit+2) + dt*moment_ray(i,j,k,idim)*uin(i,j,k,idim+1)/uin(i,j,k,1)
              end do
              do idim=1,ndim
                 rayon = 1.
                 if ((geom_dir(2) == 'phi_c' .or. geom_dir(2) == 'the_s') .and. idim==2) rayon=xcloc(i,1)
                 if ((geom_dir(3) == 'phi_s' )                            .and. idim==3) rayon=xcloc(i,1)*sin(xcloc(j,2))
                 uin(i,j,k,idim+1)  = uin(i,j,k,idim+1)  + dt*moment_ray(i,j,k,idim)/rayon
              enddo
           end do
        end do
     end do
  endif
#endif

  ! Conservative update of the variables
  if(verbose) write(*,*) 'Conservative update...'
  call update(uin,flux,flux_pre_tot,emfx,emfy,emfz,            &
       &      iu1,iu2,ju1,ju2,ku1,ku2,if1,if2,jf1,jf2,kf1,kf2, &
       &      nvar,ndim)

  ! Scatter sub-arrays to main arrays
  if(verbose) write(*,*) 'Scatter sub-arrays...'
   rho(imin:imax  ,jmin:jmax  ,kmin:kmax          ) = uin(1:nnx  ,1:nny  ,1:nnz  ,1                )
  rhou(imin:imax  ,jmin:jmax  ,kmin:kmax  ,1:n_vit) = uin(1:nnx  ,1:nny  ,1:nnz  ,2:n_vit+1        )
     E(imin:imax  ,jmin:jmax  ,kmin:kmax          ) = uin(1:nnx  ,1:nny  ,1:nnz  ,n_vit+2          )
  if (ndim==1) B(imin:imax+1,jmin:jmax  ,kmin:kmax  ,1:n_vit) = uin(1:nnx+1,1:nny  ,1:nnz  ,n_vit+3:2*n_vit+2)
  if (ndim==2) B(imin:imax+1,jmin:jmax+1,kmin:kmax  ,1:n_vit) = uin(1:nnx+1,1:nny+1,1:nnz  ,n_vit+3:2*n_vit+2)
  if (ndim==3) B(imin:imax+1,jmin:jmax+1,kmin:kmax+1,1:n_vit) = uin(1:nnx+1,1:nny+1,1:nnz+1,n_vit+3:2*n_vit+2)

#if RAYO==1 || DIF==1
   deallocate(moment_ray)
#endif

  ! Deallocate geometrical arrays
  if(verbose) write(*,*) 'Deallocate intermediate arrays...'
  deallocate(dx,dxc)
  deallocate(ds,dv)
  deallocate(xloc,xcloc)
  ! Deallocate local sub arrays
  deallocate(uin,gravin,flux,flx,flux_pre_tot,flx_pre_tot)
  deallocate(qin,bf,dq,dbf,qm,qp,qRT,qRB,qLT,qLB,uS)
  if (ndim>1) deallocate(emf,emfx,emfy,emfz)
#if GRA==1
  if (GRAV) deallocate(philoc)
#endif
#if RAYO==1 || DIF==1
  deallocate(Einloc)
#endif

  if (verbose) write(*,*) 'End of evol_mhd...'
  
  return
  
end subroutine evol_mhd
