!> \file
!! contains subroutines trace1d(), trace2d() and trace3d()
!<

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine TRACE1D
!
!> This subroutine assumed that one considers a 1D problem
!! with 3 velocities (vx,vy,vz) and 3 magnetic fields
!! (bx,by,bz).
!<
subroutine trace1d(q,dq,qm,qp,dt,iu1,iu2,ju1,ju2,ku1,ku2,nvar,ndim)

  use parameters, only : smallr,smallc
  use gammas
  use geom
  use const
  
  implicit none

  integer :: nvar,ndim
  integer :: iu1,iu2,ju1,ju2,ku1,ku2

  real, dimension(iu1:iu2,ju1:ju2,ku1:ku2,1:nvar       ) :: q  
  real, dimension(iu1:iu2,ju1:ju2,ku1:ku2,1:nvar,1:ndim) :: dq 
  real, dimension(iu1:iu2,ju1:ju2,ku1:ku2,1:nvar,1:ndim) :: qm 
  real, dimension(iu1:iu2,ju1:ju2,ku1:ku2,1:nvar,1:ndim) :: qp 
  real :: dt

  ! declare local variables
  integer :: i,j,k,n
  integer :: ilo,ihi,jlo,jhi,klo,khi
  integer :: ir,iu,iv,iw,ip,iA,iB,iC
  real    :: dtdx,xc
  real    :: r,u,v,w,p,A,B,C
  real    :: drx,dux,dvx,dwx,dpx,dAx,dBx,dCx
  real    :: sr0,su0=0.,sv0=0.,sw0=0.,sp0,sA0,sB0,sC0
  
  ilo=min(1,iu1+1); ihi=max(1,iu2-1)
  jlo=min(1,ju1+1); jhi=max(1,ju2-1)
  klo=min(1,ku1+1); khi=max(1,ku2-1)
  ir = 1; iu = 2; iv = 3 ; iw = 4 ; ip = 5; iA = 6; iB = 7; iC = 8

  do k = klo, khi
     do j = jlo, jhi
        do i = ilo, ihi

           dtdx = dt/dx(i,1)
           xc   = xcloc(i,1) 

           ! Cell centered values
           r = q(i,j,k,ir)
           u = q(i,j,k,iu)
           v = q(i,j,k,iv)
           w = q(i,j,k,iw)
           p = q(i,j,k,ip)
           A = q(i,j,k,iA)
           B = q(i,j,k,iB)
           C = q(i,j,k,iC)

           ! TVD slopes in X direction
           drx = half*dq(i,j,k,ir,1)
           dux = half*dq(i,j,k,iu,1)
           dvx = half*dq(i,j,k,iv,1)
           dwx = half*dq(i,j,k,iw,1)
           dpx = half*dq(i,j,k,ip,1)
           dBx = half*dq(i,j,k,iB,1)
           dCx = half*dq(i,j,k,iC,1)
           
           ! Source terms (including transverse derivatives)
           if (Cartesien) then
              sr0 = -u*drx-r*dux               ; sr0=sr0*dtdx
              su0 = -u*dux-(dpx+B*dBx+C*dCx)/r ; su0=su0*dtdx  
              sv0 = -u*dvx+(A*dBx)/r           ; sv0=sv0*dtdx
              sw0 = -u*dwx+(A*dCx)/r           ; sw0=sw0*dtdx
              sp0 = -u*dpx-gamma*p*dux         ; sp0=sp0*dtdx
              sB0 = -u*dBx+A*dvx-B*dux         ; sB0=sB0*dtdx
              sC0 = -u*dCx+A*dwx-C*dux         ; sC0=sC0*dtdx
           endif
           if (Cylindrique) then
              !Note that A=0 by definition in this case
              sr0 = -u*drx-r*dux               ; sr0=sr0*dtdx-r*u/xc        *half*dt
              su0 = -u*dux-(dpx+B*dBx+C*dCx)/r ; su0=su0*dtdx+(v*v-B*B/r)/xc*half*dt
              sv0 = -u*dvx                     ; sv0=sv0*dtdx-u*v/xc        *half*dt
              sw0 = -u*dwx                     ; sw0=sw0*dtdx
              sp0 = -u*dpx-gamma*p*dux         ; sp0=sp0*dtdx-gamma*p*u/xc  *half*dt
              sB0 = -u*dBx-B*dux               ; sB0=sB0*dtdx
              sC0 = -u*dCx-C*dux               ; sC0=sC0*dtdx-u*C/xc        *half*dt
           endif
           if (Spherique) then
              !Note that theta=pi/2 is assumed
              sr0 = -u*drx-r*dux               ; sr0=sr0*dtdx-two*r*u              /xc*half*dt
              su0 = -u*dux-(dpx+B*dBx+C*dCx)/r ; su0=su0*dtdx+(v*v+w*w-B*B/r-C*C/r)/xc*half*dt
              sv0 = -u*dvx+(A*dBx)/r           ; sv0=sv0*dtdx-(u*v-A*B/r)          /xc*half*dt
              sw0 = -u*dwx+(A*dCx)/r           ; sw0=sw0*dtdx-(u*w-A*C/r)          /xc*half*dt
              sp0 = -u*dpx-gamma*p*dux         ; sp0=sp0*dtdx-two*gamma*p*u        /xc*half*dt
              sB0 = -u*dBx+A*dvx-B*dux         ; sB0=sB0*dtdx-(u*B-v*A)            /xc*half*dt
              sC0 = -u*dCx+A*dwx-C*dux         ; sC0=sC0*dtdx+(w*A-u*C)            /xc*half*dt
           endif
           
           ! Cell-centered predicted states
           r = r + sr0
           u = u + su0
           v = v + sv0
           w = w + sw0
           p = p + sp0
           B = B + sB0
           C = C + sC0
           
           ! Right state at left interface
           qp(i,j,k,ir,1) = r - drx
           qp(i,j,k,iu,1) = u - dux
           qp(i,j,k,iv,1) = v - dvx
           qp(i,j,k,iw,1) = w - dwx
           qp(i,j,k,ip,1) = p - dpx
           qp(i,j,k,iA,1) = A
           qp(i,j,k,iB,1) = B - dBx
           qp(i,j,k,iC,1) = C - dCx
           qp(i,j,k,ir,1) = max(smallr, qp(i,j,k,ir,1))
           
           ! Left state at right interface
           qm(i,j,k,ir,1) = r + drx
           qm(i,j,k,iu,1) = u + dux
           qm(i,j,k,iv,1) = v + dvx
           qm(i,j,k,iw,1) = w + dwx
           qm(i,j,k,ip,1) = p + dpx
           qm(i,j,k,iA,1) = A
           qm(i,j,k,iB,1) = B + dBx
           qm(i,j,k,iC,1) = C + dCx
           qm(i,j,k,ir,1) = max(smallr, qm(i,j,k,ir,1))
           
        end do
     end do
  end do
  
  ! passive scalars
  do n = 9, nvar
     do k = klo, khi
        do j = jlo, jhi
           do i = ilo, ihi
              a   = q(i,j,k,n )           ! Cell centered values
              u   = q(i,j,k,iu)  
              dax = half * dq(i,j,k,n,1)  ! TVD slope
              sa0 = -u*dax                  ! Source terms
              a   = a + sa0*dtdx            ! Predicted state
              qp(i,j,k,n,1) = a - dax     ! Right state
              qm(i,j,k,n,1) = a + dax     ! Left state
           end do
        end do
     end do
  end do

end subroutine trace1d

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine TRACE2D
!
!> Trace in 2 dimensions.
!<
subroutine trace2d(q,bf,dq,dbf,qm,qp,qRT,qRB,qLT,qLB,dt,iu1,iu2,ju1,ju2,ku1,ku2,nvar,ndim)

  use parameters, only : smallr,smallc,n_vit
  use gammas
  use geom
  use const
  use var_loc   , only: xloc

  implicit none

  integer :: nvar,ndim
  integer :: iu1,iu2,ju1,ju2,ku1,ku2
  real    :: dt

  real, dimension(iu1:iu2  ,ju1:ju2  ,ku1:ku2  ,1:nvar       ) :: q  
  real, dimension(iu1:iu2+1,ju1:ju2+1,ku1:ku2+1,1:3          ) :: bf
  real, dimension(iu1:iu2  ,ju1:ju2  ,ku1:ku2  ,1:nvar,1:ndim) :: dq 
  real, dimension(iu1:iu2  ,ju1:ju2  ,ku1:ku2  ,1:nvar,1:ndim) :: qm 
  real, dimension(iu1:iu2  ,ju1:ju2  ,ku1:ku2  ,1:nvar,1:ndim) :: qp 
  real, dimension(iu1:iu2  ,ju1:ju2  ,ku1:ku2  ,1:3   ,1:ndim) :: dbf
  real, dimension(iu1:iu2  ,ju1:ju2  ,ku1:ku2  ,1:nvar,1:3   ) :: qRT
  real, dimension(iu1:iu2  ,ju1:ju2  ,ku1:ku2  ,1:nvar,1:3   ) :: qRB
  real, dimension(iu1:iu2  ,ju1:ju2  ,ku1:ku2  ,1:nvar,1:3   ) :: qLT
  real, dimension(iu1:iu2  ,ju1:ju2  ,ku1:ku2  ,1:nvar,1:3   ) :: qLB

  real, dimension(:,:,:), allocatable :: Ez

  ! Local variables
  integer :: i,j,k,n
  integer :: ilo,ihi,jlo,jhi,klo,khi
  integer :: ir,iu,iv,iw,ip,iA,iB,iC 
  real    :: dtdx,dtdy,xc,xL,xR,sinxL,sinxR,smallp
  real    :: r,u,v,w,p,A,B,C
  real    :: ELL,ELR,ERL,ERR
  real    :: drx,dux,dvx,dwx,dpx,dAx,dBx,dCx
  real    :: dry,duy,dvy,dwy,dpy,dAy,dBy,dCy
  real    :: sr0,su0,sv0,sw0,sp0,sA0,sB0,sC0
  real    :: AL,AR,BL,BR
  real    :: dALy,dARy
  real    :: dBLx,dBRx
  real    :: sAL0,sAR0,sBL0,sBR0
  real    :: sinxc,cotanxc
  
  allocate(Ez(iu1:iu2,ju1:ju2,ku1:ku2))

  smallp = smallr*smallc**2/gamma

  ilo=min(1,iu1+1); ihi=max(1,iu2-1)
  jlo=min(1,ju1+1); jhi=max(1,ju2-1)
  klo=min(1,ku1+1); khi=max(1,ku2-1)

  ir=1; iu=2; iv=3; ip=2+n_vit  ; iA=3+n_vit ; iB=4+n_vit 
  iw=4 ; iC=8

  do k = klo, ku2
     do j = jlo, ju2
        do i = ilo, iu2
           u = 0.25*(q(i-1,j-1,k,iu)+q(i-1,j,k,iu)+q(i,j-1,k,iu)+q(i,j,k,iu))
           v = 0.25*(q(i-1,j-1,k,iv)+q(i-1,j,k,iv)+q(i,j-1,k,iv)+q(i,j,k,iv))
           A = 0.5*(bf(i,j-1,k,1)+bf(i,j,k,1))
           B = 0.5*(bf(i-1,j,k,2)+bf(i,j,k,2))
           Ez(i,j,k)=u*B-v*A
        end do
     end do
  end do

  do k = klo, khi
     do j = jlo, jhi
        do i = ilo, ihi

           dtdx = dt/dx(i,1)
           dtdy = dt/dx(j,2)
           xc   = xcloc(i,1)
           xL   = xloc(i  ,1)
           xR   = xloc(i+1,1)
           if (Spherique) then
              sinxc   = sin(xcloc(j  ,2))
              sinxL   = sin( xloc(j  ,2))
              sinxR   = sin( xloc(j+1,2))
              cotanxc = cos(xcloc(j  ,2))/sinxc
           endif

           ! Cell centered values
           r = q(i,j,k,ir)
           u = q(i,j,k,iu)
           v = q(i,j,k,iv)
           p = q(i,j,k,ip)
           A = q(i,j,k,iA)
           B = q(i,j,k,iB)
           w = q(i,j,k,iw)            
           C = q(i,j,k,iC)            

           ! Face centered variables
           AL =  bf(i  ,j  ,k,1)
           AR =  bf(i+1,j  ,k,1)
           BL =  bf(i  ,j  ,k,2)
           BR =  bf(i  ,j+1,k,2)
           
           !electric field
           ELL = Ez(i  ,j  ,k)
           ELR = Ez(i  ,j+1,k)
           ERL = Ez(i+1,j  ,k)
           ERR = Ez(i+1,j+1,k)

           ! Cell centered TVD slopes in X direction
           drx = half * dq(i,j,k,ir,1)
           dux = half * dq(i,j,k,iu,1)
           dvx = half * dq(i,j,k,iv,1)
           dwx = half * dq(i,j,k,iw,1)
           dCx = half * dq(i,j,k,iC,1)
           dpx = half * dq(i,j,k,ip,1)
           dBx = half * dq(i,j,k,iB,1)
           
           ! Cell centered TVD slopes in Y direction
           dry = half * dq(i,j,k,ir,2)
           duy = half * dq(i,j,k,iu,2)
           dvy = half * dq(i,j,k,iv,2)
           dwy = half * dq(i,j,k,iw,2)
           dCy = half * dq(i,j,k,iC,2)
           dpy = half * dq(i,j,k,ip,2)
           dAy = half * dq(i,j,k,iA,2)
           
           ! Face centered TVD slopes in transverse direction
           dALy = half * dbf(i  ,j  ,k,1,2)
           dARy = half * dbf(i+1,j  ,k,1,2)
           dBLx = half * dbf(i  ,j  ,k,2,1)
           dBRx = half * dbf(i  ,j+1,k,2,1)
           
           ! Cell centered slopes in normal direction
           dAx = half*(AR-AL)
           dBy = half*(BR-BL)
           
           ! Source terms (including transverse derivatives)
           if (Cartesien) then
              sr0 = (-u*drx-dux*r)                *dtdx + (-v*dry-dvy*r)                *dtdy
              su0 = (-u*dux-dpx/r-B*dBx/r-C*dCx/r)*dtdx + (-v*duy+B*dAy/r)              *dtdy
              sv0 = (-u*dvx+A*dBx/r)              *dtdx + (-v*dvy-dpy/r-A*dAy/r-C*dCy/r)*dtdy
              sw0 = (-u*dwx+A*dCx/r)              *dtdx + (-v*dwy+B*dCy/r)              *dtdy
              sp0 = (-u*dpx-dux*gamma*p)          *dtdx + (-v*dpy-dvy*gamma*p)          *dtdy
              sA0 =                                       ( u*dBy+B*duy-v*dAy-A*dvy)    *dtdy
              sB0 = (-u*dBx-B*dux+v*dAx+A*dvx)    *dtdx 
              sC0 = ( w*dAx+A*dwx-u*dCx-C*dux)    *dtdx + (-v*dCy-C*dvy+w*dBy+B*dwy)    *dtdy
              !Face centered B-field
              sAL0 = +(ELR-ELL)*half*dtdy
              sAR0 = +(ERR-ERL)*half*dtdy
              sBL0 = -(ERL-ELL)*half*dtdx
              sBR0 = -(ERR-ELR)*half*dtdx
           endif
           if (Cylindrique) then
              sr0 = (-u*drx-dux*r)                *dtdx + (-v*dry-dvy*r)                /xc*dtdy
              su0 = (-u*dux-dpx/r-B*dBx/r-C*dCx/r)*dtdx + (-v*duy+B*dAy/r)              /xc*dtdy
              sv0 = (-u*dvx+A*dBx/r)              *dtdx + (-v*dvy-dpy/r-A*dAy/r-C*dCy/r)/xc*dtdy
              sw0 = (-u*dwx+A*dCx/r)              *dtdx + (-v*dwy+B*dCy/r)              /xc*dtdy
              sp0 = (-u*dpx-dux*gamma*p)          *dtdx + (-v*dpy-dvy*gamma*p)          /xc*dtdy
              sA0 =                                       ( u*dBy+B*duy-v*dAy-A*dvy)    /xc*dtdy
              sB0 = (-u*dBx-B*dux+v*dAx+A*dvx)    *dtdx 
              sC0 = ( w*dAx+A*dwx-u*dCx-C*dux)    *dtdx + (-v*dCy-C*dvy+w*dBy+B*dwy)    /xc*dtdy
              !Geometrical terms
              sr0 = sr0 - r*u         /xc*half*dt
              su0 = su0 + (v*v-B*B/r) /xc*half*dt
              sv0 = sv0 + (-u*v+A*B/r)/xc*half*dt
              sp0 = sp0 - gamma*p*u   /xc*half*dt
              sC0 = sC0 + (w*A-u*C)   /xc*half*dt
              !Face centered B-field
              sAL0 = +(ELR-ELL)/xL*half*dtdy
              sAR0 = +(ERR-ERL)/xR*half*dtdy
              sBL0 = -(ERL-ELL)   *half*dtdx
              sBR0 = -(ERR-ELR)   *half*dtdx
           endif
           if (Spherique) then
              sr0 = (-u*drx-dux*r)                *dtdx + (-v*dry-dvy*r)                /xc*dtdy
              su0 = (-u*dux-dpx/r-B*dBx/r-C*dCx/r)*dtdx + (-v*duy+B*dAy/r)              /xc*dtdy
              sv0 = (-u*dvx+A*dBx/r)              *dtdx + (-v*dvy-dpy/r-A*dAy/r-C*dCy/r)/xc*dtdy
              sw0 = (-u*dwx+A*dCx/r)              *dtdx + (-v*dwy+B*dCy/r)              /xc*dtdy
              sp0 = (-u*dpx-dux*gamma*p)          *dtdx + (-v*dpy-dvy*gamma*p)          /xc*dtdy
              sA0 =                                       ( u*dBy+B*duy-v*dAy-A*dvy)    /xc*dtdy
              sB0 = (-u*dBx-B*dux+v*dAx+A*dvx)    *dtdx 
              sC0 = ( w*dAx+A*dwx-u*dCx-C*dux)    *dtdx + (-v*dCy-C*dvy+w*dBy+B*dwy)    /xc*dtdy
              !Geometrical terms
              sr0 = sr0 - (two*u+cotanxc*v)*r             /xc*half*dt
              su0 = su0 + (v*v+w*w-(B*B+C*C)/r)           /xc*half*dt
              sv0 = sv0 + (-u*v+A*B/r+cotanxc*(w*w-C*C/r))/xc*half*dt
              sw0 = sw0 + (-u*w+A*C/r+cotanxc*(v*w-B*C/r))/xc*half*dt
              sp0 = sp0 - gamma*p*(two*u+cotanxc*v)       /xc*half*dt
              sA0 = sA0 + cotanxc*(u*B-v*A)               /xc*half*dt
              sB0 = sB0 -         (u*B-v*A)               /xc*half*dt
              sC0 = sC0 +         (w*A-u*C)               /xc*half*dt
              !Face centered B-field
              sAL0 = +(ELR*sinxR-ELL*sinxL)/xL/sinxc*half*dtdy
              sAR0 = +(ERR*sinxR-ERL*sinxL)/xR/sinxc*half*dtdy
              sBL0 = -(ERL*xR   -ELL*xL   )/xc      *half*dtdx
              sBR0 = -(ERR*xR   -ELR*xL   )/xc      *half*dtdx
           endif
           
           !Update in time the  primitive variables
           r = r + sr0
           u = u + su0
           v = v + sv0
           w = w + sw0
           p = p + sp0
           A = A + sA0
           B = B + sB0
           C = C + sC0
           
           AL = AL + sAL0
           AR = AR + sAR0
           BL = BL + sBL0
           BR = BR + sBR0
           
           ! Right state at left interface
           qp(i,j,k,ir,1) = r - drx  
           qp(i,j,k,iu,1) = u - dux
           qp(i,j,k,iv,1) = v - dvx
           qp(i,j,k,iw,1) = w - dwx
           qp(i,j,k,ip,1) = p - dpx
           qp(i,j,k,iA,1) = AL
           qp(i,j,k,iB,1) = B - dBx
           qp(i,j,k,iC,1) = C - dCx
           qp(i,j,k,ir,1) = max(smallr, qp(i,j,k,ir,1))
           qp(i,j,k,ip,1) = max(smallp, qp(i,j,k,ip,1))
           
           ! Left state at right interface
           qm(i,j,k,ir,1) = r + drx
           qm(i,j,k,iu,1) = u + dux
           qm(i,j,k,iv,1) = v + dvx
           qm(i,j,k,iw,1) = w + dwx
           qm(i,j,k,ip,1) = p + dpx  
           qm(i,j,k,iA,1) = AR
           qm(i,j,k,iB,1) = B + dBx
           qm(i,j,k,iC,1) = C + dCx
           qm(i,j,k,ir,1) = max(smallr, qm(i,j,k,ir,1))
           qm(i,j,k,ip,1) = max(smallp, qm(i,j,k,ip,1))
           
           ! Top state at bottom interface
           qp(i,j,k,ir,2) = r - dry  
           qp(i,j,k,iu,2) = u - duy 
           qp(i,j,k,iv,2) = v - dvy 
           qp(i,j,k,iw,2) = w - dwy 
           qp(i,j,k,ip,2) = p - dpy  
           qp(i,j,k,iA,2) = A - dAy
           qp(i,j,k,iB,2) = BL
           qp(i,j,k,iC,2) = C - dCy
           qp(i,j,k,ir ,2) = max(smallr, qp(i,j,k,ir,2))
           qp(i,j,k,ip ,2) = max(smallp, qp(i,j,k,ip,2))
           
           ! Bottom state at top interface
           qm(i,j,k,ir,2) = r + dry
           qm(i,j,k,iu,2) = u + duy
           qm(i,j,k,iv,2) = v + dvy
           qm(i,j,k,iw,2) = w + dwy
           qm(i,j,k,ip,2) = p + dpy
           qm(i,j,k,iA,2) = A + dAy
           qm(i,j,k,iB,2) = BR
           qm(i,j,k,iC,2) = C + dCy
           qm(i,j,k,ir ,2) = max(smallr, qm(i,j,k,ir,2))
           qm(i,j,k,ip ,2) = max(smallp, qm(i,j,k,ip,2))
           
           
           ! Right-top state (RT->LL)
           qRT(i,j,k,ir,3) = r + (+drx+dry)
           qRT(i,j,k,iu,3) = u + (+dux+duy)
           qRT(i,j,k,iv,3) = v + (+dvx+dvy)
           qRT(i,j,k,iw,3) = w + (+dwx+dwy)
           qRT(i,j,k,ip,3) = p + (+dpx+dpy)
           qRT(i,j,k,iA,3) = AR+ (   +dARy)
           qRT(i,j,k,iB,3) = BR+ (+dBRx   )
           qRT(i,j,k,iC,3) = C + (+dCx+dCy)
           qRT(i,j,k,ir ,3) = max(smallr, qRT(i,j,k,ir,3))
           qRT(i,j,k,ip ,3) = max(smallp, qRT(i,j,k,ip,3))
           
           ! Right-Bottom state (RB->LR)
           qRB(i,j,k,ir,3) = r + (+drx-dry)
           qRB(i,j,k,iu,3) = u + (+dux-duy)
           qRB(i,j,k,iv,3) = v + (+dvx-dvy)
           qRB(i,j,k,iw,3) = w + (+dwx-dwy)
           qRB(i,j,k,ip,3) = p + (+dpx-dpy)
           qRB(i,j,k,iA,3) = AR+ (   -dARy)
           qRB(i,j,k,iB,3) = BL+ (+dBLx   )
           qRB(i,j,k,iC,3) = C + (+dCx-dCy)
           qRB(i,j,k,ir,3) = max(smallr, qRB(i,j,k,ir,3))
           qRB(i,j,k,ip,3) = max(smallp, qRB(i,j,k,ip,3))
           
           ! Left-Bottom state (LB->RR)
           qLB(i,j,k,ir,3) = r + (-drx-dry)
           qLB(i,j,k,iu,3) = u + (-dux-duy)
           qLB(i,j,k,iv,3) = v + (-dvx-dvy)
           qLB(i,j,k,iw,3) = w + (-dwx-dwy)
           qLB(i,j,k,ip,3) = p + (-dpx-dpy)
           qLB(i,j,k,iA,3) = AL+ (   -dALy)
           qLB(i,j,k,iB,3) = BL+ (-dBLx   )
           qLB(i,j,k,iC,3) = C + (-dCx-dCy)
           qLB(i,j,k,ir,3) = max(smallr, qLB(i,j,k,ir,3))
           qLB(i,j,k,ip,3) = max(smallp, qLB(i,j,k,ip,3))
           
           !Left-Top state (LT->RL)
           qLT(i,j,k,ir,3) = r + (-drx+dry)
           qLT(i,j,k,iu,3) = u + (-dux+duy)
           qLT(i,j,k,iv,3) = v + (-dvx+dvy)
           qLT(i,j,k,iw,3) = w + (-dwx+dwy)
           qLT(i,j,k,ip,3) = p + (-dpx+dpy)
           qLT(i,j,k,iA,3) = AL+ (   +dALy)
           qLT(i,j,k,iB,3) = BR+ (-dBRx   )
           qLT(i,j,k,iC,3) = C + (-dCx+dCy)
           qLT(i,j,k,ir,3) = max(smallr, qLT(i,j,k,ir,3))
           qLT(i,j,k,ip,3) = max(smallp, qLT(i,j,k,ip,3))
           
        end do
     end do
  end do

  ! passive scalars
  do n = 2*n_vit+3, nvar
     do k = klo, khi
        do j = jlo, jhi
           do i = ilo, ihi
              r   = q(i,j,k,n )               ! Cell centered values
              u   = q(i,j,k,iu)
              v   = q(i,j,k,iv)
              drx = half * dq(i,j,k,n,1)      ! TVD slopes
              dry = half * dq(i,j,k,n,2)
              sr0 = -u*drx*dtdx -v*dry*dtdy   ! Source terms
              r = r + sr0
              qp(i,j,k,n,1) = r - drx         ! Right state
              qm(i,j,k,n,1) = r + drx         ! Left state
              qp(i,j,k,n,2) = r - dry         ! Top state
              qm(i,j,k,n,2) = r + dry         ! Bottom state
           end do
        end do
     end do
  end do

  deallocate(Ez)
  
  return

end subroutine trace2d

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine TRACE3D
!
!> Trace in 3 dimensions.
!<
subroutine  trace3d(q,bf,dq,dbf,qm,qp,qRT,qRB,qLT,qLB,dt,iu1,iu2,ju1,ju2,ku1,ku2,nvar,ndim)

  use parameters, only : smallr,smallc
  use gammas
  use const
  use geom
  use var_loc   , only : xloc

  implicit none

  integer :: nvar,ndim
  integer :: iu1,iu2,ju1,ju2,ku1,ku2
  real    :: dt

  real, dimension(iu1:iu2  ,ju1:ju2  ,ku1:ku2  ,1:nvar       ) :: q  
  real, dimension(iu1:iu2+1,ju1:ju2+1,ku1:ku2+1,1:3          ) :: bf
  real, dimension(iu1:iu2  ,ju1:ju2  ,ku1:ku2  ,1:nvar,1:ndim) :: dq 
  real, dimension(iu1:iu2  ,ju1:ju2  ,ku1:ku2  ,1:nvar,1:ndim) :: qm 
  real, dimension(iu1:iu2  ,ju1:ju2  ,ku1:ku2  ,1:nvar,1:ndim) :: qp 
  real, dimension(iu1:iu2  ,ju1:ju2  ,ku1:ku2  ,1:3   ,1:ndim) :: dbf
  real, dimension(iu1:iu2  ,ju1:ju2  ,ku1:ku2  ,1:nvar,1:3   ) :: qRT
  real, dimension(iu1:iu2  ,ju1:ju2  ,ku1:ku2  ,1:nvar,1:3   ) :: qRB
  real, dimension(iu1:iu2  ,ju1:ju2  ,ku1:ku2  ,1:nvar,1:3   ) :: qLT
  real, dimension(iu1:iu2  ,ju1:ju2  ,ku1:ku2  ,1:nvar,1:3   ) :: qLB

  ! Declare local variables
  real, dimension(:,:,:), allocatable :: Ex,Ey,Ez

  integer :: i,j,k,n
  integer :: ilo,ihi,jlo,jhi,klo,khi
  integer :: ir,iu,iv,iw,ip,iA,iB,iC 
  real    :: dtdx,dtdy,dtdz,xc,xL,xR,sinxc,sinxL,sinxR,cotanxc,smallp
  real    :: r,u,v,w,p,A,B,C
  real    :: ELL,ELR,ERL,ERR
  real    :: FLL,FLR,FRL,FRR
  real    :: GLL,GLR,GRL,GRR
  real    :: drx,dux,dvx,dwx,dpx,dAx,dBx,dCx
  real    :: dry,duy,dvy,dwy,dpy,dAy,dBy,dCy
  real    :: drz,duz,dvz,dwz,dpz,dAz,dBz,dCz
  real    :: sr0,su0,sv0,sw0,sp0,sA0,sB0,sC0
  real    :: AL,AR,BL,BR,CL,CR
  real    :: dALy,dARy,dALz,dARz
  real    :: dBLx,dBRx,dBLz,dBRz
  real    :: dCLx,dCRx,dCLy,dCRy
  real    :: sAL0,sAR0,sBL0,sBR0,sCL0,sCR0

  allocate(Ex(iu1:iu2,ju1:ju2,ku1:ku2))
  allocate(Ey(iu1:iu2,ju1:ju2,ku1:ku2))
  allocate(Ez(iu1:iu2,ju1:ju2,ku1:ku2))

  smallp = smallr*smallc**2/gamma

  ilo=min(1,iu1+1); ihi=max(1,iu2-1)
  jlo=min(1,ju1+1); jhi=max(1,ju2-1)
  klo=min(1,ku1+1); khi=max(1,ku2-1)
  ir=1; iu=2; iv=3; iw=4; ip=5; ia=6; ib=7; ic=8 

  do k = klo, ku2
     do j = jlo, ju2
        do i = ilo, iu2
           v = 0.25*(q(i,j-1,k-1,iv)+q(i,j-1,k,iv)+q(i,j,k-1,iv)+q(i,j,k,iv))
           w = 0.25*(q(i,j-1,k-1,iw)+q(i,j-1,k,iw)+q(i,j,k-1,iw)+q(i,j,k,iw))
           B = 0.5*(bf(i,j,k-1,2)+bf(i,j,k,2))
           C = 0.5*(bf(i,j-1,k,3)+bf(i,j,k,3))
           Ex(i,j,k)=v*C-w*B
           
           u = 0.25*(q(i-1,j,k-1,iu)+q(i-1,j,k,iu)+q(i,j,k-1,iu)+q(i,j,k,iu))
           w = 0.25*(q(i-1,j,k-1,iw)+q(i-1,j,k,iw)+q(i,j,k-1,iw)+q(i,j,k,iw))
           A = 0.5*(bf(i,j,k-1,1)+bf(i,j,k,1))
           C = 0.5*(bf(i-1,j,k,3)+bf(i,j,k,3))
           Ey(i,j,k)=w*A-u*C
           
           u = 0.25*(q(i-1,j-1,k,iu)+q(i-1,j,k,iu)+q(i,j-1,k,iu)+q(i,j,k,iu))
           v = 0.25*(q(i-1,j-1,k,iv)+q(i-1,j,k,iv)+q(i,j-1,k,iv)+q(i,j,k,iv))
           A = 0.5*(bf(i,j-1,k,1)+bf(i,j,k,1))
           B = 0.5*(bf(i-1,j,k,2)+bf(i,j,k,2))
           Ez(i,j,k)=u*B-v*A
        end do
     end do
  end do

  do k = klo, khi
     do j = jlo, jhi
        do i = ilo, ihi

           dtdx = dt/dx(i,1)
           dtdy = dt/dx(j,2)
           dtdz = dt/dx(k,3)
           xc   = xcloc(i  ,1)
           xL   =  xloc(i  ,1)
           xR   =  xloc(i+1,1)
           if (Spherique) then
              sinxc   = sin(xcloc(j  ,2))
              sinxL   = sin( xloc(j  ,2))
              sinxR   = sin( xloc(j+1,2))
              cotanxc = cos(xcloc(j  ,2))/sinxc
           endif

           ! Cell centered values
           r =    q(i,j,k,ir)
           u =    q(i,j,k,iu)
           v =    q(i,j,k,iv)
           w =    q(i,j,k,iw)            
           p =    q(i,j,k,ip)
           A =    q(i,j,k,iA)
           B =    q(i,j,k,iB)
           C =    q(i,j,k,iC)            
           
           ! Cell centered TVD slopes in X direction
           drx = half * dq(i,j,k,ir,1)
           dux = half * dq(i,j,k,iu,1)
           dvx = half * dq(i,j,k,iv,1)
           dwx = half * dq(i,j,k,iw,1)
           dpx = half * dq(i,j,k,ip,1)
           dBx = half * dq(i,j,k,iB,1)
           dCx = half * dq(i,j,k,iC,1)
           
           ! Cell centered TVD slopes in Y direction
           dry = half * dq(i,j,k,ir,2)
           duy = half * dq(i,j,k,iu,2)
           dvy = half * dq(i,j,k,iv,2)
           dwy = half * dq(i,j,k,iw,2)
           dpy = half * dq(i,j,k,ip,2)
           dAy = half * dq(i,j,k,iA,2)
           dCy = half * dq(i,j,k,iC,2)
           
           ! Cell centered TVD slopes in Y direction
           drz = half * dq(i,j,k,ir,3)
           duz = half * dq(i,j,k,iu,3)
           dvz = half * dq(i,j,k,iv,3)
           dwz = half * dq(i,j,k,iw,3)
           dpz = half * dq(i,j,k,ip,3)
           dAz = half * dq(i,j,k,iA,3)
           dBz = half * dq(i,j,k,iB,3)
           
           ! Face centered variables
           AL =  bf(i  ,j  ,k  ,1)
           AR =  bf(i+1,j  ,k  ,1)
           BL =  bf(i  ,j  ,k  ,2)
           BR =  bf(i  ,j+1,k  ,2)
           CL =  bf(i  ,j  ,k  ,3)
           CR =  bf(i  ,j  ,k+1,3)
           
           ! Cell centered slopes in normal direction
           dAx = half*(AR-AL)
           dBy = half*(BR-BL)
           dCz = half*(CR-CL)
           
           ! Face centered TVD slopes in transverse direction
           dALy = half * dbf(i  ,j  ,k  ,1,2)
           dARy = half * dbf(i+1,j  ,k  ,1,2)
           dALz = half * dbf(i  ,j  ,k  ,1,3)
           dARz = half * dbf(i+1,j  ,k  ,1,3)
           
           dBLx = half * dbf(i  ,j  ,k  ,2,1)
           dBRx = half * dbf(i  ,j+1,k  ,2,1)
           dBLz = half * dbf(i  ,j  ,k  ,2,3)
           dBRz = half * dbf(i  ,j+1,k  ,2,3)
           
           dCLx = half * dbf(i  ,j  ,k  ,3,1)
           dCRx = half * dbf(i  ,j  ,k+1,3,1)
           dCLy = half * dbf(i  ,j  ,k  ,3,2)
           dCRy = half * dbf(i  ,j  ,k+1,3,2)
           
           ! Edge centered electric field in X, Y and Z directions
           ELL = Ex(i,j  ,k  )
           ELR = Ex(i,j  ,k+1)
           ERL = Ex(i,j+1,k  )
           ERR = Ex(i,j+1,k+1)
           
           FLL = Ey(i  ,j,k  )
           FLR = Ey(i  ,j,k+1)
           FRL = Ey(i+1,j,k  )
           FRR = Ey(i+1,j,k+1)
           
           GLL = Ez(i  ,j  ,k)
           GLR = Ez(i  ,j+1,k)
           GRL = Ez(i+1,j  ,k)
           GRR = Ez(i+1,j+1,k)

           ! Source terms (including transverse derivatives)
           if (Cartesien) then
              sr0 = (-u*drx-dux*r)              *dtdx + (-v*dry-dvy*r)              *dtdy + (-w*drz-dwz*r)              *dtdz 
              su0 = (-u*dux-(dpx+B*dBx+C*dCx)/r)*dtdx + (-v*duy+B*dAy/r)            *dtdy + (-w*duz+C*dAz/r)            *dtdz 
              sv0 = (-u*dvx+A*dBx/r)            *dtdx + (-v*dvy-(dpy+A*dAy+C*dCy)/r)*dtdy + (-w*dvz+C*dBz/r)            *dtdz
              sw0 = (-u*dwx+A*dCx/r)            *dtdx + (-v*dwy+B*dCy/r)            *dtdy + (-w*dwz-(dpz+A*dAz+B*dBz)/r)*dtdz 
              sp0 = (-u*dpx-dux*gamma*p)        *dtdx + (-v*dpy-dvy*gamma*p)        *dtdy + (-w*dpz-dwz*gamma*p)        *dtdz
              sA0 =                                     (u*dBy+B*duy-v*dAy-A*dvy)   *dtdy + (u*dCz+C*duz-w*dAz-A*dwz)   *dtdz
              sB0 = (v*dAx+A*dvx-u*dBx-B*dux)   *dtdx +                                     (v*dCz+C*dvz-w*dBz-B*dwz)   *dtdz 
              sC0 = (w*dAx+A*dwx-u*dCx-C*dux)   *dtdx + (w*dBy+B*dwy-v*dCy-C*dvy)   *dtdy
              !Face-centered B-field
              sAL0 = +(GLR-GLL)*dtdy*half -(FLR-FLL)*dtdz*half
              sAR0 = +(GRR-GRL)*dtdy*half -(FRR-FRL)*dtdz*half
              sBL0 = -(GRL-GLL)*dtdx*half +(ELR-ELL)*dtdz*half
              sBR0 = -(GRR-GLR)*dtdx*half +(ERR-ERL)*dtdz*half
              sCL0 = +(FRL-FLL)*dtdx*half -(ERL-ELL)*dtdy*half
              sCR0 = +(FRR-FLR)*dtdx*half -(ERR-ELR)*dtdy*half
           endif
           if (Cylindrique) then
              sr0 = (-u*drx-dux*r)              *dtdx + (-v*dry-dvy*r)              /xc*dtdy + (-w*drz-dwz*r)              *dtdz 
              su0 = (-u*dux-(dpx+B*dBx+C*dCx)/r)*dtdx + (-v*duy+B*dAy/r)            /xc*dtdy + (-w*duz+C*dAz/r)            *dtdz 
              sv0 = (-u*dvx+A*dBx/r)            *dtdx + (-v*dvy-(dpy+A*dAy+C*dCy)/r)/xc*dtdy + (-w*dvz+C*dBz/r)            *dtdz
              sw0 = (-u*dwx+A*dCx/r)            *dtdx + (-v*dwy+B*dCy/r)            /xc*dtdy + (-w*dwz-(dpz+A*dAz+B*dBz)/r)*dtdz 
              sp0 = (-u*dpx-dux*gamma*p)        *dtdx + (-v*dpy-dvy*gamma*p)        /xc*dtdy + (-w*dpz-dwz*gamma*p)        *dtdz
              sA0 =                                     (u*dBy+B*duy-v*dAy-A*dvy)   /xc*dtdy + (u*dCz+C*duz-w*dAz-A*dwz)   *dtdz
              sB0 = (v*dAx+A*dvx-u*dBx-B*dux)   *dtdx +                                        (v*dCz+C*dvz-w*dBz-B*dwz)   *dtdz 
              sC0 = (w*dAx+A*dwx-u*dCx-C*dux)   *dtdx + (w*dBy+B*dwy-v*dCy-C*dvy)   /xc*dtdy
              !Geometrical terms
              sr0 = sr0 - r*u         /xc*half*dt
              su0 = su0 + (v*v-B*B/r) /xc*half*dt
              sv0 = sv0 + (-u*v+A*B/r)/xc*half*dt
              sp0 = sp0 - gamma*p*u   /xc*half*dt
              sC0 = sC0 + (w*A-u*C)   /xc*half*dt
              !Face-centered B-field
              sAL0 = +(GLR   -GLL   )/xL*dtdy*half -(FLR-FLL)   *dtdz*half
              sAR0 = +(GRR   -GRL   )/xR*dtdy*half -(FRR-FRL)   *dtdz*half
              sBL0 = -(GRL   -GLL   )   *dtdx*half +(ELR-ELL)   *dtdz*half
              sBR0 = -(GRR   -GLR   )   *dtdx*half +(ERR-ERL)   *dtdz*half
              sCL0 = +(FRL*xR-FLL*xL)/xc*dtdx*half -(ERL-ELL)/xc*dtdy*half
              sCR0 = +(FRR*xR-FLR*xL)/xc*dtdx*half -(ERR-ELR)/xc*dtdy*half
           endif
           if (Spherique) then
              sr0 = (-u*drx-dux*r)              *dtdx + (-v*dry-dvy*r)              /xc*dtdy + (-w*drz-dwz*r)              /xc/sinxc*dtdz 
              su0 = (-u*dux-(dpx+B*dBx+C*dCx)/r)*dtdx + (-v*duy+B*dAy/r)            /xc*dtdy + (-w*duz+C*dAz/r)            /xc/sinxc*dtdz 
              sv0 = (-u*dvx+A*dBx/r)            *dtdx + (-v*dvy-(dpy+A*dAy+C*dCy)/r)/xc*dtdy + (-w*dvz+C*dBz/r)            /xc/sinxc*dtdz
              sw0 = (-u*dwx+A*dCx/r)            *dtdx + (-v*dwy+B*dCy/r)            /xc*dtdy + (-w*dwz-(dpz+A*dAz+B*dBz)/r)/xc/sinxc*dtdz 
              sp0 = (-u*dpx-dux*gamma*p)        *dtdx + (-v*dpy-dvy*gamma*p)        /xc*dtdy + (-w*dpz-dwz*gamma*p)        /xc/sinxc*dtdz
              sA0 =                                     (u*dBy+B*duy-v*dAy-A*dvy)   /xc*dtdy + (u*dCz+C*duz-w*dAz-A*dwz)   /xc/sinxc*dtdz
              sB0 = (v*dAx+A*dvx-u*dBx-B*dux)   *dtdx +                                        (v*dCz+C*dvz-w*dBz-B*dwz)   /xc/sinxc*dtdz 
              sC0 = (w*dAx+A*dwx-u*dCx-C*dux)   *dtdx + (w*dBy+B*dwy-v*dCy-C*dvy)   /xc*dtdy
              !Geometrical terms
              sr0 = sr0 - (two*u+cotanxc*v)*r             /xc*half*dt
              su0 = su0 + (v*v+w*w-(B*B+C*C)/r)           /xc*half*dt
              sv0 = sv0 + (-u*v+A*B/r+cotanxc*(w*w-C*C/r))/xc*half*dt
              sw0 = sw0 + (-u*w+A*C/r+cotanxc*(v*w-B*C/r))/xc*half*dt
              sp0 = sp0 - gamma*p*(two*u+cotanxc*v)       /xc*half*dt
              sA0 = sA0 + cotanxc*(u*B-v*A)               /xc*half*dt
              sB0 = sB0 -         (u*B-v*A)               /xc*half*dt
              sC0 = sC0 +         (w*A-u*C)               /xc*half*dt
              !Face-centered B-field
              sAL0 = +(GLR*sinxR-GLL*sinxL)/xL/sinxc*half*dtdy - (FLR-FLL)/xc/sinxc*half*dtdz
              sAR0 = +(GRR*sinxR-GRL*sinxL)/xR/sinxc*half*dtdy - (FRR-FRL)/xc/sinxc*half*dtdz
              sBL0 = -(GRL*xR   -GLL*xL   )/xc      *half*dtdx + (ELR-ELL)/xc/sinxc*half*dtdz
              sBR0 = -(GRR*xR   -GLR*xL   )/xc      *half*dtdx + (ERR-ERL)/xc/sinxc*half*dtdz
              sCL0 = +(FRL*xR   -FLL*xL   )/xc      *half*dtdx - (ERL-ELL)/xc      *half*dtdy
              sCR0 = +(FRR*xR   -FLR*xL   )/xc      *half*dtdx - (ERR-ELR)/xc      *half*dtdy
           endif

           
           !Update in time the  primitive variables
           r = r + sr0
           u = u + su0
           v = v + sv0
           w = w + sw0
           p = p + sp0
           A = A + sA0
           B = B + sB0
           C = C + sC0
           
           AL = AL + sAL0
           AR = AR + sAR0
           BL = BL + sBL0
           BR = BR + sBR0
           CL = CL + sCL0
           CR = CR + sCR0
           
           ! Face averaged right state at left interface
           qp(i,j,k,ir,1) = r - drx
           qp(i,j,k,iu,1) = u - dux
           qp(i,j,k,iv,1) = v - dvx
           qp(i,j,k,iw,1) = w - dwx
           qp(i,j,k,ip,1) = p - dpx
           qp(i,j,k,iA,1) = AL     
           qp(i,j,k,iB,1) = B - dBx
           qp(i,j,k,iC,1) = C - dCx
           qp(i,j,k,ir,1) = max(smallr, qp(i,j,k,ir,1))
           qp(i,j,k,ip,1) = max(smallp, qp(i,j,k,ip,1))
           
           ! Face averaged left state at right interface
           qm(i,j,k,ir,1) = r + drx
           qm(i,j,k,iu,1) = u + dux
           qm(i,j,k,iv,1) = v + dvx
           qm(i,j,k,iw,1) = w + dwx
           qm(i,j,k,ip,1) = p + dpx
           qm(i,j,k,iA,1) = AR     
           qm(i,j,k,iB,1) = B + dBx
           qm(i,j,k,iC,1) = C + dCx
           qm(i,j,k,ir,1) = max(smallr, qm(i,j,k,ir,1))
           qm(i,j,k,ip,1) = max(smallp, qm(i,j,k,ip,1))
           
           ! Face averaged top state at bottom interface
           qp(i,j,k,ir,2) = r - dry
           qp(i,j,k,iu,2) = u - duy
           qp(i,j,k,iv,2) = v - dvy
           qp(i,j,k,iw,2) = w - dwy
           qp(i,j,k,ip,2) = p - dpy
           qp(i,j,k,iA,2) = A - dAy
           qp(i,j,k,iB,2) = BL     
           qp(i,j,k,iC,2) = C - dCy
           qp(i,j,k,ir,2) = max(smallr, qp(i,j,k,ir,2))
           qp(i,j,k,ip,2) = max(smallp, qp(i,j,k,ip,2))
           
           ! Face averaged bottom state at top interface
           qm(i,j,k,ir,2) = r + dry
           qm(i,j,k,iu,2) = u + duy
           qm(i,j,k,iv,2) = v + dvy
           qm(i,j,k,iw,2) = w + dwy
           qm(i,j,k,ip,2) = p + dpy
           qm(i,j,k,iA,2) = A + dAy
           qm(i,j,k,iB,2) = BR     
           qm(i,j,k,iC,2) = C + dCy
           qm(i,j,k,ir,2) = max(smallr, qm(i,j,k,ir,2))
           qm(i,j,k,ip,2) = max(smallp, qm(i,j,k,ip,2))
           
           ! Face averaged front state at back interface
           qp(i,j,k,ir,3) = r - drz
           qp(i,j,k,iu,3) = u - duz
           qp(i,j,k,iv,3) = v - dvz
           qp(i,j,k,iw,3) = w - dwz
           qp(i,j,k,ip,3) = p - dpz
           qp(i,j,k,iA,3) = A - dAz
           qp(i,j,k,iB,3) = B - dBz
           qp(i,j,k,iC,3) = CL     
           qp(i,j,k,ir,3) = max(smallr, qp(i,j,k,ir,3))
           qp(i,j,k,ip,3) = max(smallp, qp(i,j,k,ip,3))
           
           ! Face averaged back state at front interface
           qm(i,j,k,ir,3) = r + drz
           qm(i,j,k,iu,3) = u + duz
           qm(i,j,k,iv,3) = v + dvz
           qm(i,j,k,iw,3) = w + dwz
           qm(i,j,k,ip,3) = p + dpz
           qm(i,j,k,iA,3) = A + dAz
           qm(i,j,k,iB,3) = B + dBz
           qm(i,j,k,iC,3) = CR     
           qm(i,j,k,ir,3) = max(smallr, qm(i,j,k,ir,3))
           qm(i,j,k,ip,3) = max(smallp, qm(i,j,k,ip,3))
           
           ! X-edge averaged right-top corner state (RT->LL)
           qRT(i,j,k,ir,1) = r + (+dry+drz)
           qRT(i,j,k,iu,1) = u + (+duy+duz)
           qRT(i,j,k,iv,1) = v + (+dvy+dvz)
           qRT(i,j,k,iw,1) = w + (+dwy+dwz)
           qRT(i,j,k,ip,1) = p + (+dpy+dpz)
           qRT(i,j,k,iA,1) = A + (+dAy+dAz)
           qRT(i,j,k,iB,1) = BR+ (   +dBRz)
           qRT(i,j,k,iC,1) = CR+ (+dCRy   )
           qRT(i,j,k,ir,1) = max(smallr, qRT(i,j,k,ir,1))
           qRT(i,j,k,ip,1) = max(smallp, qRT(i,j,k,ip,1))
           
           ! X-edge averaged right-bottom corner state (RB->LR)
           qRB(i,j,k,ir,1) = r + (+dry-drz)
           qRB(i,j,k,iu,1) = u + (+duy-duz)
           qRB(i,j,k,iv,1) = v + (+dvy-dvz)
           qRB(i,j,k,iw,1) = w + (+dwy-dwz)
           qRB(i,j,k,ip,1) = p + (+dpy-dpz)
           qRB(i,j,k,iA,1) = A + (+dAy-dAz)
           qRB(i,j,k,iB,1) = BR+ (   -dBRz)
           qRB(i,j,k,iC,1) = CL+ (+dCLy   )
           qRB(i,j,k,ir,1) = max(smallr, qRB(i,j,k,ir,1))
           qRB(i,j,k,ip,1) = max(smallp, qRB(i,j,k,ip,1))
           
           ! X-edge averaged left-top corner state (LT->RL)
           qLT(i,j,k,ir,1) = r + (-dry+drz)
           qLT(i,j,k,iu,1) = u + (-duy+duz)
           qLT(i,j,k,iv,1) = v + (-dvy+dvz)
           qLT(i,j,k,iw,1) = w + (-dwy+dwz)
           qLT(i,j,k,ip,1) = p + (-dpy+dpz)
           qLT(i,j,k,iA,1) = A + (-dAy+dAz)
           qLT(i,j,k,iB,1) = BL+ (   +dBLz)
           qLT(i,j,k,iC,1) = CR+ (-dCRy   )
           qLT(i,j,k,ir,1) = max(smallr, qLT(i,j,k,ir,1))
           qLT(i,j,k,ip,1) = max(smallp, qLT(i,j,k,ip,1))
           
           ! X-edge averaged left-bottom corner state (LB->RR)
           qLB(i,j,k,ir,1) = r + (-dry-drz)
           qLB(i,j,k,iu,1) = u + (-duy-duz)
           qLB(i,j,k,iv,1) = v + (-dvy-dvz)
           qLB(i,j,k,iw,1) = w + (-dwy-dwz)
           qLB(i,j,k,ip,1) = p + (-dpy-dpz)
           qLB(i,j,k,iA,1) = A + (-dAy-dAz)
           qLB(i,j,k,iB,1) = BL+ (   -dBLz)
           qLB(i,j,k,iC,1) = CL+ (-dCLy   )
           qLB(i,j,k,ir,1) = max(smallr, qLB(i,j,k,ir,1))
           qLB(i,j,k,ip,1) = max(smallp, qLB(i,j,k,ip,1))
           
           ! Y-edge averaged right-top corner state (RT->LL)
           qRT(i,j,k,ir,2) = r + (+drx+drz)
           qRT(i,j,k,iu,2) = u + (+dux+duz)
           qRT(i,j,k,iv,2) = v + (+dvx+dvz)
           qRT(i,j,k,iw,2) = w + (+dwx+dwz)
           qRT(i,j,k,ip,2) = p + (+dpx+dpz)
           qRT(i,j,k,iA,2) = AR+ (   +dARz)
           qRT(i,j,k,iB,2) = B + (+dBx+dBz)
           qRT(i,j,k,iC,2) = CR+ (+dCRx   )
           qRT(i,j,k,ir,2) = max(smallr, qRT(i,j,k,ir,2))
           qRT(i,j,k,ip,2) = max(smallp, qRT(i,j,k,ip,2))
           
           ! Y-edge averaged right-bottom corner state (RB->LR)
           qRB(i,j,k,ir,2) = r + (+drx-drz)
           qRB(i,j,k,iu,2) = u + (+dux-duz)
           qRB(i,j,k,iv,2) = v + (+dvx-dvz)
           qRB(i,j,k,iw,2) = w + (+dwx-dwz)
           qRB(i,j,k,ip,2) = p + (+dpx-dpz)
           qRB(i,j,k,iA,2) = AR+ (   -dARz)
           qRB(i,j,k,iB,2) = B + (+dBx-dBz)
           qRB(i,j,k,iC,2) = CL+ (+dCLx   )
           qRB(i,j,k,ir,2) = max(smallr, qRB(i,j,k,ir,2))
           qRB(i,j,k,ip,2) = max(smallp, qRB(i,j,k,ip,2))
           
           ! Y-edge averaged left-top corner state (LT->RL)
           qLT(i,j,k,ir,2) = r + (-drx+drz)
           qLT(i,j,k,iu,2) = u + (-dux+duz)
           qLT(i,j,k,iv,2) = v + (-dvx+dvz)
           qLT(i,j,k,iw,2) = w + (-dwx+dwz)
           qLT(i,j,k,ip,2) = p + (-dpx+dpz)
           qLT(i,j,k,iA,2) = AL+ (   +dALz)
           qLT(i,j,k,iB,2) = B + (-dBx+dBz)
           qLT(i,j,k,iC,2) = CR+ (-dCRx   )
           qLT(i,j,k,ir,2) = max(smallr, qLT(i,j,k,ir,2))
           qLT(i,j,k,ip,2) = max(smallp, qLT(i,j,k,ip,2))
           
           ! Y-edge averaged left-bottom corner state (LB->RR)
           qLB(i,j,k,ir,2) = r + (-drx-drz)
           qLB(i,j,k,iu,2) = u + (-dux-duz)
           qLB(i,j,k,iv,2) = v + (-dvx-dvz)
           qLB(i,j,k,iw,2) = w + (-dwx-dwz)
           qLB(i,j,k,ip,2) = p + (-dpx-dpz)
           qLB(i,j,k,iA,2) = AL+ (   -dALz)
           qLB(i,j,k,iB,2) = B + (-dBx-dBz)
           qLB(i,j,k,iC,2) = CL+ (-dCLx   )
           qLB(i,j,k,ir,2) = max(smallr, qLB(i,j,k,ir,2))
           qLB(i,j,k,ip,2) = max(smallp, qLB(i,j,k,ip,2))
           
           ! Z-edge averaged right-top corner state (RT->LL)
           qRT(i,j,k,ir,3) = r + (+drx+dry)
           qRT(i,j,k,iu,3) = u + (+dux+duy)
           qRT(i,j,k,iv,3) = v + (+dvx+dvy)
           qRT(i,j,k,iw,3) = w + (+dwx+dwy)
           qRT(i,j,k,ip,3) = p + (+dpx+dpy)
           qRT(i,j,k,iA,3) = AR+ (   +dARy)
           qRT(i,j,k,iB,3) = BR+ (+dBRx   )
           qRT(i,j,k,iC,3) = C + (+dCx+dCy)
           qRT(i,j,k,ir,3) = max(smallr, qRT(i,j,k,ir,3))
           qRT(i,j,k,ip,3) = max(smallp, qRT(i,j,k,ip,3))
           
           ! Z-edge averaged right-bottom corner state (RB->LR)
           qRB(i,j,k,ir,3) = r + (+drx-dry)
           qRB(i,j,k,iu,3) = u + (+dux-duy)
           qRB(i,j,k,iv,3) = v + (+dvx-dvy)
           qRB(i,j,k,iw,3) = w + (+dwx-dwy)
           qRB(i,j,k,ip,3) = p + (+dpx-dpy)
           qRB(i,j,k,iA,3) = AR+ (   -dARy)
           qRB(i,j,k,iB,3) = BL+ (+dBLx   )
           qRB(i,j,k,iC,3) = C + (+dCx-dCy)
           qRB(i,j,k,ir,3) = max(smallr, qRB(i,j,k,ir,3))
           qRB(i,j,k,ip,3) = max(smallp, qRB(i,j,k,ip,3))
           
           ! Z-edge averaged left-top corner state (LT->RL)
           qLT(i,j,k,ir,3) = r + (-drx+dry)
           qLT(i,j,k,iu,3) = u + (-dux+duy)
           qLT(i,j,k,iv,3) = v + (-dvx+dvy)
           qLT(i,j,k,iw,3) = w + (-dwx+dwy)
           qLT(i,j,k,ip,3) = p + (-dpx+dpy)
           qLT(i,j,k,iA,3) = AL+ (   +dALy)
           qLT(i,j,k,iB,3) = BR+ (-dBRx   )
           qLT(i,j,k,iC,3) = C + (-dCx+dCy)
           qLT(i,j,k,ir,3) = max(smallr, qLT(i,j,k,ir,3))
           qLT(i,j,k,ip,3) = max(smallp, qLT(i,j,k,ip,3))
           
           ! Z-edge averaged left-bottom corner state (LB->RR)
           qLB(i,j,k,ir,3) = r + (-drx-dry)
           qLB(i,j,k,iu,3) = u + (-dux-duy)
           qLB(i,j,k,iv,3) = v + (-dvx-dvy)
           qLB(i,j,k,iw,3) = w + (-dwx-dwy)
           qLB(i,j,k,ip,3) = p + (-dpx-dpy)
           qLB(i,j,k,iA,3) = AL+ (   -dALy)
           qLB(i,j,k,iB,3) = BL+ (-dBLx   )
           qLB(i,j,k,iC,3) = C + (-dCx-dCy)
           qLB(i,j,k,ir,3) = max(smallr, qLB(i,j,k,ir,3))
           qLB(i,j,k,ip,3) = max(smallp, qLB(i,j,k,ip,3))
           
        end do
     end do
  end do

  ! Passive scalars
  do n = 9, nvar
     do k = klo, khi
        do j = jlo, jhi
           do i = ilo, ihi
              r   = q(i,j,k,n )              ! Cell centered values
              u   = q(i,j,k,iu)
              v   = q(i,j,k,iv)
              w   = q(i,j,k,iw)
              drx = half * dq(i,j,k,n,1)     ! TVD slopes
              dry = half * dq(i,j,k,n,2)
              drz = half * dq(i,j,k,n,3)
              sr0 = -u*drx*dtdx -v*dry*dtdy -w*drz*dtdz    ! Source terms
              qp(i,j,k,n,1) = r - drx + sr0  ! Right state
              qm(i,j,k,n,1) = r + drx + sr0  ! Left state
              qp(i,j,k,n,2) = r - dry + sr0  ! Top state
              qm(i,j,k,n,2) = r + dry + sr0  ! Bottom state
              qp(i,j,k,n,3) = r - drz + sr0  ! Front state
              qm(i,j,k,n,3) = r + drz + sr0  ! Back state
           end do
        end do
     end do
  end do

  deallocate(Ex,Ey,Ez)
  
  return

end subroutine trace3d
