!> \file
!! contains subroutine update()
!<

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine UPDATE
!
!> Updates the conservative variables after the mhd step
!! using the computed mhd fluxes.
!<
subroutine update(uin,flux,flux_pre_tot,emfx,emfy,emfz,            &
           &      iu1,iu2,ju1,ju2,ku1,ku2,if1,if2,jf1,jf2,kf1,kf2, &
           &      nvar,ndim)
           
  use parameters, only : n_vit
  use divers    , only : verbose
  use geom      , only : dx,xcloc,Cartesien,Cylindrique,Spherique,dv
  use var_loc   , only : xloc
  
  implicit none

  integer :: nvar,ndim
  integer :: iu1,iu2,ju1,ju2,ku1,ku2
  integer :: if1,if2,jf1,jf2,kf1,kf2
  integer :: i,j,k,idim,ibx,iby,ibz
  integer :: i0,j0,k0,ihi,jhi,khi,ivit
  
  real, dimension(iu1:iu2,ju1:ju2,ku1:ku2,1:nvar+3       ) :: uin 
  real, dimension(if1:if2,jf1:jf2,kf1:kf2,1:nvar  ,1:ndim) :: flux
  real, dimension(if1:if2,jf1:jf2,kf1:kf2         ,1:ndim) :: flux_pre_tot
  real, dimension(iu1:iu2,ju1:ju2,ku1:ku2                ) :: emfx,emfy,emfz
  real, dimension(3      ,2      ,3                      ) :: radius
  real                                                     :: xL,xR,xc,sinxL,sinxR,sinxc

  if (verbose) write (*,*) 'Start of conservative mhd update'

  ibx=6 ; iby=7 ; ibz=8

  ihi=max(1,if2-1)
  jhi=max(1,jf2-1)
  khi=max(1,kf2-1)

  do k=1,khi
     do j=1,jhi
        do i=1,ihi
           
           radius = 1.
           if (Cylindrique .or. Spherique) then
              radius(2,1:2,1:3) = xcloc(i  ,1)
              radius(2,1  ,1  ) =  xloc(i  ,1)
              radius(2,2  ,1  ) =  xloc(i+1,1)
           endif
              
           if (Spherique) then
              radius(3,1  ,1  ) =  xloc(i  ,1) * sin(xcloc(j  ,2))
              radius(3,2  ,1  ) =  xloc(i+1,1) * sin(xcloc(j  ,2))
              radius(3,1  ,2  ) = xcloc(i  ,1) * sin( xloc(j  ,2))
              radius(3,2  ,2  ) = xcloc(i  ,1) * sin( xloc(j+1,2))
              radius(3,1  ,3  ) = xcloc(i  ,1) * sin(xcloc(j  ,2))
              radius(3,2  ,3  ) = xcloc(i  ,1) * sin(xcloc(j  ,2))
           endif

           do idim=1,ndim

              i0=0; j0=0; k0=0 
              if(idim==1)i0=1
              if(idim==2)j0=1
              if(idim==3)k0=1
              
              !rho
              uin(i,j,k,1)=uin(i,j,k,1) + (flux(i   ,j   ,k   ,1,idim) &
                   &                    -  flux(i+i0,j+j0,k+k0,1,idim))/dv(i,j,k)

              !rho*v
              do ivit=2,4
                 uin(i,j,k,ivit)=uin(i,j,k,ivit) + (flux(i   ,j   ,k   ,ivit,idim)*radius(ivit-1,1,idim)  &
                   &                             -  flux(i+i0,j+j0,k+k0,ivit,idim)*radius(ivit-1,2,idim)) &
                   &                                                    /dv(i,j,k)/radius(ivit-1,2,3)
              end do

              !Etot
              uin(i,j,k,5)=uin(i,j,k,5) + (flux(i   ,j   ,k   ,5,idim) &
                   &                    -  flux(i+i0,j+j0,k+k0,5,idim))/dv(i,j,k)

              !Cell-centered B-field (only in 1 & 2D)
              if (ndim==1) then
                 uin(i,j,k,n_vit+3:nvar)=uin(i,j,k,n_vit+3:nvar)+ &
                      &          (flux(i   ,j   ,k   ,n_vit+3:nvar,idim) &
                      &          -flux(i+i0,j+j0,k+k0,n_vit+3:nvar,idim))/dv(i,j,k)
              else if (ndim==2) then
                 uin(i,j,k,nvar)=uin(i,j,k,nvar)+ &
                      &          (flux(i   ,j   ,k   ,nvar,idim) &
                      &          -flux(i+i0,j+j0,k+k0,nvar,idim))/dv(i,j,k)
              endif

           end do

           !Add presure gradient and geometrical terms
                       uin(i,j,k,2) = uin(i,j,k,2) + flux_pre_tot(i,j,k,1)/dv(i,j,k)
           if (ndim>1) uin(i,j,k,3) = uin(i,j,k,3) + flux_pre_tot(i,j,k,2)/dv(i,j,k)
           if (ndim>2) uin(i,j,k,4) = uin(i,j,k,4) + flux_pre_tot(i,j,k,3)/dv(i,j,k)

        end do
     end do
  end do

  !Constrained transport for face-centered B-field
  if (Cartesien) then
     if (ndim>=2) then
        do k=1,khi
           do j=1,jhi+1
              do i=1,ihi+1
                 uin(i,j,k,ibx)=uin(i,j,k,ibx)+(emfz(i  ,j+1,k  )-emfz(i,j,k))/dx(j,2)
                 uin(i,j,k,iby)=uin(i,j,k,iby)-(emfz(i+1,j  ,k  )-emfz(i,j,k))/dx(i,1)
              end do
           end do
        end do
     endif
     if (ndim==3) then
        do k=1,khi+1
           do j=1,jhi+1
              do i=1,ihi+1
                 uin(i,j,k,ibx)=uin(i,j,k,ibx)-(emfy(i  ,j  ,k+1)-emfy(i,j,k))/dx(k,3)
                 uin(i,j,k,iby)=uin(i,j,k,iby)+(emfx(i  ,j  ,k+1)-emfx(i,j,k))/dx(k,3)
                 
                 uin(i,j,k,ibz)=uin(i,j,k,ibz)+(emfy(i+1,j  ,k  )-emfy(i,j,k))/dx(i,1) &
                               &              -(emfx(i  ,j+1,k  )-emfx(i,j,k))/dx(j,2)
              end do
           end do
        end do
     endif
  endif
  if (Cylindrique) then
     if (ndim>=2) then
        do k=1,khi
           do j=1,jhi+1
              do i=1,ihi+1
                 uin(i,j,k,ibx)=uin(i,j,k,ibx)+(emfz(i  ,j+1,k  )-emfz(i,j,k))/dx(j,2)/xloc(i,1)
                 uin(i,j,k,iby)=uin(i,j,k,iby)-(emfz(i+1,j  ,k  )-emfz(i,j,k))/dx(i,1)
              end do
           end do
        end do
     endif
     if (ndim==3) then
        do k=1,khi+1
           do j=1,jhi+1
              do i=1,ihi+1
                 uin(i,j,k,ibx)=uin(i,j,k,ibx)-(emfy(i  ,j  ,k+1)-emfy(i,j,k))/dx(k,3)
                 uin(i,j,k,iby)=uin(i,j,k,iby)+(emfx(i  ,j  ,k+1)-emfx(i,j,k))/dx(k,3)
                 
                 uin(i,j,k,ibz)=uin(i,j,k,ibz)+(emfy(i+1,j  ,k  )*xloc(i+1,1)-emfy(i,j,k)*xloc(i,1))/xcloc(i,1)/dx(i,1) &
                                            & -(emfx(i  ,j+1,k  )            -emfx(i,j,k)          )/xcloc(i,1)/dx(j,2)
              end do
           end do
        end do
     endif
  endif
  if (Spherique) then
     if (ndim>=2) then
        do k=1,khi
           do j=1,jhi+1
              sinxL = sin(xloc (j  ,2))
              sinxR = sin(xloc (j+1,2))
              sinxc = sin(xcloc(j  ,2))
              do i=1,ihi+1
                 xL =  xloc(i  ,1)
                 xR =  xloc(i+1,1)
                 xc = xcloc(i  ,1)
                 uin(i,j,k,ibx)=uin(i,j,k,ibx)+(emfz(i  ,j+1,k  )*sinxR-emfz(i,j,k)*sinxL)/dx(j,2)/xL/sinxc
                 uin(i,j,k,iby)=uin(i,j,k,iby)-(emfz(i+1,j  ,k  )*xR   -emfz(i,j,k)*xL   )/dx(i,1)/xc
              end do
           end do
        end do
     endif
     if (ndim==3) then
        do k=1,khi+1
           do j=1,jhi+1
              sinxL = sin(xloc (j  ,2))
              sinxR = sin(xloc (j+1,2))
              sinxc = sin(xcloc(j  ,2))
              do i=1,ihi+1
                 xL =  xloc(i  ,1)
                 xR =  xloc(i+1,1)
                 xc = xcloc(i  ,1)
                 uin(i,j,k,ibx)=uin(i,j,k,ibx)-(emfy(i  ,j  ,k+1)-emfy(i,j,k))/xL/sinxc/dx(k,3)
                 uin(i,j,k,iby)=uin(i,j,k,iby)+(emfx(i  ,j  ,k+1)-emfx(i,j,k))/xc/sinxL/dx(k,3)
                 
                 uin(i,j,k,ibz)=uin(i,j,k,ibz)+(emfy(i+1,j  ,k  )*xR-emfy(i,j,k)*xL)/xc/dx(i,1) &
                      &                       -(emfx(i  ,j+1,k  )   -emfx(i,j,k)   )/xc/dx(j,2)
              end do
           end do
        end do
     endif
  endif

  if (verbose) write (*,*) 'End of conservative mhd update'

  return
  
end subroutine update
