!> \file
!! Contains subroutines mat_prod3d(), com_1D_xin(),
!! com_2D_xin(), com_3D_xin() and limites_xin()
!<

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine MAT_PROD3D
!
!> Computes the A*xin product and stocks it in the xout
!! vector.
!!
!! In order to treat the boundary conditions properly, the
!! xin and xout vectors have a ghost cell on each side.
!!
!! Moreover, the stocking formula for the xin and xout
!! vectors is:
!! x((i-imin+(j-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+ivar)
!! corresponds to the value of the variable ivar (1,2,3,4,5
!! for Tgaz, Eray, Frayx, Frayy, Frayz respectively) in the
!! (i,j,k) cell.
!<
subroutine mat_prod3d(xin,xout)

  use parameters, only : ndim,nxmin,nxmax,nx,nvar_ray,pi
  use divers_ray, only : alpha,precond_ray,cal_valp,valp_min,rad_trans_model,slope_type_ray
  use divers    , only : mu,dt,verbose,hydro
  use cputime   , only : temps_comm
  use unites    , only : c,c2
  use geom      , only : x,dx,ds,dv,dxc,xcloc,cylindrique,spherique,geom_dir
  use var_loc   , only : xloc
  use varray    , only : eray,fray,kappa_abs_ray,kappa_emi_ray,kappa_sca_ray,erayold,frayold
  use var       , only : tgaz,rho
  use varold    , only : tgazold
  use cl_ray    , only : icl_ray,er_l,fr_l
  use tabgmres  , only : b,imin,imax,jmin,jmax,kmin,kmax

  implicit none

#if PARA==1
  include 'mpif.h'
  real    :: tr1,tr2
  integer :: ierr
#endif

  integer, dimension(2,3,2,3) :: iv
  integer                     :: i,j,k,ivar,jvar,nn,jjmin,jjmax,jjpas,iloc
  integer                     :: iface,idp,idm,i1
  integer                     :: ip,jp,kp,im,jm,km,idim,jdim,kdim

  real, dimension(-nvar_ray+1:(imax-imin+(jmax-jmin)*(imax-imin+1)+(kmax-kmin)*(imax-imin+1)*(jmax-jmin+1))*nvar_ray) :: xin
  real, dimension(          1: nx(1)                *nx(2)                                  *nx(3)         *nvar_ray) :: xout
  real, dimension(nxmin(1):nxmax(1),nxmin(2):nxmax(2),nxmin(3):nxmax(3),1:ndim       ) :: peray,pfray1,pdummy
  real, dimension(nxmin(1):nxmax(1),nxmin(2):nxmax(2),nxmin(3):nxmax(3),1:ndim,1:ndim) :: pfray
  real, dimension(nxmin(1):nxmax(1),nxmin(2):nxmax(2),nxmin(3):nxmax(3)              ) :: eray0,tgaz0
  real, dimension(nxmin(1):nxmax(1),nxmin(2):nxmax(2),nxmin(3):nxmax(3),1:ndim       ) :: fray0
  real, dimension(nvar_ray,2*nvar_ray) :: a1d
  real, dimension(nvar_ray,2*nvar_ray) :: a2d
  real, dimension(nvar_ray,2*nvar_ray) :: a3d
  real, dimension(1:3,1:3,2,ndim)      :: praym,prayp,deddp,deddm
  real, dimension(        2,ndim)      :: eps,ap,am,lm,lp,lmp
  real, dimension(     2,ndim)         :: eraym,erayp,deray
  real, dimension(ndim,2,ndim)         :: fraym,frayp,dfray
  real, dimension(     2,ndim)         :: ffm, ffp
  real, dimension(3,3       )          :: Dedd,Dedd_dE
  real, dimension(3,3,1:ndim)          :: Dedd_dF
  real, dimension(3,3,       2,1:ndim) :: DeddM_dE,DeddP_dE
  real, dimension(3,3,1:ndim,2,1:ndim) :: DeddM_dF,DeddP_dF
  real, dimension(3,3       ,2)        :: Dedd_1d,Dedd_dE_1d,Dedd_2d,Dedd_dE_2d,Dedd_3d,Dedd_dE_3d
  real, dimension(3,3,1:ndim,2)        :: Dedd_dF_1d,Dedd_dF_2d,Dedd_dF_3d
  real, dimension(1:5,1:5)             :: mat,inv
  real, dimension(2,1:ndim)            :: dSloc
  real, dimension(     3,2,3     )     :: Rayon
  real                                 :: drp,drm
  real                                 :: ff1,ff2
  real                                 :: thetam,thetap,interpol_valp
  real                                 :: xx,conv,derive_source_ray

  logical :: chi_uniforme,derive_chi

!*********************************************************************************************************************************
!  Dedd,Dedd_dE (i,j)
!  Dedd_dF      (i,j,k)
!
!  DeddM_dE, DeddP_dE  (i,j  ,iface,idim)
!  DeddM_dF, DeddP_dF  (i,j,k,iface,idim)
!  
!  Dedd_1d,Dedd_dE_1d,Dedd_2d,Dedd_dE_2d,Dedd_3d,Dedd_dE_3d  (i,j  ,iface)
!  Dedd_dF_1d,Dedd_dF_2d,Dedd_dF_3d                          (i,j,k,iface)
!
!   iface : face index            (=1:   2) with 1 = back ; 2 = front
!   idim  : direction considered  (=1:ndim)
!
!   i,j   : tensorial coordinates       (=1:3   )
!   k     : flux direction considered   (=1:ndim)
!
!*********************************************************************************************************************************

  if (verbose) write(*,*) 'Start of mat_prod3d'

  ! In order to have a uniform chi in the cells
  Chi_uniforme = .false.

  ! In order to include the derivatives in the matrix
  derive_chi = .true. ! besoin dans cal_Dedd

  ! geometry
  nn = max(nx(1),nx(2),nx(3))
  allocate(xloc(-1:nn+3,ndim))
  
                 xloc(-1:nx(1)+3,1) = x(-1:nx(1)+3,1)   
  if (ndim.gt.1) xloc(-1:nx(2)+3,2) = x(-1:nx(2)+3,2)   
  if (ndim.gt.2) xloc(-1:nx(3)+3,3) = x(-1:nx(3)+3,3)   

  call cal_geom(nx(1),nx(2),nx(3))

  ! neighbour offsets
  ! Iv(i,j,k,l) :: i = front or back face -- j = face direction -- k = front or back cell -- l = cell coordinate
  Iv = 0
  do idim=1,3
     Iv(1,idim,1,idim) = -1
     Iv(2,idim,2,idim) = +1
  enddo

#if PARA==1
  tr1 = MPI_Wtime(ierr)
  ! Communications
  if (ndim.eq.1) then
     call com_1D_xin(xin)
  else if (ndim.eq.2) then
     call com_2D_xin(xin)
  else if (ndim.eq.3) then
     call com_3D_xin(xin)
  endif
  tr2 = MPI_Wtime(ierr)
  temps_comm = temps_comm + tr2-tr1
#endif

  ! Boundary conditions
  call limites_xin(xin)

  !~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  !         !!!!!!!!!!!!!!!     CAUTION             !!!!!!!!!!!!!!!!!!
  ! Do NOT re-compute kappa since it was alread done in cal_b3d abd it needs to be identical
  !~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
!  call cal_kappa_ray(rho,Tgaz,kappa_emi_ray,kappa_sca_ray,nxmin(1),nxmax(1),nxmin(2),nxmax(2),nxmin(3),nxmax(3))
!  kappa_abs_ray = kappa_emi_ray

  Eray0 = Erayold
  Fray0 = Frayold
  Tgaz0 = Tgazold


  if (verbose) write(*,*) '     -- Eray slope --'
  Call slopes(Eray0,peray,pdummy,nxmin(1),nxmax(1),nxmin(2),nxmax(2),nxmin(3),nxmax(3),slope_type_ray)
  if (verbose) write(*,*) '     -- Fray slope --'
  do idim =1,ndim
     Call slopes(Fray0(nxmin(1):nxmax(1),nxmin(2):nxmax(2),nxmin(3):nxmax(3),idim),pfray1,pdummy,&
                 nxmin(1),nxmax(1),nxmin(2),nxmax(2),nxmin(3),nxmax(3),slope_type_ray)
     pfray(nxmin(1):nxmax(1),nxmin(2):nxmax(2),nxmin(3):nxmax(3),idim,1:ndim) = pfray1(nxmin(1):nxmax(1),nxmin(2):nxmax(2),nxmin(3):nxmax(3),1:ndim)
  enddo


  do k = 1,nx(3)
     do j = 1,nx(2)
        do i = 1,nx(1)

           PrayM = 0. ; PrayP = 0.

           if (verbose) write(*,*) '     -- Interfaces values--'
   
           do idim  = 1,ndim
           do iface = 1,2
   
              im = i + Iv(iface,idim,1,1) ; jm = j + Iv(iface,idim,1,2); km = k + Iv(iface,idim,1,3)
              ip = i + Iv(iface,idim,2,1) ; jp = j + Iv(iface,idim,2,2); kp = k + Iv(iface,idim,2,3)
   
              if(idim.eq.1) then
                 idm = im
                 idp = ip
              else if(idim.eq.2) then
                 idm = jm
                 idp = jp
              else if(idim.eq.3) then
                 idm = km
                 idp = kp
              endif
   
              drp = 0.5*(x(idp  ,idim)-x(idm,idim))
              drm = 0.5*(x(idp+1,idim)-x(idp,idim)) 
   
              ErayM(iface,idim) = Eray0(im,jm,km) + drp*peray(im,jm,km,idim)  
              ErayP(iface,idim) = Eray0(ip,jp,kp) - drm*peray(ip,jp,kp,idim)
   
              if(iface.eq.1) then
                 dEray(iface,idim) = ErayP(iface,idim) - Eray0(i,j,k) 
              else
                 dEray(iface,idim) = ErayM(iface,idim) - Eray0(i,j,k) 
              endif
   
              ffM(iface,idim) = 0. ; ffP(iface,idim) = 0.
              do jdim = 1,ndim
                 FrayM(jdim,iface,idim) = Fray0(im,jm,km,jdim) + drp*pfray(im,jm,km,jdim,idim)
                 FrayP(jdim,iface,idim) = Fray0(ip,jp,kp,jdim) - drm*pfray(ip,jp,kp,jdim,idim)
   
                 if(iface.eq.1) then
                    dFray(jdim,iface,idim) = FrayP(jdim,iface,idim)- Fray0(i,j,k,jdim)
                 else
                    dFray(jdim,iface,idim) = FrayM(jdim,iface,idim)- Fray0(i,j,k,jdim)
                 endif
   
                 ffM(iface,idim) = ffM(iface,idim) + FrayM(jdim,iface,idim)**2
                 ffP(iface,idim) = ffP(iface,idim) + FrayP(jdim,iface,idim)**2
              enddo
              ffM(iface,idim) = sqrt(ffM(iface,idim))/(c*ErayM(iface,idim))
              ffP(iface,idim) = sqrt(ffP(iface,idim))/(c*ErayP(iface,idim))
   
              ! Calculate theta incidence angle to interpolate eigenvalues
              ! theta needs to be calculated here before ffM is changed if it is greater than 1
              if(ffM(iface,idim).gt.1.e-5) then
                 xx=FrayM(idim,iface,idim)/(c*ErayM(iface,idim))/ffM(iface,idim)
                 if(abs(xx).gt.1.) then
                    thetaM=pi/2.*(1.-sign(1.,xx))
                 else
                    thetaM=acos(xx)
                 endif
              else
                 thetaM=0.
              endif
              if(ffP(iface,idim).gt.1.e-5) then
                 xx=FrayP(idim,iface,idim)/(c*ErayP(iface,idim))/ffP(iface,idim)
                 if(abs(xx).gt.1.) then
                    thetaP=pi/2.*(1.-sign(1.,xx))
                 else
                    thetaP=acos(xx)
                 endif
              else
                 thetaP=0.
              endif

              ! If ffM > 1 then set it to 1 - WARNING: NOT VERY SATISFACTORY!! Ed
              if (ffM(iface,idim).gt. 1.) ffM(iface,idim)= 1.
              if (ffP(iface,idim).gt. 1.) ffP(iface,idim)= 1.


              ! Compute Eddington tensor and its derivatives at the cell centre and interfaces
              if (Chi_uniforme) then 
                 call cal_Dedd(Eray0(ip,jp,kp),Fray0(ip,jp,kp,1:ndim),DeddP(:,:,iface,idim),DeddP_dE(:,:,iface,idim),DeddP_dF(:,:,:,iface,idim) , Derive_chi)
                 call cal_Dedd(Eray0(im,jm,km),Fray0(im,jm,km,1:ndim),DeddM(:,:,iface,idim),DeddM_dE(:,:,iface,idim),DeddM_dF(:,:,:,iface,idim) , Derive_chi)
              else
                 call cal_Dedd(ErayP(iface,idim),FrayP(1:ndim,iface,idim),DeddP(:,:,iface,idim),DeddP_dE(:,:,iface,idim),DeddP_dF(:,:,:,iface,idim) , Derive_chi)
                 call cal_Dedd(ErayM(iface,idim),FrayM(1:ndim,iface,idim),DeddM(:,:,iface,idim),DeddM_dE(:,:,iface,idim),DeddM_dF(:,:,:,iface,idim) , Derive_chi)
              endif
   
              do jdim=1,ndim
              do kdim=1,ndim
                 PrayM(jdim,kdim,iface,idim) = DeddM(jdim,kdim,iface,idim)*ErayM(iface,idim)
                 PrayP(jdim,kdim,iface,idim) = DeddP(jdim,kdim,iface,idim)*ErayP(iface,idim)
              enddo
              enddo


              ! Compute epsilon for asymptotic preserving scheme
              if (verbose) write(*,*) '     -- compute epsilon --'
   
              ff1 = ffM(iface,idim)
              ff2 = ffP(iface,idim)

              if( ((kappa_abs_ray(im,jm,km)+kappa_sca_ray(im,jm,km))*rho(im,jm,km)).ne.0. .and. &
                  ((kappa_abs_ray(ip,jp,kp)+kappa_sca_ray(ip,jp,kp))*rho(ip,jp,kp)).ne.0.       ) then
                 eps(iface,idim) = max( 1./((kappa_abs_ray(im,jm,km)+kappa_sca_ray(im,jm,km))*rho(im,jm,km)*(2.*drm)) &
                      &                ,1./((kappa_abs_ray(ip,jp,kp)+kappa_sca_ray(ip,jp,kp))*rho(ip,jp,kp)*(2.*drp)) ) 
                 eps(iface,idim) = max(eps(iface,idim),2.*abs(ff1),2.*abs(ff2)) 
                 eps(iface,idim) = min(1.,eps(iface,idim)) 
              else
                 eps(iface,idim) = 1.
              endif
              
              ! Compute eigenvalues at the cell interfaces
              if (verbose)  write(*,*) '     -- compute eigenvalues --'
              
              if (cal_valp) then
   
                 select case(rad_trans_model)
         
                 case('P1')
         
                    am(iface,idim)=min(-valp_min,-1./sqrt(3.))
                    ap(iface,idim)=max(+valp_min, 1./sqrt(3.))
            
                 case('M1')
         
                    am(iface,idim)=min(-valp_min,interpol_valp(ffM(iface,idim),thetaM,eps(iface,idim),1), &
                                                 interpol_valp(ffP(iface,idim),thetaP,eps(iface,idim),1))
                    ap(iface,idim)=max(+valp_min,interpol_valp(ffM(iface,idim),thetaM,eps(iface,idim),4), &
                                                 interpol_valp(ffP(iface,idim),thetaP,eps(iface,idim),4))
         
                 end select
         
              else
                 am(iface,idim) = -1.
                 ap(iface,idim) =  1.
              endif
              
              if(ap(iface,idim).ne.am(iface,idim)) then
                 lm (iface,idim) = am(iface,idim)               /(ap(iface,idim)-am(iface,idim))
                 lp (iface,idim) =                ap(iface,idim)/(ap(iface,idim)-am(iface,idim))
                 lmp(iface,idim) = am(iface,idim)*ap(iface,idim)/(ap(iface,idim)-am(iface,idim))
              else !alors ap=am=0.
                 lm (iface,idim) = -1./2
                 lp (iface,idim) =  1./2
                 
                 lmp(iface,idim) =  0.
              endif
   
           enddo
           enddo

           !*************************************
           !         Chi derivatives
           !*************************************
           if (verbose)  write(*,*) '     -- Compute chi derivatives --'
   
           call cal_Dedd(Eray0(i-1,j,k),Fray0(i-1,j,k,1:ndim),Dedd_1d(:,:,1),Dedd_dE_1d(:,:,1),Dedd_dF_1d(:,:,:,1),Derive_chi )
           call cal_Dedd(Eray0(i  ,j,k),Fray0(i  ,j,k,1:ndim),Dedd          ,Dedd_dE          ,Dedd_dF            ,Derive_chi )
           call cal_Dedd(Eray0(i+1,j,k),Fray0(i+1,j,k,1:ndim),Dedd_1d(:,:,2),Dedd_dE_1d(:,:,2),Dedd_dF_1d(:,:,:,2),Derive_chi )
           if(ndim.gt.1) then
              call cal_Dedd(Eray0(i,j-1,k),Fray0(i,j-1,k,1:ndim),Dedd_2d(:,:,1),Dedd_dE_2d(:,:,1),Dedd_dF_2d(:,:,:,1),Derive_chi )
              call cal_Dedd(Eray0(i,j+1,k),Fray0(i,j+1,k,1:ndim),Dedd_2d(:,:,2),Dedd_dE_2d(:,:,2),Dedd_dF_2d(:,:,:,2),Derive_chi )
              if(ndim.gt.2) then
                 call cal_Dedd(Eray0(i,j,k-1),Fray0(i,j,k-1,1:ndim),Dedd_3d(:,:,1),Dedd_dE_3d(:,:,1),Dedd_dF_3d(:,:,:,1),Derive_chi )
                 call cal_Dedd(Eray0(i,j,k+1),Fray0(i,j,k+1,1:ndim),Dedd_3d(:,:,2),Dedd_dE_3d(:,:,2),Dedd_dF_3d(:,:,:,2),Derive_chi )
              endif
           endif


           ! Initialise geometry
           do idim = 1,ndim
              ip = i ; jp = j ; kp = k
              if(idim==1) ip = i+1
              if(idim==2) jp = j+1
              if(idim==3) kp = k+1
              dSloc(1,idim) = dS(i ,j ,k ,idim)
              dSloc(2,idim) = dS(ip,jp,kp,idim)
           enddo

           rayon = 1.
           if (geom_dir(2) == 'phi_c' .or. geom_dir(2) == 'the_s') then
              rayon(2,1:2,1:3) = xcloc(i  ,1)
              rayon(2,1  ,1  ) =  xloc(i  ,1)
              rayon(2,2  ,1  ) =  xloc(i+1,1)
           endif
   
           if (geom_dir(3) == 'phi_s' ) then
              rayon(3,1  ,1  ) =  xloc(i  ,1) * sin(xcloc(j  ,2))
              rayon(3,2  ,1  ) =  xloc(i+1,1) * sin(xcloc(j  ,2))
              rayon(3,1  ,2  ) = xcloc(i  ,1) * sin( xloc(j  ,2))
              rayon(3,2  ,2  ) = xcloc(i  ,1) * sin( xloc(j+1,2))
              rayon(3,1  ,3  ) = xcloc(i  ,1) * sin(xcloc(j  ,2))
              rayon(3,2  ,3  ) = xcloc(i  ,1) * sin(xcloc(j  ,2))
           endif

           if (verbose)  write(*,*) '     -- compute matrix --'

           mat = 0.
           do idim = 1,5
              mat(idim,idim) = 1.
           enddo

           ! Er equation without the source term
           mat(2,1) = 0.
           mat(2,2) = 1. 
           do idim = 1,ndim
              mat(2,2     ) = mat(2,2) - dt/dV(i,j,k)*c*(lmp(1,idim)*eps(1,idim)*dSloc(1,idim)+lmp(2,idim)*eps(2,idim)*dSloc(2,idim))
              mat(2,2+idim) =            dt/dV(i,j,k)  *(lm (1,idim)            *dSloc(1,idim)+lp (2,idim)            *dSloc(2,idim))
           enddo
           
           ! Tgaz equation (using the coefficients from the Er equation without the source term)
           call heat_capacity(rho(i,j,k),Tgaz0(i,j,k),conv,1,1,1,1,1,1)
           conv=1./conv

           mat(1,1  ) = 1.
           mat(1,2:5) = mat(2,2:5)* conv 
           
           ! Finally include the energy source term: rho c (kappa_em aT^4-kappa_abs E)
           
           !  T4 = Tgazold(i,j,k)**4 *(1. + 4.*(Tgaz(i,j,k)-tgazold(i,j,k))/Tgazold(i,j,k)) 
           !  T4 = -3.*Tgazold(i,j,k)**4 + 4.* Tgazold(i,j,k)**3*Tgaz(i,j,k)

           mat(2,1) =          - dt*kappa_emi_ray(i,j,k)*rho(i,j,k)*c*derive_source_ray(Tgaz0(i,j,k))
           mat(2,2) = mat(2,2) + dt*kappa_abs_ray(i,j,k)*rho(i,j,k)*c

           ! Fr equation using only dtFr + c2(Div . Pr) = 0 terms
           do idim = 1,ndim
              mat(2+idim,1     ) = 0.
              mat(2+idim,2     ) = 0.
              mat(2+idim,2+idim) = 1.
              jjmin=1
              jjmax=ndim
              jjpas=1
              if(idim.eq.1 .and. (geom_dir(1) == 'rc' .or. geom_dir(1) == 'rs')) then
                 jjmin=2
                 jdim=1
                 mat(2+idim,2     ) = mat(2+idim,2     ) + alpha*c2*dt/(xloc(i+1,1)-xloc(i,1))* &
                      ( ( lp(2,jdim)*eps(2,jdim)**2+(1.-eps(2,jdim)**2)/2) &
                      - (-lm(1,jdim)*eps(1,jdim)**2+(1.-eps(1,jdim)**2)/2) ) * Dedd(idim,jdim)
                 mat(2+idim,2+idim) = mat(2+idim,2+idim) - c*dt/(xloc(i+1,1)-xloc(i,1))* &
                      (lmp(1,jdim)*eps(1,jdim) &
                      +lmp(2,jdim)*eps(2,jdim) )  
              endif
              if(idim.eq.2 .and. geom_dir(2) == 'the_s') then
                 jjpas=2
                 jdim=1
                 mat(2+idim,2     ) = mat(2+idim,2     ) + alpha*c2*dt/(xcloc(i,1)*(xloc(j+1,2)-xloc(j,2)))* &
                      ( ( lp(2,jdim)*eps(2,jdim)**2+(1.-eps(2,jdim)**2)/2) &
                      - (-lm(1,jdim)*eps(1,jdim)**2+(1.-eps(1,jdim)**2)/2) ) * Dedd(idim,jdim)
                 mat(2+idim,2+idim) = mat(2+idim,2+idim) - c*dt/(xcloc(i,1)*(xloc(j+1,2)-xloc(j,2)))* &
                      (lmp(1,jdim)*eps(1,jdim) &
                      +lmp(2,jdim)*eps(2,jdim) )  
              endif
              do jdim = jjmin,jjmax,jjpas
                 mat(2+idim,2     ) = mat(2+idim,2     ) + alpha*c2*dt/dV(i,j,k)* &
                      ( ( lp(2,jdim)*eps(2,jdim)**2+(1.-eps(2,jdim)**2)/2)*dSloc(2,jdim)*Rayon(idim,2,jdim)/Rayon(idim,2,3) &
                      - (-lm(1,jdim)*eps(1,jdim)**2+(1.-eps(1,jdim)**2)/2)*dSloc(1,jdim)*Rayon(idim,1,jdim)/Rayon(idim,2,3) ) * Dedd(idim,jdim)
                !ed   ( ( lp(2,jdim)*eps(2,jdim)**2+(1.-eps(2,jdim)**2)/2)*dSloc(2,jdim)*Rayon(idim,2,jdim)/Rayon(idim,2,3) * DeddM(idim,jdim,2,jdim ) &
                !ed   - (-lm(1,jdim)*eps(1,jdim)**2+(1.-eps(1,jdim)**2)/2)*dSloc(1,jdim)*Rayon(idim,1,jdim)/Rayon(idim,2,3) * DeddP(idim,jdim,1,jdim ) ) 

                 mat(2+idim,2+idim) = mat(2+idim,2+idim) - c*dt/dV(i,j,k)* &
                      (lmp(1,jdim)*eps(1,jdim)*dSloc(1,jdim)*Rayon(idim,1,jdim)/Rayon(idim,2,3) &
                      +lmp(2,jdim)*eps(2,jdim)*dSloc(2,jdim)*Rayon(idim,2,jdim)/Rayon(idim,2,3) )  
              enddo
           enddo


           ! Geometric sources
           if (cylindrique) then
              iloc=3
              if(geom_dir(2) == 'phi_c') iloc=2
              mat(3,2) = mat(3,2) + c2*dt/dV(i,j,k)* (Dedd(1,1) - Dedd(iloc,iloc)) * (dSloc(2,1)-dSloc(1,1))
           endif
           if (spherique) then
                                         mat(3,2) = mat(3,2) + c2*dt/dV(i,j,k)*                                 &
                   (2.*Dedd(1,1) - Dedd(2,2) - Dedd(3,3)) * 0.5*(dSloc(2,1)-dSloc(1,1))
              if(geom_dir(2) == 'the_s') mat(4,2) = mat(4,2) + c2*dt/dV(i,j,k)* cos(xcloc(j,2))/sin(xcloc(j,2))*&
                   (               Dedd(2,2) - Dedd(3,3)) * 0.5*(dSloc(2,1)-dSloc(1,1))
           endif

           do idim=1,ndim
              ! Include the flux source term: -kappa_abs rho c F
              mat(2+idim,2+idim) = mat(2+idim,2+idim) + alpha*c*dt*(kappa_abs_ray(i,j,k)+kappa_sca_ray(i,j,k))*rho(i,j,k)
           enddo

           !*************************************************
           !     Derivative terms in the matrix
           !
           !      !!!!       CAUTION       !!!!
           ! To be done after mat(*,*)*Eray0 + mat(*,*)*Fray0
           !*************************************************
           ! Fr equation using only dtFr + c2(Div . Pr) = 0 terms
           do idim = 1,ndim
              jjmin=1
              jjmax=ndim
              jjpas=1
              if(idim.eq.1 .and. (geom_dir(1) == 'rc' .or. geom_dir(1) == 'rs')) then
                 jjmin=2
                 jdim=1
                 mat(2+idim,2     ) = mat(2+idim,2     ) + c2*dt/(xloc(i+1,1)-xloc(i,1))* &
                      ( ( lp(2,jdim)*eps(2,jdim)**2+(1.-eps(2,jdim)**2)/2) &
                      - (-lm(1,jdim)*eps(1,jdim)**2+(1.-eps(1,jdim)**2)/2) ) * Dedd_dE(idim,jdim     )  ! d/dEray
                 do kdim=1,ndim
                    mat(2+idim,2+kdim) = mat(2+idim,2+kdim) + c2*dt/(xloc(i+1,1)-xloc(i,1))* &
                         ( ( lp(2,jdim)*eps(2,jdim)**2+(1.-eps(2,jdim)**2)/2) &
                         - (-lm(1,jdim)*eps(1,jdim)**2+(1.-eps(1,jdim)**2)/2) ) * Dedd_dF(idim,jdim,kdim)  ! d/dFray
                 enddo
              endif
              if(idim.eq.2 .and. geom_dir(2) == 'the_s') then
                 jjpas=2
                 jdim=2
                 mat(2+idim,2     ) = mat(2+idim,2     ) + c2*dt/(xcloc(i,1)*(xloc(j+1,2)-xloc(j,2)))* &
                      ( ( lp(2,jdim)*eps(2,jdim)**2+(1.-eps(2,jdim)**2)/2) &
                      - (-lm(1,jdim)*eps(1,jdim)**2+(1.-eps(1,jdim)**2)/2) ) * Dedd_dE(idim,jdim     )  ! d/dEray
                 do kdim=1,ndim
                    mat(2+idim,2+kdim) = mat(2+idim,2+kdim) + c2*dt/(xcloc(i,1)*(xloc(j+1,2)-xloc(j,2)))* &
                         ( ( lp(2,jdim)*eps(2,jdim)**2+(1.-eps(2,jdim)**2)/2) &
                         - (-lm(1,jdim)*eps(1,jdim)**2+(1.-eps(1,jdim)**2)/2) ) * Dedd_dF(idim,jdim,kdim)  ! d/dFray
                 enddo
              endif
              do jdim = jjmin,jjmax,jjpas
                 ! The Dedd derivatives
                 mat(2+idim,2     ) = mat(2+idim,2     ) + c2*dt/dV(i,j,k)* &
                      ( ( lp(2,jdim)*eps(2,jdim)**2+(1.-eps(2,jdim)**2)/2)*dSloc(2,jdim)*Rayon(idim,2,jdim)/Rayon(idim,2,3) &
                      - (-lm(1,jdim)*eps(1,jdim)**2+(1.-eps(1,jdim)**2)/2)*dSloc(1,jdim)*Rayon(idim,1,jdim)/Rayon(idim,2,3) ) * Dedd_dE(idim,jdim     )  ! d/dEray
                 !ed  ( ( lp(2,jdim)*eps(2,jdim)**2+(1.-eps(2,jdim)**2)/2)*dSloc(2,jdim)*Rayon(idim,2,jdim)/Rayon(idim,2,3) * DeddM_dE(idim,jdim,2,jdim )&
                 !ed  - (-lm(1,jdim)*eps(1,jdim)**2+(1.-eps(1,jdim)**2)/2)*dSloc(1,jdim)*Rayon(idim,1,jdim)/Rayon(idim,2,3) * DeddP_dE(idim,jdim,1,jdim ) ) ! d/dEray
                 do kdim=1,ndim
                    mat(2+idim,2+kdim) = mat(2+idim,2+kdim) + c2*dt/dV(i,j,k)* &
                         ( ( lp(2,jdim)*eps(2,jdim)**2+(1.-eps(2,jdim)**2)/2)*dSloc(2,jdim)*Rayon(idim,2,jdim)/Rayon(idim,2,3) &
                         - (-lm(1,jdim)*eps(1,jdim)**2+(1.-eps(1,jdim)**2)/2)*dSloc(1,jdim)*Rayon(idim,1,jdim)/Rayon(idim,2,3) ) * Dedd_dF(idim,jdim,kdim)  ! d/dFray
                    !ed    ( ( lp(2,jdim)*eps(2,jdim)**2+(1.-eps(2,jdim)**2)/2)*dSloc(2,jdim)*Rayon(idim,2,jdim)/Rayon(idim,2,3) * DeddM_dF(idim,jdim,kdim,2,jdim )&
                    !ed    - (-lm(1,jdim)*eps(1,jdim)**2+(1.-eps(1,jdim)**2)/2)*dSloc(1,jdim)*Rayon(idim,1,jdim)/Rayon(idim,2,3) * DeddP_dF(idim,jdim,kdim,1,jdim ) )  ! d/dFray
                 enddo
              enddo
           enddo

        
           ! Geometric sources
           if(cylindrique) then
              iloc=3
              if(geom_dir(2) == 'phi_c') iloc=2
              mat(3,2     ) = mat(3,2     ) + c2*dt/dV(i,j,k)* (Dedd_dE(1,1) - Dedd_dE(iloc,iloc)) * (dSloc(2,1)-dSloc(1,1))
              do idim=1,ndim
                 mat(3,2+idim) = mat(3,2+idim) + c2*dt/dV(i,j,k)* (Dedd_dF(1,1,idim) - Dedd_dF(iloc,iloc,idim)) * (dSloc(2,1)-dSloc(1,1))
              enddo
           endif
           if (spherique) then
              mat(3,2     ) = mat(3,2     ) + c2*dt/dV(i,j,k)* (2.*Dedd_dE(1,1     ) - Dedd_dE(2,2     ) - Dedd_dE(3,3     )) * 0.5*(dSloc(2,1)-dSloc(1,1))
              do idim=1,ndim
                 mat(3,2+idim) = mat(3,2+idim) + c2*dt/dV(i,j,k)* (2.*Dedd_dF(1,1,idim) - Dedd_dF(2,2,idim) - Dedd_dF(3,3,idim)) * 0.5*(dSloc(2,1)-dSloc(1,1))
              enddo
              if(geom_dir(2) == 'the_s') then
                 mat(4,2     ) = mat(4,2     ) + c2*dt/dV(i,j,k)* cos(xcloc(j,2))/sin(xcloc(j,2))*(Dedd_dE(2,2     ) - Dedd_dE(3,3     )) * 0.5*(dSloc(2,1)-dSloc(1,1))
                 do idim=1,ndim
                    mat(4,2+idim) = mat(4,2+idim) + c2*dt/dV(i,j,k)* cos(xcloc(j,2))/sin(xcloc(j,2))*(Dedd_dF(2,2,idim) - Dedd_dF(3,3,idim)) * 0.5*(dSloc(2,1)-dSloc(1,1))
                 enddo
              endif
           endif
           

           ! non-diagonal terms
           
           a1d = 0. ; a2d = 0. ; a3d = 0.

           ! Temperature equation
           a1d(1,2)  =  dt/dV(i,j,k)*c*lmp(1,1)*eps(1,1)*conv*dSloc(1,1) !*ErayM(1)   
           a1d(1,3)  = -dt/dV(i,j,k)*   lp(1,1)         *conv*dSloc(1,1) !*FrayM(1,1) 

           ! Radiative energy equation
           a1d(2,2)  =  dt/dV(i,j,k)*c*lmp(1,1)*eps(1,1)*dSloc(1,1)!*ErayM(1)                
           a1d(2,3)  = -dt/dV(i,j,k)*   lp(1,1)         *dSloc(1,1)!*FrayM(1,1)

           ! Radiative flux equation
           do idim=1,ndim
              if(idim.eq.1 .and. (geom_dir(1) == 'rc' .or. geom_dir(1) == 'rs')) then
                 a1d(2+idim,2) = -dt/(xloc(i+1,1)-xloc(i,1))*   c2*( lp(1,1)*eps(1,1)**2+(1.-eps(1,1)**2)/2 ) * (DeddM(idim,1,1,1)+Dedd_dE_1d(idim,1,1)) *Rayon(idim,1,1)/Rayon(idim,2,3)
                 do jdim=1,ndim
                    a1d(2+idim,2+jdim) = -dt/(xloc(i+1,1)-xloc(i,1)) *   c2*( lp(1,1)*eps(1,1)**2+(1.-eps(1,1)**2)/2 ) *  Dedd_dF_1d(idim,1,jdim,1)      *Rayon(idim,1,1)/Rayon(idim,2,3)
                 enddo
                 a1d(2+idim,2+idim) = a1d(2+idim,2+idim) + dt/(xloc(i+1,1)-xloc(i,1))*c*lmp(1,1)*eps(1,1)                                                *Rayon(idim,1,1)/Rayon(idim,2,3)
              else
                 a1d(2+idim,2) = -dt/dV(i,j,k)*   c2*( lp(1,1)*eps(1,1)**2+(1.-eps(1,1)**2)/2 ) * (DeddM(idim,1,1,1)+Dedd_dE_1d(idim,1,1))*dSloc(1,1) *Rayon(idim,1,1)/Rayon(idim,2,3)
                 do jdim=1,ndim
                    a1d(2+idim,2+jdim) = -dt/dV(i,j,k)*   c2*( lp(1,1)*eps(1,1)**2+(1.-eps(1,1)**2)/2 ) *  Dedd_dF_1d(idim,1,jdim,1)*dSloc(1,1)       *Rayon(idim,1,1)/Rayon(idim,2,3)
                 enddo
                 a1d(2+idim,2+idim) = a1d(2+idim,2+idim) + dt/dV(i,j,k)*c*lmp(1,1)*eps(1,1)*dSloc(1,1)                                                *Rayon(idim,1,1)/Rayon(idim,2,3)
              endif
           enddo

           ! Temperature equation
           a1d(1,nvar_ray+2)  =  dt/dV(i,j,k)*c*lmp(2,1)*eps(2,1)*conv*dSloc(2,1) !ErayP(2)   
           a1d(1,nvar_ray+3)  = -dt/dV(i,j,k)*   lm(2,1)         *conv*dSloc(2,1) !FrayP(1,2) 

           ! Radiative energy equation
           a1d(2,nvar_ray+2)  =  dt/dV(i,j,k)*c*lmp(2,1)*eps(2,1)*dSloc(2,1) !*ErayP(2)
           a1d(2,nvar_ray+3)  = -dt/dV(i,j,k)*   lm(2,1)         *dSloc(2,1) !*FrayP(1,2)

           ! Radiative flux equation
           do idim=1,ndim
              if(idim.eq.1 .and. (geom_dir(1) == 'rc' .or. geom_dir(1) == 'rs')) then
                 a1d(2+idim,nvar_ray+2) = dt/(xloc(i+1,1)-xloc(i,1))*   c2*(-lm(2,1)*eps(2,1)**2+(1.-eps(2,1)**2)/2 ) * (DeddP(idim,1,2,1)+Dedd_dE_1d(idim,1,2)) *Rayon(idim,2,1)/Rayon(idim,2,3)
                 do jdim=1,ndim
                    a1d(2+idim,nvar_ray+2+jdim) = dt/(xloc(i+1,1)-xloc(i,1))* c2*(-lm(2,1)*eps(2,1)**2+(1.-eps(2,1)**2)/2 ) * Dedd_dF_1d(idim,1,jdim,2)          *Rayon(idim,2,1)/Rayon(idim,2,3)
                 enddo
                 a1d(2+idim,nvar_ray+2+idim) = a1d(2+idim,nvar_ray+2+idim) + dt/(xloc(i+1,1)-xloc(i,1))*c* lmp(2,1)*eps(2,1)                                     *Rayon(idim,2,1)/Rayon(idim,2,3)
              else
                 a1d(2+idim,nvar_ray+2) = dt/dV(i,j,k)*   c2*(-lm(2,1)*eps(2,1)**2+(1.-eps(2,1)**2)/2 ) * (DeddP(idim,1,2,1)+Dedd_dE_1d(idim,1,2))*dSloc(2,1) *Rayon(idim,2,1)/Rayon(idim,2,3)
                 do jdim=1,ndim
                    a1d(2+idim,nvar_ray+2+jdim) = dt/dV(i,j,k)* c2*(-lm(2,1)*eps(2,1)**2+(1.-eps(2,1)**2)/2 ) * Dedd_dF_1d(idim,1,jdim,2)*dSloc(2,1)          *Rayon(idim,2,1)/Rayon(idim,2,3)
                 enddo
                 a1d(2+idim,nvar_ray+2+idim) = a1d(2+idim,nvar_ray+2+idim) + dt/dV(i,j,k)*c* lmp(2,1)*eps(2,1)*dSloc(2,1)                                     *Rayon(idim,2,1)/Rayon(idim,2,3)
              endif
           enddo

           if(ndim.gt.1) then
              ! coupling with j-1 terms
              a2d(1,1) = 0.
              a2d(1,2) =  dt/dV(i,j,k)*c*lmp(1,2)*eps(1,2)*conv*dSloc(1,2)
              a2d(1,3) = 0.
              a2d(1,4) = -dt/dV(i,j,k)*   lp(1,2)         *conv*dSloc(1,2)

              a2d(2,1) = 0.
              a2d(2,2) =  dt/dV(i,j,k)*c*lmp(1,2)*eps(1,2)*dSloc(1,2)
              a2d(2,3) = 0.
              a2d(2,4) = -dt/dV(i,j,k)*   lp(1,2)         *dSloc(1,2)

              do idim=1,ndim
                 if(idim.eq.2 .and. geom_dir(2)=='the_s') then
                    a2d(2+idim,2) = -dt/(xcloc(i,1)*(xloc(j+1,2)-xloc(j,2)))*   c2*( lp(1,2)*eps(1,2)**2+(1.-eps(1,2)**2)/2 ) * (DeddM(idim,2,1,2)+Dedd_dE_2d(idim,2,1))  *Rayon(idim,1,2)/Rayon(idim,2,3)
                    do jdim=1,ndim
                       a2d(2+idim,2+jdim) = -dt/(xcloc(i,1)*(xloc(j+1,2)-xloc(j,2)))*   c2*( lp(1,2)*eps(1,2)**2+(1.-eps(1,2)**2)/2 ) * Dedd_dF_2d(idim,2,jdim,1)         *Rayon(idim,1,2)/Rayon(idim,2,3)
                    enddo
                    a2d(2+idim,2+idim) = a2d(2+idim,2+idim) + dt/(xcloc(i,1)*(xloc(j+1,2)-xloc(j,2)))*c* lmp(1,2)*eps(1,2)                                                *Rayon(idim,1,2)/Rayon(idim,2,3)
                 else
                    a2d(2+idim,2) = -dt/dV(i,j,k)*   c2*( lp(1,2)*eps(1,2)**2+(1.-eps(1,2)**2)/2 ) * (DeddM(idim,2,1,2)+Dedd_dE_2d(idim,2,1))*dSloc(1,2)      *Rayon(idim,1,2)/Rayon(idim,2,3)
                    do jdim=1,ndim
                       a2d(2+idim,2+jdim) = -dt/dV(i,j,k)*   c2*( lp(1,2)*eps(1,2)**2+(1.-eps(1,2)**2)/2 ) * Dedd_dF_2d(idim,2,jdim,1)*dSloc(1,2)             *Rayon(idim,1,2)/Rayon(idim,2,3)
                    enddo
                    a2d(2+idim,2+idim) = a2d(2+idim,2+idim) + dt/dV(i,j,k)*c* lmp(1,2)*eps(1,2)*dSloc(1,2)                                                    *Rayon(idim,1,2)/Rayon(idim,2,3)
                 endif
              enddo

              ! coupling with j+1 terms
              a2d(1,nvar_ray+1) = 0.
              a2d(1,nvar_ray+2) =  dt/dV(i,j,k)*c*lmp(2,2)*eps(2,2)*conv*dSloc(2,2)
              a2d(1,nvar_ray+3) = 0.
              a2d(1,nvar_ray+4) = -dt/dV(i,j,k)*   lm(2,2)         *conv*dSloc(2,2)

              a2d(2,nvar_ray+1) = 0.
              a2d(2,nvar_ray+2) =  dt/dV(i,j,k)*c*lmp(2,2)*eps(2,2)*dSloc(2,2)
              a2d(2,nvar_ray+3) = 0.
              a2d(2,nvar_ray+4) = -dt/dV(i,j,k)*   lm(2,2)         *dSloc(2,2)

              do idim=1,ndim
                 if(idim.eq.2 .and. geom_dir(2)=='the_s') then
                    a2d(2+idim,nvar_ray+2) = dt/(xcloc(i,1)*(xloc(j+1,2)-xloc(j,2)))*   c2*(-lm(2,2)*eps(2,2)**2+(1.-eps(2,2)**2)/2 ) * (DeddP(idim,2,2,2)+Dedd_dE_2d(idim,2,2))*Rayon(idim,2,2)/Rayon(idim,2,3)
                    do jdim=1,ndim
                       a2d(2+idim,nvar_ray+2+jdim) = dt/(xcloc(i,1)*(xloc(j+1,2)-xloc(j,2)))*   c2*(-lm(2,2)*eps(2,2)**2+(1.-eps(2,2)**2)/2 ) * Dedd_dF_2d(idim,2,jdim,2)       *Rayon(idim,2,2)/Rayon(idim,2,3)
                    enddo
                    a2d(2+idim,nvar_ray+2+idim) = a2d(2+idim,nvar_ray+2+idim) + dt/(xcloc(i,1)*(xloc(j+1,2)-xloc(j,2)))*c* lmp(2,2)*eps(2,2)                                    *Rayon(idim,2,2)/Rayon(idim,2,3)
                 else
                    a2d(2+idim,nvar_ray+2) = dt/dV(i,j,k)*   c2*(-lm(2,2)*eps(2,2)**2+(1.-eps(2,2)**2)/2 ) * (DeddP(idim,2,2,2)+Dedd_dE_2d(idim,2,2))*dSloc(2,2)  *Rayon(idim,2,2)/Rayon(idim,2,3)
                    do jdim=1,ndim
                       a2d(2+idim,nvar_ray+2+jdim) = dt/dV(i,j,k)*   c2*(-lm(2,2)*eps(2,2)**2+(1.-eps(2,2)**2)/2 ) * Dedd_dF_2d(idim,2,jdim,2)*dSloc(2,2)         *Rayon(idim,2,2)/Rayon(idim,2,3)
                    enddo
                    a2d(2+idim,nvar_ray+2+idim) = a2d(2+idim,nvar_ray+2+idim) + dt/dV(i,j,k)*c* lmp(2,2)*eps(2,2)*dSloc(2,2)                                      *Rayon(idim,2,2)/Rayon(idim,2,3)
                 endif
              enddo

              if(ndim.gt.2) then
                 ! coupling with k-1 terms
                 a3d(1,1) = 0.
                 a3d(1,2) =  dt/dV(i,j,k)*c*lmp(1,3)*eps(1,3)*conv*dSloc(1,3)
                 a3d(1,3) = 0.
                 a3d(1,4) = 0.
                 a3d(1,5) = -dt/dV(i,j,k)*   lp(1,3)         *conv*dSloc(1,3)

                 a3d(2,1) = 0.
                 a3d(2,2) =  dt/dV(i,j,k)*c*lmp(1,3)*eps(1,3)*dSloc(1,3)
                 a3d(2,3) = 0.
                 a3d(2,4) = 0.
                 a3d(2,5) = -dt/dV(i,j,k)  * lp(1,3)         *dSloc(1,3)

                 do idim=1,ndim
                    a3d(2+idim,2) = -dt/dV(i,j,k)*   c2*( lp(1,3)*eps(1,3)**2+(1.-eps(1,3)**2)/2 ) * (DeddM(idim,3,1,3)+Dedd_dE_3d(idim,3,1))*dSloc(1,3)      *Rayon(idim,1,3)/Rayon(idim,2,3)
                    do jdim=1,ndim
                       a3d(2+idim,2+jdim) =  -dt/dV(i,j,k)*   c2*( lp(1,3)*eps(1,3)**2+(1.-eps(1,3)**2)/2 ) * Dedd_dF_3d(idim,3,jdim,1)*dSloc(1,3)            *Rayon(idim,1,3)/Rayon(idim,2,3)
                    enddo
                    a3d(2+idim,2+idim) = a3d(2+idim,2+idim) + dt/dV(i,j,k)*c * lmp(1,3)*eps(1,3)*dSloc(1,3)                                                   *Rayon(idim,1,3)/Rayon(idim,2,3)
                 enddo

                 ! coupling with k+1 terms
                 a3d(1,nvar_ray+1) = 0.
                 a3d(1,nvar_ray+2) =  dt/dV(i,j,k)*c*lmp(2,3)*eps(2,3)*conv*dSloc(2,3)
                 a3d(1,nvar_ray+3) = 0.
                 a3d(1,nvar_ray+4) = 0.
                 a3d(1,nvar_ray+5) = -dt/dV(i,j,k)  * lm(2,3)         *conv*dSloc(2,3)

                 a3d(2,nvar_ray+1) = 0.
                 a3d(2,nvar_ray+2) =  dt/dV(i,j,k)*c*lmp(2,3)*eps(2,3)*dSloc(2,3)
                 a3d(2,nvar_ray+3) = 0.
                 a3d(2,nvar_ray+4) = 0.
                 a3d(2,nvar_ray+5) = -dt/dV(i,j,k)  * lm(2,3)         *dSloc(2,3)

                 do idim=1,ndim
                    a3d(2+idim,nvar_ray+2) =  dt/dV(i,j,k)*   c2*(-lm(2,3)*eps(2,3)**2+(1.-eps(2,3)**2)/2 )  *(DeddP(idim,3,2,3)+Dedd_dE_3d(idim,3,2))*dSloc(2,3) *Rayon(idim,2,3)/Rayon(idim,2,3)
                    do jdim=1,ndim
                       a3d(2+idim,nvar_ray+2+jdim) = dt/dV(i,j,k)*   c2*(-lm(2,3)*eps(2,3)**2+(1.-eps(2,3)**2)/2 )  *Dedd_dF_3d(idim,3,jdim,2)*dSloc(2,3)         *Rayon(idim,2,3)/Rayon(idim,2,3)
                    enddo
                    a3d(2+idim,nvar_ray+2+idim) = a3d(2+idim,nvar_ray+2+idim) + dt/dV(i,j,k)*c * lmp(2,3)*eps(2,3)*dSloc(2,3)                                     *Rayon(idim,2,3)/Rayon(idim,2,3)
                 enddo
              endif
           endif

           !***********************************************************
           !              Boundary conditions
           !***********************************************************
           if(i==1) then
              if(icl_ray(1)==1) then
                 mat(1:nvar_ray,1:nvar_ray) = mat(1:nvar_ray,1:nvar_ray) +     a1d(1:nvar_ray,     1:     nvar_ray)
              else if(icl_ray(1)==3) then
                 mat(1:nvar_ray,1:nvar_ray) = mat(1:nvar_ray,1:nvar_ray) +     a1d(1:nvar_ray,     1:     nvar_ray)
                 mat(1:nvar_ray,3     ) = mat(1:nvar_ray,3     ) - 2.* a1d(1:nvar_ray,     3          )
              endif
              if(icl_ray(1).ne.2 .and. icl_ray(1).ne.0) a1d(1:nvar_ray,1:nvar_ray) = 0.
           else if(i==nx(1)) then
              if(icl_ray(2)==1) then
                 mat(1:nvar_ray,1:nvar_ray) = mat(1:nvar_ray,1:nvar_ray) +     a1d(1:nvar_ray,nvar_ray+1:nvar_ray+nvar_ray)
              else if(icl_ray(2)==3) then
                 mat(1:nvar_ray,1:nvar_ray) = mat(1:nvar_ray,1:nvar_ray) +     a1d(1:nvar_ray,nvar_ray+1:nvar_ray+nvar_ray)
                 mat(1:nvar_ray,3     ) = mat(1:nvar_ray,3     ) - 2.* a1d(1:nvar_ray,nvar_ray+3          )
              endif
              if(icl_ray(2).ne.2 .and. icl_ray(2).ne.0) a1d(1:nvar_ray,nvar_ray+1:nvar_ray+nvar_ray) = 0.
           endif
           if(ndim.gt.1) then
              if(j==1) then
                 if(icl_ray(3)==1) then
                    mat(1:nvar_ray,1:nvar_ray) = mat(1:nvar_ray,1:nvar_ray) +     a2d(1:nvar_ray,     1:     nvar_ray)
                 else if(icl_ray(3)==3) then
                    mat(1:nvar_ray,1:nvar_ray) = mat(1:nvar_ray,1:nvar_ray) +     a2d(1:nvar_ray,     1:     nvar_ray)
                    mat(1:nvar_ray,4     ) = mat(1:nvar_ray,4     ) - 2.* a2d(1:nvar_ray,     4          )
                 endif
                 if(icl_ray(3).ne.2 .and. icl_ray(3).ne.0) a2d(1:nvar_ray,1:nvar_ray) = 0.
              else if(j==nx(2)) then
                 if(icl_ray(4)==1) then
                    mat(1:nvar_ray,1:nvar_ray) = mat(1:nvar_ray,1:nvar_ray) +     a2d(1:nvar_ray,nvar_ray+1:nvar_ray+nvar_ray)
                 else if(icl_ray(4)==3) then
                    mat(1:nvar_ray,1:nvar_ray) = mat(1:nvar_ray,1:nvar_ray) +     a2d(1:nvar_ray,nvar_ray+1:nvar_ray+nvar_ray)
                    mat(1:nvar_ray,4     ) = mat(1:nvar_ray,4     ) - 2.* a2d(1:nvar_ray,nvar_ray+4          )
                 endif
                 if(icl_ray(4).ne.2 .and. icl_ray(4).ne.0) a2d(1:nvar_ray,nvar_ray+1:nvar_ray+nvar_ray) = 0.
              endif
           endif
           if(ndim.gt.2) then
              if(k==1) then
                 if(icl_ray(5)==1) then
                    mat(1:nvar_ray,1:nvar_ray) = mat(1:nvar_ray,1:nvar_ray) +     a3d(1:nvar_ray,     1:     nvar_ray)
                 else if(icl_ray(5)==3) then
                    mat(1:nvar_ray,1:nvar_ray) = mat(1:nvar_ray,1:nvar_ray) +     a3d(1:nvar_ray,     1:     nvar_ray)
                    mat(1:nvar_ray,5     ) = mat(1:nvar_ray,5     ) - 2.* a3d(1:nvar_ray,     5          )
                 endif
                 if(icl_ray(5).ne.2 .and. icl_ray(5).ne.0) a3d(1:nvar_ray,1:nvar_ray) = 0.
              else if(k==nx(3)) then
                 if(icl_ray(6)==1) then
                    mat(1:nvar_ray,1:nvar_ray) = mat(1:nvar_ray,1:nvar_ray) +     a3d(1:nvar_ray,nvar_ray+1:nvar_ray+nvar_ray)
                 else if(icl_ray(6)==3) then
                    mat(1:nvar_ray,1:nvar_ray) = mat(1:nvar_ray,1:nvar_ray) +     a3d(1:nvar_ray,nvar_ray+1:nvar_ray+nvar_ray)
                    mat(1:nvar_ray,5     ) = mat(1:nvar_ray,5     ) - 2.* a3d(1:nvar_ray,nvar_ray+5          )
                 endif
                 if(icl_ray(6).ne.2 .and. icl_ray(6).ne.0) a3d(1:nvar_ray,nvar_ray+1:nvar_ray+nvar_ray) = 0.
              endif
           endif


           !**************  Normalised equations   ****************!
           !*******************************************************!
           !**        eqi / Xoldi et Aij=dFi/dXj *Xoldj          **!
           !**       to get     Aij (deltaX/Xold)j = eqi         **!
           !*******************************************************!
           !*******************************************************!
           mat(1  ,1:5) = mat(1  ,1:5)/   Tgazold(i,j,k)
           mat(2  ,1:5) = mat(2  ,1:5)/   Erayold(i,j,k)
           mat(3:5,1:5) = mat(3:5,1:5)/(c*Erayold(i,j,k))

           mat(1:5,1  ) = mat(1:5,1  ) *   Tgazold(i,j,k)
           mat(1:5,2  ) = mat(1:5,2  ) *   Erayold(i,j,k)
           mat(1:5,3:5) = mat(1:5,3:5) * c*Erayold(i,j,k)

           a1d(1     ,1:2*nvar_ray)=a1d(1     ,1:2*nvar_ray)/   Tgazold(i,j,k)
           a1d(2     ,1:2*nvar_ray)=a1d(2     ,1:2*nvar_ray)/   Erayold(i,j,k)
           a1d(3:nvar_ray,1:2*nvar_ray)=a1d(3:nvar_ray,1:2*nvar_ray)/(c*Erayold(i,j,k))

           do i1=1,2
              a1d(1:nvar_ray,nvar_ray*(i1-1)+1                 ) = a1d(1:nvar_ray,nvar_ray*(i1-1)+1                 ) *   Tgazold(i+2*i1-3,j,k) 
              a1d(1:nvar_ray,nvar_ray*(i1-1)+2                 ) = a1d(1:nvar_ray,nvar_ray*(i1-1)+2                 ) *   Erayold(i+2*i1-3,j,k)
              a1d(1:nvar_ray,nvar_ray*(i1-1)+3:nvar_ray*(i1-1)+nvar_ray) = a1d(1:nvar_ray,nvar_ray*(i1-1)+3:nvar_ray*(i1-1)+nvar_ray) * c*Erayold(i+2*i1-3,j,k)
           enddo

           if(ndim.gt.1) then
              a2d(1     ,1:2*nvar_ray)=a2d(1     ,1:2*nvar_ray)/   Tgazold(i,j,k)
              a2d(2     ,1:2*nvar_ray)=a2d(2     ,1:2*nvar_ray)/   Erayold(i,j,k)
              a2d(3:nvar_ray,1:2*nvar_ray)=a2d(3:nvar_ray,1:2*nvar_ray)/(c*Erayold(i,j,k))

              do i1=1,2
                 a2d(1:nvar_ray,nvar_ray*(i1-1)+1                 ) = a2d(1:nvar_ray,nvar_ray*(i1-1)+1                 ) *   Tgazold(i,j+2*i1-3,k) 
                 a2d(1:nvar_ray,nvar_ray*(i1-1)+2                 ) = a2d(1:nvar_ray,nvar_ray*(i1-1)+2                 ) *   Erayold(i,j+2*i1-3,k)
                 a2d(1:nvar_ray,nvar_ray*(i1-1)+3:nvar_ray*(i1-1)+nvar_ray) = a2d(1:nvar_ray,nvar_ray*(i1-1)+3:nvar_ray*(i1-1)+nvar_ray) * c*Erayold(i,j+2*i1-3,k)
              enddo
              
              if(ndim.gt.2) then
                 a3d(1     ,1:2*nvar_ray)=a3d(1     ,1:2*nvar_ray)/   Tgazold(i,j,k)
                 a3d(2     ,1:2*nvar_ray)=a3d(2     ,1:2*nvar_ray)/   Erayold(i,j,k)
                 a3d(3:nvar_ray,1:2*nvar_ray)=a3d(3:nvar_ray,1:2*nvar_ray)/(c*Erayold(i,j,k))
                 
                 do i1=1,2
                    a3d(1:nvar_ray,nvar_ray*(i1-1)+1                 ) = a3d(1:nvar_ray,nvar_ray*(i1-1)+1                 ) *   Tgazold(i,j,k+2*i1-3) 
                    a3d(1:nvar_ray,nvar_ray*(i1-1)+2                 ) = a3d(1:nvar_ray,nvar_ray*(i1-1)+2                 ) *   Erayold(i,j,k+2*i1-3)
                    a3d(1:nvar_ray,nvar_ray*(i1-1)+3:nvar_ray*(i1-1)+nvar_ray) = a3d(1:nvar_ray,nvar_ray*(i1-1)+3:nvar_ray*(i1-1)+nvar_ray) * c*Erayold(i,j,k+2*i1-3)
                 enddo
              endif
           endif

           if(precond_ray) then
              call essai_inv5x5(mat,inv)
              mat (1:nvar_ray,     1:     nvar_ray) = matmul( inv(1:nvar_ray,1:nvar_ray),mat (1:nvar_ray,     1:     nvar_ray) )
              a1d (1:nvar_ray,     1:     nvar_ray) = matmul( inv(1:nvar_ray,1:nvar_ray),a1d (1:nvar_ray,     1:     nvar_ray) )
              a1d (1:nvar_ray,nvar_ray+1:nvar_ray+nvar_ray) = matmul( inv(1:nvar_ray,1:nvar_ray),a1d (1:nvar_ray,nvar_ray+1:nvar_ray+nvar_ray) )
              if(ndim.gt.1) then
                 a2d (1:nvar_ray,     1:     nvar_ray) = matmul( inv(1:nvar_ray,1:nvar_ray),a2d (1:nvar_ray,     1:     nvar_ray) )
                 a2d (1:nvar_ray,nvar_ray+1:nvar_ray+nvar_ray) = matmul( inv(1:nvar_ray,1:nvar_ray),a2d (1:nvar_ray,nvar_ray+1:nvar_ray+nvar_ray) )
                 if(ndim.gt.2) then
                    a3d (1:nvar_ray,     1:     nvar_ray) = matmul( inv(1:nvar_ray,1:nvar_ray),a3d (1:nvar_ray,     1:     nvar_ray) )
                    a3d (1:nvar_ray,nvar_ray+1:nvar_ray+nvar_ray) = matmul( inv(1:nvar_ray,1:nvar_ray),a3d (1:nvar_ray,nvar_ray+1:nvar_ray+nvar_ray) )
                 endif
              endif
           endif
      
           do ivar=1,nvar_ray
              xout((i+(j-1)*nx(1)+(k-1)*nx(1)*nx(2)-1)*nvar_ray+ivar) = 0.
              do jvar=1,nvar_ray
                 xout((i+(j-1)*nx(1)+(k-1)*nx(1)*nx(2)-1)*nvar_ray+ivar) = xout((i+(j-1)*nx(1)+(k-1)*nx(1)*nx(2)-1)*nvar_ray+ivar) &
                      &       + a1d(ivar,     jvar)*xin((i-1-imin+(j-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+jvar) &
                      &       + mat(ivar,jvar     )*xin((i  -imin+(j-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+jvar) &
                      &       + a1d(ivar,nvar_ray+jvar)*xin((i+1-imin+(j-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+jvar)
              enddo
           enddo
      
           if(ndim.gt.1) then
              do ivar=1,nvar_ray
                 do jvar=1,nvar_ray
                    xout((i+(j-1)*nx(1)+(k-1)*nx(1)*nx(2)-1)*nvar_ray+ivar) = xout((i+(j-1)*nx(1)+(k-1)*nx(1)*nx(2)-1)*nvar_ray+ivar) &
                         &       + a2d(ivar,     jvar)*xin((i-imin+(j-1-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+jvar) &
                         &       + a2d(ivar,nvar_ray+jvar)*xin((i-imin+(j+1-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+jvar)
                 enddo
              enddo
           endif
           if(ndim.gt.2) then
              do ivar=1,nvar_ray
                 do jvar=1,nvar_ray
                    xout((i+(j-1)*nx(1)+(k-1)*nx(1)*nx(2)-1)*nvar_ray+ivar) = xout((i+(j-1)*nx(1)+(k-1)*nx(1)*nx(2)-1)*nvar_ray+ivar) &
                         &       + a3d(ivar,     jvar)*xin((i-imin+(j-jmin)*(imax-imin+1)+(k-1-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+jvar) &
                         &       + a3d(ivar,nvar_ray+jvar)*xin((i-imin+(j-jmin)*(imax-imin+1)+(k+1-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+jvar)
                 enddo
              enddo
           endif

        enddo
     enddo
  enddo ! End of i,j,k loop


  ! Array deallocation
  deallocate(dx,dxc)
  deallocate(ds,dv)
  deallocate(xloc,xcloc)

  if (verbose) write(*,*) 'End of mat_prod3d'

end subroutine mat_prod3d
!###########################################################
!###########################################################
!###########################################################
!###########################################################

#if PARA==1

!  Subroutine COM_1D_XIN
!
!> Communications 1D for the xin vector.
!<
subroutine com_1D_xin(xin)

  use communication
  use parameters
  use para
  use tabgmres     , only : imin,imax,jmin,jmax,kmin,kmax

  implicit none

  include 'mpif.h'

  real, dimension(-nvar_ray+1:(imax-imin+(jmax-jmin)*(imax-imin+1)+(kmax-kmin)*(imax-imin+1)*(jmax-jmin+1))*nvar_ray) :: xin

  integer, parameter                    :: LEFT=1,RIGHT=2
  integer                               :: icount,ierr
  integer                               :: destination, emetteur
  integer, dimension(4  )               :: Req
  integer, dimension(MPI_STATUS_SIZE,4) :: Stat

  integer :: nnx

  nnx = nx(1)

  icount   = nvar_ray*2
  Req(1:4) = MPI_REQUEST_NULL

  face1_e(1,1,1,1) = xin(1) 
  face1_e(1,1,1,2) = xin(2) 
  face1_e(1,1,1,3) = xin(3) 

  face2_e(1,1,1,1) = xin((nnx-1)*nvar_ray+1) 
  face2_e(1,1,1,2) = xin((nnx-1)*nvar_ray+2) 
  face2_e(1,1,1,3) = xin((nnx-1)*nvar_ray+3) 

  ! send data

  destination = voisin(-1,0,0)
  call MPI_Isend(face1_e,icount,MPI_DOUBLE_PRECISION,destination,LEFT ,MPI_Comm_world,Req(1),ierr)

  destination = voisin(+1,0,0)
  call MPI_Isend(face2_e,icount,MPI_DOUBLE_PRECISION,destination,RIGHT,MPI_COMM_WORLD,Req(2),ierr)

  ! receive data
  
  emetteur = voisin(+1,0,0)
  call MPI_Irecv(face2_r,icount,MPI_DOUBLE_PRECISION,emetteur   ,LEFT ,MPI_COMM_WORLD,Req(3),ierr)
  
  emetteur = voisin(-1,0,0)
  call MPI_Irecv(face1_r,icount,MPI_DOUBLE_PRECISION,emetteur   ,RIGHT,MPI_COMM_WORLD,Req(4),ierr)
  
  call MPI_Waitall(4,Req,Stat,ierr)

  xin(-nvar_ray+1) = face1_r(1,1,1,1)
  xin(-nvar_ray+2) = face1_r(1,1,1,2)
  xin(-nvar_ray+3) = face1_r(1,1,1,3)
  
  xin( nnx   *nvar_ray+1) = face2_r(1,1,1,1)
  xin( nnx   *nvar_ray+2) = face2_r(1,1,1,2)
  xin( nnx   *nvar_ray+3) = face2_r(1,1,1,3)

  return
  
end subroutine com_1D_xin

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine COM_2D_XIN
!
!> Communications 2D for the xin vector.
!<
subroutine com_2D_xin(xin)

  use communication
  use parameters
  use para
  use tabgmres     ,only : imin,imax,jmin,jmax,kmin,kmax

  implicit none

  include 'mpif.h'

  real, dimension(-nvar_ray+1:(imax-imin+(jmax-jmin)*(imax-imin+1)+(kmax-kmin)*(imax-imin+1)*(jmax-jmin+1))*nvar_ray) :: xin

  integer, parameter                     :: LEFT=1,RIGHT=2,DOWN=3, UP=4, RD=13, RU=14, LD=23, LU = 24
  integer                                :: i,j,ierr
  integer                                :: ic_f12,ic_f34,ic_coin
  integer                                :: destination, emetteur
  integer, dimension(16  )               :: Req
  integer, dimension(MPI_STATUS_SIZE,16) :: Stat

  integer :: nnx, nny

  nnx = nx(1) ; nny = nx(2)

  ic_f12  = nvar_ray*nny*2
  ic_f34  = nvar_ray*nnx*2
  ic_coin = nvar_ray*2*2

  Req(1:16) = MPI_REQUEST_NULL

  ! 4 faces
  do j=1,nny
     face1_e(1,j,1,1) = xin((1-imin+(j-jmin)*(imax-imin+1)-1)*nvar_ray+1)
     face1_e(1,j,1,2) = xin((1-imin+(j-jmin)*(imax-imin+1)-1)*nvar_ray+2)
     face1_e(1,j,1,3) = xin((1-imin+(j-jmin)*(imax-imin+1)-1)*nvar_ray+3)
     face1_e(1,j,1,4) = xin((1-imin+(j-jmin)*(imax-imin+1)-1)*nvar_ray+4)
     
     face2_e(1,j,1,1) = xin((nnx-imin+(j-jmin)*(imax-imin+1)-1)*nvar_ray+1)
     face2_e(1,j,1,2) = xin((nnx-imin+(j-jmin)*(imax-imin+1)-1)*nvar_ray+2)
     face2_e(1,j,1,3) = xin((nnx-imin+(j-jmin)*(imax-imin+1)-1)*nvar_ray+3)
     face2_e(1,j,1,4) = xin((nnx-imin+(j-jmin)*(imax-imin+1)-1)*nvar_ray+4)
  enddo
  do i=1,nnx
     face3_e(i,1,1,1) = xin((i-imin+(1-jmin)*(imax-imin+1)-1)*nvar_ray+1)
     face3_e(i,1,1,2) = xin((i-imin+(1-jmin)*(imax-imin+1)-1)*nvar_ray+2)
     face3_e(i,1,1,3) = xin((i-imin+(1-jmin)*(imax-imin+1)-1)*nvar_ray+3)
     face3_e(i,1,1,4) = xin((i-imin+(1-jmin)*(imax-imin+1)-1)*nvar_ray+4)
     
     face4_e(i,1,1,1) = xin((i-imin+(nny-jmin)*(imax-imin+1)-1)*nvar_ray+1)
     face4_e(i,1,1,2) = xin((i-imin+(nny-jmin)*(imax-imin+1)-1)*nvar_ray+2)
     face4_e(i,1,1,3) = xin((i-imin+(nny-jmin)*(imax-imin+1)-1)*nvar_ray+3)
     face4_e(i,1,1,4) = xin((i-imin+(nny-jmin)*(imax-imin+1)-1)*nvar_ray+4)
  enddo

  ! 4 corners
  coin13_e(1,1,1,1) = xin((1-imin+(1-jmin)*(imax-imin+1)-1)*nvar_ray+1)
  coin13_e(1,1,1,2) = xin((1-imin+(1-jmin)*(imax-imin+1)-1)*nvar_ray+2)
  coin13_e(1,1,1,3) = xin((1-imin+(1-jmin)*(imax-imin+1)-1)*nvar_ray+3)
  coin13_e(1,1,1,4) = xin((1-imin+(1-jmin)*(imax-imin+1)-1)*nvar_ray+4)

  coin23_e(1,1,1,1) = xin((nnx-imin+(1-jmin)*(imax-imin+1)-1)*nvar_ray+1)
  coin23_e(1,1,1,2) = xin((nnx-imin+(1-jmin)*(imax-imin+1)-1)*nvar_ray+2)
  coin23_e(1,1,1,3) = xin((nnx-imin+(1-jmin)*(imax-imin+1)-1)*nvar_ray+3)
  coin23_e(1,1,1,4) = xin((nnx-imin+(1-jmin)*(imax-imin+1)-1)*nvar_ray+4)
                                                                    
  coin14_e(1,1,1,1) = xin((1-imin+(nny-jmin)*(imax-imin+1)-1)*nvar_ray+1)
  coin14_e(1,1,1,2) = xin((1-imin+(nny-jmin)*(imax-imin+1)-1)*nvar_ray+2)
  coin14_e(1,1,1,3) = xin((1-imin+(nny-jmin)*(imax-imin+1)-1)*nvar_ray+3)
  coin14_e(1,1,1,4) = xin((1-imin+(nny-jmin)*(imax-imin+1)-1)*nvar_ray+4)
                                                                    
  coin24_e(1,1,1,1) = xin((nnx-imin+(nny-jmin)*(imax-imin+1)-1)*nvar_ray+1)
  coin24_e(1,1,1,2) = xin((nnx-imin+(nny-jmin)*(imax-imin+1)-1)*nvar_ray+2)
  coin24_e(1,1,1,3) = xin((nnx-imin+(nny-jmin)*(imax-imin+1)-1)*nvar_ray+3)
  coin24_e(1,1,1,4) = xin((nnx-imin+(nny-jmin)*(imax-imin+1)-1)*nvar_ray+4)

  ! send data
            
  destination = voisin(-1, 0,0)
  call MPI_Isend(face1_e ,ic_f12 ,MPI_DOUBLE_PRECISION,destination,LEFT ,MPI_Comm_world,Req( 1),ierr)

  destination = voisin(+1, 0,0)
  call MPI_Isend(face2_e ,ic_f12 ,MPI_DOUBLE_PRECISION,destination,RIGHT,MPI_COMM_WORLD,Req( 2),ierr)

  destination = voisin( 0,-1,0)
  call MPI_Isend(face3_e ,ic_f34 ,MPI_DOUBLE_PRECISION,destination,DOWN ,MPI_COMM_WORLD,Req( 3),ierr)

  destination = voisin( 0,+1,0)
  call MPI_Isend(face4_e ,ic_f34 ,MPI_DOUBLE_PRECISION,destination,UP   ,MPI_COMM_WORLD,Req( 4),ierr)

  destination = voisin(-1,-1,0)
  call MPI_Isend(coin13_e,ic_coin,MPI_DOUBLE_PRECISION,destination,LD   ,MPI_COMM_WORLD,Req( 5),ierr)

  destination = voisin(-1,+1,0)
  call MPI_Isend(coin14_e,ic_coin,MPI_DOUBLE_PRECISION,destination,LU   ,MPI_COMM_WORLD,Req( 6),ierr)

  destination = voisin(+1,-1,0)
  call MPI_Isend(coin23_e,ic_coin,MPI_DOUBLE_PRECISION,destination,RD   ,MPI_COMM_WORLD,Req( 7),ierr)

  destination = voisin(+1,+1,0)
  call MPI_Isend(coin24_e,ic_coin,MPI_DOUBLE_PRECISION,destination,RU   ,MPI_COMM_WORLD,Req( 8),ierr)

  ! receive data
  
  emetteur = voisin(+1, 0,0)
  call MPI_Irecv(face2_r ,ic_f12 ,MPI_DOUBLE_PRECISION,emetteur   ,LEFT ,MPI_COMM_WORLD,Req( 9),ierr)
  
  emetteur = voisin(-1, 0,0)
  call MPI_Irecv(face1_r ,ic_f12 ,MPI_DOUBLE_PRECISION,emetteur   ,RIGHT,MPI_COMM_WORLD,Req(10),ierr)
  
  emetteur = voisin( 0,+1,0)
  call MPI_Irecv(face4_r ,ic_f34 ,MPI_DOUBLE_PRECISION,emetteur   ,DOWN ,MPI_COMM_WORLD,Req(11),ierr)
  
  emetteur = voisin( 0,-1,0)
  call MPI_Irecv(face3_r ,ic_f34 ,MPI_DOUBLE_PRECISION,emetteur   ,UP   ,MPI_COMM_WORLD,Req(12),ierr)
  
  emetteur = voisin(+1,+1,0)
  call MPI_Irecv(coin24_r,ic_coin,MPI_DOUBLE_PRECISION,emetteur   ,LD   ,MPI_COMM_WORLD,Req(13),ierr)
  
  emetteur = voisin(+1,-1,0)
  call MPI_Irecv(coin23_r,ic_coin,MPI_DOUBLE_PRECISION,emetteur   ,LU   ,MPI_COMM_WORLD,Req(14),ierr)

  emetteur = voisin(-1,+1,0)
  call MPI_Irecv(coin14_r,ic_coin,MPI_DOUBLE_PRECISION,emetteur   ,RD   ,MPI_COMM_WORLD,Req(15),ierr)

  emetteur = voisin(-1,-1,0)
  call MPI_Irecv(coin13_r,ic_coin,MPI_DOUBLE_PRECISION,emetteur   ,RU   ,MPI_COMM_WORLD,Req(16),ierr)

  call MPI_Waitall(16,Req,Stat,ierr)

  ! 4 faces
  do j=1,nny
     xin((0-imin+(j-jmin)*(imax-imin+1)-1)*nvar_ray+1) = face1_r(1,j,1,1)
     xin((0-imin+(j-jmin)*(imax-imin+1)-1)*nvar_ray+2) = face1_r(1,j,1,2)
     xin((0-imin+(j-jmin)*(imax-imin+1)-1)*nvar_ray+3) = face1_r(1,j,1,3)
     xin((0-imin+(j-jmin)*(imax-imin+1)-1)*nvar_ray+4) = face1_r(1,j,1,4)

     xin((nnx+1-imin+(j-jmin)*(imax-imin+1)-1)*nvar_ray+1) = face2_r(1,j,1,1)
     xin((nnx+1-imin+(j-jmin)*(imax-imin+1)-1)*nvar_ray+2) = face2_r(1,j,1,2)
     xin((nnx+1-imin+(j-jmin)*(imax-imin+1)-1)*nvar_ray+3) = face2_r(1,j,1,3)
     xin((nnx+1-imin+(j-jmin)*(imax-imin+1)-1)*nvar_ray+4) = face2_r(1,j,1,4)
  enddo
  do i=1,nnx
     xin((i-imin+(0-jmin)*(imax-imin+1)-1)*nvar_ray+1) = face3_r(i,1,1,1)
     xin((i-imin+(0-jmin)*(imax-imin+1)-1)*nvar_ray+2) = face3_r(i,1,1,2)
     xin((i-imin+(0-jmin)*(imax-imin+1)-1)*nvar_ray+3) = face3_r(i,1,1,3)
     xin((i-imin+(0-jmin)*(imax-imin+1)-1)*nvar_ray+4) = face3_r(i,1,1,4)

     xin((i-imin+(nny+1-jmin)*(imax-imin+1)-1)*nvar_ray+1) = face4_r(i,1,1,1)
     xin((i-imin+(nny+1-jmin)*(imax-imin+1)-1)*nvar_ray+2) = face4_r(i,1,1,2)
     xin((i-imin+(nny+1-jmin)*(imax-imin+1)-1)*nvar_ray+3) = face4_r(i,1,1,3)
     xin((i-imin+(nny+1-jmin)*(imax-imin+1)-1)*nvar_ray+4) = face4_r(i,1,1,4)
  enddo
  
  ! 4 corners
  xin((0-imin+(0-jmin)*(imax-imin+1)-1)*nvar_ray+1) = coin13_r(1,1,1,1)
  xin((0-imin+(0-jmin)*(imax-imin+1)-1)*nvar_ray+2) = coin13_r(1,1,1,2)
  xin((0-imin+(0-jmin)*(imax-imin+1)-1)*nvar_ray+3) = coin13_r(1,1,1,3)
  xin((0-imin+(0-jmin)*(imax-imin+1)-1)*nvar_ray+4) = coin13_r(1,1,1,4)

  xin((0-imin+(nny+1-jmin)*(imax-imin+1)-1)*nvar_ray+1) = coin14_r(1,1,1,1)
  xin((0-imin+(nny+1-jmin)*(imax-imin+1)-1)*nvar_ray+2) = coin14_r(1,1,1,2)
  xin((0-imin+(nny+1-jmin)*(imax-imin+1)-1)*nvar_ray+3) = coin14_r(1,1,1,3)
  xin((0-imin+(nny+1-jmin)*(imax-imin+1)-1)*nvar_ray+4) = coin14_r(1,1,1,4)

  xin((nnx+1-imin+(0-jmin)*(imax-imin+1)-1)*nvar_ray+1) = coin23_r(1,1,1,1)
  xin((nnx+1-imin+(0-jmin)*(imax-imin+1)-1)*nvar_ray+2) = coin23_r(1,1,1,2)
  xin((nnx+1-imin+(0-jmin)*(imax-imin+1)-1)*nvar_ray+3) = coin23_r(1,1,1,3)
  xin((nnx+1-imin+(0-jmin)*(imax-imin+1)-1)*nvar_ray+4) = coin23_r(1,1,1,4)

  xin((nnx+1-imin+(nny+1-jmin)*(imax-imin+1)-1)*nvar_ray+1) = coin24_r(1,1,1,1)
  xin((nnx+1-imin+(nny+1-jmin)*(imax-imin+1)-1)*nvar_ray+2) = coin24_r(1,1,1,2)
  xin((nnx+1-imin+(nny+1-jmin)*(imax-imin+1)-1)*nvar_ray+3) = coin24_r(1,1,1,3)
  xin((nnx+1-imin+(nny+1-jmin)*(imax-imin+1)-1)*nvar_ray+4) = coin24_r(1,1,1,4)
  
  return
  
end subroutine com_2D_xin

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine COM_3D_XIN
!
!> Communications 3D for the xin vector.
!<
subroutine com_3D_xin(xin)

  use communication
  use parameters
  use para
  use tabgmres     , only : imin,imax,jmin,jmax,kmin,kmax

  implicit none

  include 'mpif.h'

  real, dimension(-nvar_ray+1:(imax-imin+(jmax-jmin)*(imax-imin+1)+(kmax-kmin)*(imax-imin+1)*(jmax-jmin+1))*nvar_ray) :: xin

  integer, parameter                     :: F1=1, F2=2, F3=3, F4=4, F5=5, F6=6
  integer, parameter                     :: AR13=13, AR14=14, AR15=15, AR16=16
  integer, parameter                     :: AR23=23, AR24=24, AR25=25, AR26=26
  integer, parameter                     :: AR35=35, AR36=36
  integer, parameter                     :: AR45=45, AR46=46
  integer, parameter                     :: C135=135, C136=136, C145=145, C146=146
  integer, parameter                     :: C235=235, C236=236, C245=245, C246=246
  integer                                :: i,j,k,ierr
  integer                                :: ic_f12,ic_f34,ic_f56, ic_arx, ic_ary, ic_arz, ic_coin
  integer                                :: destination, emetteur
  integer, dimension(52  )               :: Req
  integer, dimension(MPI_STATUS_SIZE,52) :: Stat

  integer :: nnx, nny, nnz

  nnx = nx(1) ; nny = nx(2) ; nnz = nx(3)

  ic_f12  = nvar_ray*nny*nnz*2
  ic_f34  = nvar_ray*nnx*nnz*2
  ic_f56  = nvar_ray*nnx*nny*2

  ic_arx  = nvar_ray*2*2*nnx
  ic_ary  = nvar_ray*2*2*nny
  ic_arz  = nvar_ray*2*2*nnz

  ic_coin = nvar_ray*2*2*2

  Req(1:52) = MPI_REQUEST_NULL

  ! 6 faces
  do j=1,nny
     do k=1,nnz
        face1_e(1,j,k,1) = xin((1-imin+(j-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+1)
        face1_e(1,j,k,2) = xin((1-imin+(j-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+2)
        face1_e(1,j,k,3) = xin((1-imin+(j-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+3)
        face1_e(1,j,k,4) = xin((1-imin+(j-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+4)
        face1_e(1,j,k,5) = xin((1-imin+(j-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+5)

        face2_e(1,j,k,1) = xin((nnx-imin+(j-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+1)
        face2_e(1,j,k,2) = xin((nnx-imin+(j-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+2)
        face2_e(1,j,k,3) = xin((nnx-imin+(j-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+3)
        face2_e(1,j,k,4) = xin((nnx-imin+(j-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+4)
        face2_e(1,j,k,5) = xin((nnx-imin+(j-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+5)
     enddo
  enddo
  do i=1,nnx
     do k=1,nnz
        face3_e(i,1,k,1) = xin((i-imin+(1-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+1)
        face3_e(i,1,k,2) = xin((i-imin+(1-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+2)
        face3_e(i,1,k,3) = xin((i-imin+(1-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+3)
        face3_e(i,1,k,4) = xin((i-imin+(1-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+4)
        face3_e(i,1,k,5) = xin((i-imin+(1-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+5)
        
        face4_e(i,1,k,1) = xin((i-imin+(nny-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+1)
        face4_e(i,1,k,2) = xin((i-imin+(nny-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+2)
        face4_e(i,1,k,3) = xin((i-imin+(nny-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+3)
        face4_e(i,1,k,4) = xin((i-imin+(nny-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+4)
        face4_e(i,1,k,5) = xin((i-imin+(nny-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+5)
     enddo
  enddo
  do i=1,nnx
     do j=1,nny
        face5_e(i,j,1,1) = xin((i-imin+(j-jmin)*(imax-imin+1)+(1-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+1)
        face5_e(i,j,1,2) = xin((i-imin+(j-jmin)*(imax-imin+1)+(1-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+2)
        face5_e(i,j,1,3) = xin((i-imin+(j-jmin)*(imax-imin+1)+(1-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+3)
        face5_e(i,j,1,4) = xin((i-imin+(j-jmin)*(imax-imin+1)+(1-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+4)
        face5_e(i,j,1,5) = xin((i-imin+(j-jmin)*(imax-imin+1)+(1-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+5)
                  
        face6_e(i,j,1,1) = xin((i-imin+(j-jmin)*(imax-imin+1)+(nnz-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+1)
        face6_e(i,j,1,2) = xin((i-imin+(j-jmin)*(imax-imin+1)+(nnz-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+2)
        face6_e(i,j,1,3) = xin((i-imin+(j-jmin)*(imax-imin+1)+(nnz-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+3)
        face6_e(i,j,1,4) = xin((i-imin+(j-jmin)*(imax-imin+1)+(nnz-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+4)
        face6_e(i,j,1,5) = xin((i-imin+(j-jmin)*(imax-imin+1)+(nnz-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+5)
     enddo
  enddo
  
  ! 12 edges
  do k=1,nnz
     ar13_e(1, 1, k, 1  ) = xin((1-imin+(1-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+1)
     ar13_e(1, 1, k, 2  ) = xin((1-imin+(1-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+2)
     ar13_e(1, 1, k, 3  ) = xin((1-imin+(1-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+3)
     ar13_e(1, 1, k, 4  ) = xin((1-imin+(1-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+4)
     ar13_e(1, 1, k, 5  ) = xin((1-imin+(1-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+5)
     
     ar14_e(1, 1, k, 1  ) = xin((1-imin+(nny-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+1)
     ar14_e(1, 1, k, 2  ) = xin((1-imin+(nny-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+2)
     ar14_e(1, 1, k, 3  ) = xin((1-imin+(nny-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+3)
     ar14_e(1, 1, k, 4  ) = xin((1-imin+(nny-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+4)
     ar14_e(1, 1, k, 5  ) = xin((1-imin+(nny-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+5)

     ar23_e(1, 1, k, 1  ) = xin((nnx-imin+(1-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+1)
     ar23_e(1, 1, k, 2  ) = xin((nnx-imin+(1-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+2)
     ar23_e(1, 1, k, 3  ) = xin((nnx-imin+(1-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+3)
     ar23_e(1, 1, k, 4  ) = xin((nnx-imin+(1-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+4)
     ar23_e(1, 1, k, 5  ) = xin((nnx-imin+(1-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+5)

     ar24_e(1, 1, k, 1  ) = xin((nnx-imin+(nny-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+1)
     ar24_e(1, 1, k, 2  ) = xin((nnx-imin+(nny-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+2)
     ar24_e(1, 1, k, 3  ) = xin((nnx-imin+(nny-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+3)
     ar24_e(1, 1, k, 4  ) = xin((nnx-imin+(nny-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+4)
     ar24_e(1, 1, k, 5  ) = xin((nnx-imin+(nny-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+5)
  enddo
  do j=1,nny
     ar15_e(1, j, 1, 1  ) = xin((1-imin+(j-jmin)*(imax-imin+1)+(1-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+1)
     ar15_e(1, j, 1, 2  ) = xin((1-imin+(j-jmin)*(imax-imin+1)+(1-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+2)
     ar15_e(1, j, 1, 3  ) = xin((1-imin+(j-jmin)*(imax-imin+1)+(1-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+3)
     ar15_e(1, j, 1, 4  ) = xin((1-imin+(j-jmin)*(imax-imin+1)+(1-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+4)
     ar15_e(1, j, 1, 5  ) = xin((1-imin+(j-jmin)*(imax-imin+1)+(1-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+5)
     
     ar25_e(1, j, 1, 1  ) = xin((nnx-imin+(j-jmin)*(imax-imin+1)+(1-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+1)
     ar25_e(1, j, 1, 2  ) = xin((nnx-imin+(j-jmin)*(imax-imin+1)+(1-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+2)
     ar25_e(1, j, 1, 3  ) = xin((nnx-imin+(j-jmin)*(imax-imin+1)+(1-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+3)
     ar25_e(1, j, 1, 4  ) = xin((nnx-imin+(j-jmin)*(imax-imin+1)+(1-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+4)
     ar25_e(1, j, 1, 5  ) = xin((nnx-imin+(j-jmin)*(imax-imin+1)+(1-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+5)

     ar16_e(1, j, 1, 1  ) = xin((1-imin+(j-jmin)*(imax-imin+1)+(nnz-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+1)
     ar16_e(1, j, 1, 2  ) = xin((1-imin+(j-jmin)*(imax-imin+1)+(nnz-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+2)
     ar16_e(1, j, 1, 3  ) = xin((1-imin+(j-jmin)*(imax-imin+1)+(nnz-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+3)
     ar16_e(1, j, 1, 4  ) = xin((1-imin+(j-jmin)*(imax-imin+1)+(nnz-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+4)
     ar16_e(1, j, 1, 5  ) = xin((1-imin+(j-jmin)*(imax-imin+1)+(nnz-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+5)
     
     ar26_e(1, j, 1, 1  ) = xin((nnx-imin+(j-jmin)*(imax-imin+1)+(nnz-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+1)
     ar26_e(1, j, 1, 2  ) = xin((nnx-imin+(j-jmin)*(imax-imin+1)+(nnz-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+2)
     ar26_e(1, j, 1, 3  ) = xin((nnx-imin+(j-jmin)*(imax-imin+1)+(nnz-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+3)
     ar26_e(1, j, 1, 4  ) = xin((nnx-imin+(j-jmin)*(imax-imin+1)+(nnz-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+4)
     ar26_e(1, j, 1, 5  ) = xin((nnx-imin+(j-jmin)*(imax-imin+1)+(nnz-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+5)
  enddo
  do i=1,nnx
     ar35_e(i, 1, 1, 1  ) = xin((i-imin+(1-jmin)*(imax-imin+1)+(1-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+1)
     ar35_e(i, 1, 1, 2  ) = xin((i-imin+(1-jmin)*(imax-imin+1)+(1-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+2)
     ar35_e(i, 1, 1, 3  ) = xin((i-imin+(1-jmin)*(imax-imin+1)+(1-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+3)
     ar35_e(i, 1, 1, 4  ) = xin((i-imin+(1-jmin)*(imax-imin+1)+(1-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+4)
     ar35_e(i, 1, 1, 5  ) = xin((i-imin+(1-jmin)*(imax-imin+1)+(1-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+5)
     
     ar45_e(i, 1, 1, 1  ) = xin((i-imin+(nny-jmin)*(imax-imin+1)+(1-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+1)
     ar45_e(i, 1, 1, 2  ) = xin((i-imin+(nny-jmin)*(imax-imin+1)+(1-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+2)
     ar45_e(i, 1, 1, 3  ) = xin((i-imin+(nny-jmin)*(imax-imin+1)+(1-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+3)
     ar45_e(i, 1, 1, 4  ) = xin((i-imin+(nny-jmin)*(imax-imin+1)+(1-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+4)
     ar45_e(i, 1, 1, 5  ) = xin((i-imin+(nny-jmin)*(imax-imin+1)+(1-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+5)

     ar36_e(i, 1, 1, 1  ) = xin((i-imin+(1-jmin)*(imax-imin+1)+(nnz-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+1)
     ar36_e(i, 1, 1, 2  ) = xin((i-imin+(1-jmin)*(imax-imin+1)+(nnz-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+2)
     ar36_e(i, 1, 1, 3  ) = xin((i-imin+(1-jmin)*(imax-imin+1)+(nnz-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+3)
     ar36_e(i, 1, 1, 4  ) = xin((i-imin+(1-jmin)*(imax-imin+1)+(nnz-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+4)
     ar36_e(i, 1, 1, 5  ) = xin((i-imin+(1-jmin)*(imax-imin+1)+(nnz-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+5)
     
     ar46_e(i, 1, 1, 1  ) = xin((i-imin+(nny-jmin)*(imax-imin+1)+(nnz-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+1)
     ar46_e(i, 1, 1, 2  ) = xin((i-imin+(nny-jmin)*(imax-imin+1)+(nnz-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+2)
     ar46_e(i, 1, 1, 3  ) = xin((i-imin+(nny-jmin)*(imax-imin+1)+(nnz-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+3)
     ar46_e(i, 1, 1, 4  ) = xin((i-imin+(nny-jmin)*(imax-imin+1)+(nnz-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+4)
     ar46_e(i, 1, 1, 5  ) = xin((i-imin+(nny-jmin)*(imax-imin+1)+(nnz-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+5)
  enddo
 

  ! 8 corners
  coin135_e(1, 1, 1, 1) = xin((1-imin+(1-jmin)*(imax-imin+1)+(1-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+1)
  coin135_e(1, 1, 1, 2) = xin((1-imin+(1-jmin)*(imax-imin+1)+(1-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+2)
  coin135_e(1, 1, 1, 3) = xin((1-imin+(1-jmin)*(imax-imin+1)+(1-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+3)
  coin135_e(1, 1, 1, 4) = xin((1-imin+(1-jmin)*(imax-imin+1)+(1-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+4)
  coin135_e(1, 1, 1, 5) = xin((1-imin+(1-jmin)*(imax-imin+1)+(1-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+5)

  coin136_e(1, 1, 1, 1) = xin((1-imin+(1-jmin)*(imax-imin+1)+(nnz-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+1)
  coin136_e(1, 1, 1, 2) = xin((1-imin+(1-jmin)*(imax-imin+1)+(nnz-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+2)
  coin136_e(1, 1, 1, 3) = xin((1-imin+(1-jmin)*(imax-imin+1)+(nnz-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+3)
  coin136_e(1, 1, 1, 4) = xin((1-imin+(1-jmin)*(imax-imin+1)+(nnz-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+4)
  coin136_e(1, 1, 1, 5) = xin((1-imin+(1-jmin)*(imax-imin+1)+(nnz-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+5)

  coin145_e(1, 1, 1, 1) = xin((1-imin+(nny-jmin)*(imax-imin+1)+(1-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+1)
  coin145_e(1, 1, 1, 2) = xin((1-imin+(nny-jmin)*(imax-imin+1)+(1-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+2)
  coin145_e(1, 1, 1, 3) = xin((1-imin+(nny-jmin)*(imax-imin+1)+(1-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+3)
  coin145_e(1, 1, 1, 4) = xin((1-imin+(nny-jmin)*(imax-imin+1)+(1-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+4)
  coin145_e(1, 1, 1, 5) = xin((1-imin+(nny-jmin)*(imax-imin+1)+(1-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+5)

  coin146_e(1, 1, 1, 1) = xin((1-imin+(nny-jmin)*(imax-imin+1)+(nnz-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+1)
  coin146_e(1, 1, 1, 2) = xin((1-imin+(nny-jmin)*(imax-imin+1)+(nnz-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+2)
  coin146_e(1, 1, 1, 3) = xin((1-imin+(nny-jmin)*(imax-imin+1)+(nnz-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+3)
  coin146_e(1, 1, 1, 4) = xin((1-imin+(nny-jmin)*(imax-imin+1)+(nnz-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+4)
  coin146_e(1, 1, 1, 5) = xin((1-imin+(nny-jmin)*(imax-imin+1)+(nnz-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+5)

  coin235_e(1, 1, 1, 1) = xin((nnx-imin+(1-jmin)*(imax-imin+1)+(1-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+1)
  coin235_e(1, 1, 1, 2) = xin((nnx-imin+(1-jmin)*(imax-imin+1)+(1-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+2)
  coin235_e(1, 1, 1, 3) = xin((nnx-imin+(1-jmin)*(imax-imin+1)+(1-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+3)
  coin235_e(1, 1, 1, 4) = xin((nnx-imin+(1-jmin)*(imax-imin+1)+(1-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+4)
  coin235_e(1, 1, 1, 5) = xin((nnx-imin+(1-jmin)*(imax-imin+1)+(1-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+5)

  coin236_e(1, 1, 1, 1) = xin((nnx-imin+(1-jmin)*(imax-imin+1)+(nnz-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+1)
  coin236_e(1, 1, 1, 2) = xin((nnx-imin+(1-jmin)*(imax-imin+1)+(nnz-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+2)
  coin236_e(1, 1, 1, 3) = xin((nnx-imin+(1-jmin)*(imax-imin+1)+(nnz-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+3)
  coin236_e(1, 1, 1, 4) = xin((nnx-imin+(1-jmin)*(imax-imin+1)+(nnz-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+4)
  coin236_e(1, 1, 1, 5) = xin((nnx-imin+(1-jmin)*(imax-imin+1)+(nnz-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+5)

  coin245_e(1, 1, 1, 1) = xin((nnx-imin+(nny-jmin)*(imax-imin+1)+(1-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+1)
  coin245_e(1, 1, 1, 2) = xin((nnx-imin+(nny-jmin)*(imax-imin+1)+(1-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+2)
  coin245_e(1, 1, 1, 3) = xin((nnx-imin+(nny-jmin)*(imax-imin+1)+(1-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+3)
  coin245_e(1, 1, 1, 4) = xin((nnx-imin+(nny-jmin)*(imax-imin+1)+(1-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+4)
  coin245_e(1, 1, 1, 5) = xin((nnx-imin+(nny-jmin)*(imax-imin+1)+(1-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+5)

  coin246_e(1, 1, 1, 1) = xin((nnx-imin+(nny-jmin)*(imax-imin+1)+(nnz-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+1)
  coin246_e(1, 1, 1, 2) = xin((nnx-imin+(nny-jmin)*(imax-imin+1)+(nnz-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+2)
  coin246_e(1, 1, 1, 3) = xin((nnx-imin+(nny-jmin)*(imax-imin+1)+(nnz-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+3)
  coin246_e(1, 1, 1, 4) = xin((nnx-imin+(nny-jmin)*(imax-imin+1)+(nnz-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+4)
  coin246_e(1, 1, 1, 5) = xin((nnx-imin+(nny-jmin)*(imax-imin+1)+(nnz-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+5)


  ! send data

  ! Faces
  destination = voisin(-1, 0, 0)
  call MPI_Isend(face1_e,ic_f12, MPI_DOUBLE_PRECISION,destination,F1 ,MPI_Comm_world,Req( 1),ierr)

  destination = voisin(+1, 0, 0)
  call MPI_Isend(face2_e,ic_f12, MPI_DOUBLE_PRECISION,destination,F2 ,MPI_COMM_WORLD,Req( 2),ierr)

  destination = voisin( 0,-1, 0)
  call MPI_Isend(face3_e,ic_f34, MPI_DOUBLE_PRECISION,destination,F3 ,MPI_COMM_WORLD,Req( 3),ierr)

  destination = voisin( 0,+1, 0)
  call MPI_Isend(face4_e,ic_f34, MPI_DOUBLE_PRECISION,destination,F4 ,MPI_COMM_WORLD,Req( 4),ierr)

  destination = voisin( 0, 0,-1)
  call MPI_Isend(face5_e,ic_f56, MPI_DOUBLE_PRECISION,destination,F5 ,MPI_COMM_WORLD,Req( 5),ierr)

  destination = voisin( 0, 0,+1)
  call MPI_Isend(face6_e,ic_f56, MPI_DOUBLE_PRECISION,destination,F6 ,MPI_COMM_WORLD,Req( 6),ierr)



  ! Edges
  destination = voisin(-1,-1,0)
  call MPI_Isend(ar13_e,ic_arz, MPI_DOUBLE_PRECISION,destination,AR13   ,MPI_COMM_WORLD,Req( 7),ierr)

  destination = voisin(-1,+1,0)
  call MPI_Isend(ar14_e,ic_arz, MPI_DOUBLE_PRECISION,destination,AR14   ,MPI_COMM_WORLD,Req( 8),ierr)

  destination = voisin(+1,-1,0)
  call MPI_Isend(ar23_e,ic_arz, MPI_DOUBLE_PRECISION,destination,AR23   ,MPI_COMM_WORLD,Req( 9),ierr)

  destination = voisin(+1,+1,0)
  call MPI_Isend(ar24_e,ic_arz, MPI_DOUBLE_PRECISION,destination,AR24   ,MPI_COMM_WORLD,Req(10),ierr)


  destination = voisin(-1, 0,-1)
  call MPI_Isend(ar15_e,ic_ary, MPI_DOUBLE_PRECISION,destination,AR15   ,MPI_COMM_WORLD,Req(11),ierr)

  destination = voisin(-1, 0,+1)
  call MPI_Isend(ar16_e,ic_ary, MPI_DOUBLE_PRECISION,destination,AR16   ,MPI_COMM_WORLD,Req(12),ierr)

  destination = voisin(+1, 0,-1)
  call MPI_Isend(ar25_e,ic_ary, MPI_DOUBLE_PRECISION,destination,AR25   ,MPI_COMM_WORLD,Req(13),ierr)

  destination = voisin(+1, 0,+1)
  call MPI_Isend(ar26_e,ic_ary, MPI_DOUBLE_PRECISION,destination,AR26   ,MPI_COMM_WORLD,Req(14),ierr)


  destination = voisin( 0,-1,-1)
  call MPI_Isend(ar35_e,ic_arx, MPI_DOUBLE_PRECISION,destination,AR35   ,MPI_COMM_WORLD,Req(15),ierr)

  destination = voisin( 0,-1,+1)
  call MPI_Isend(ar36_e,ic_arx, MPI_DOUBLE_PRECISION,destination,AR36   ,MPI_COMM_WORLD,Req(16),ierr)

  destination = voisin( 0,+1,-1)
  call MPI_Isend(ar45_e,ic_arx, MPI_DOUBLE_PRECISION,destination,AR45   ,MPI_COMM_WORLD,Req(17),ierr)

  destination = voisin( 0,+1,+1)
  call MPI_Isend(ar46_e,ic_arx, MPI_DOUBLE_PRECISION,destination,AR46   ,MPI_COMM_WORLD,Req(18),ierr)


  ! Corners
  destination = voisin( -1,-1,-1)
  call MPI_Isend(coin135_e,ic_coin, MPI_DOUBLE_PRECISION,destination,C135   ,MPI_COMM_WORLD,Req(19),ierr)

  destination = voisin( -1,-1,+1)
  call MPI_Isend(coin136_e,ic_coin, MPI_DOUBLE_PRECISION,destination,C136   ,MPI_COMM_WORLD,Req(20),ierr)

  destination = voisin( -1,+1,-1)
  call MPI_Isend(coin145_e,ic_coin, MPI_DOUBLE_PRECISION,destination,C145   ,MPI_COMM_WORLD,Req(21),ierr)

  destination = voisin( -1,+1,+1)
  call MPI_Isend(coin146_e,ic_coin, MPI_DOUBLE_PRECISION,destination,C146   ,MPI_COMM_WORLD,Req(22),ierr)

  destination = voisin( +1,-1,-1)
  call MPI_Isend(coin235_e,ic_coin, MPI_DOUBLE_PRECISION,destination,C235   ,MPI_COMM_WORLD,Req(23),ierr)

  destination = voisin( +1,-1,+1)
  call MPI_Isend(coin236_e,ic_coin, MPI_DOUBLE_PRECISION,destination,C236   ,MPI_COMM_WORLD,Req(24),ierr)

  destination = voisin( +1,+1,-1)
  call MPI_Isend(coin245_e,ic_coin, MPI_DOUBLE_PRECISION,destination,C245   ,MPI_COMM_WORLD,Req(25),ierr)

  destination = voisin( +1,+1,+1)
  call MPI_Isend(coin246_e,ic_coin, MPI_DOUBLE_PRECISION,destination,C246   ,MPI_COMM_WORLD,Req(26),ierr)

 

  ! receive data
  

  ! Faces
  emetteur = voisin(+1, 0, 0)
  call MPI_Irecv(face2_r,ic_f12, MPI_DOUBLE_PRECISION,emetteur,F1 ,MPI_Comm_world,Req(27),ierr)

  emetteur = voisin(-1, 0, 0)
  call MPI_Irecv(face1_r,ic_f12, MPI_DOUBLE_PRECISION,emetteur,F2 ,MPI_COMM_WORLD,Req(28),ierr)

  emetteur = voisin( 0,+1, 0)
  call MPI_Irecv(face4_r,ic_f34, MPI_DOUBLE_PRECISION,emetteur,F3 ,MPI_COMM_WORLD,Req(29),ierr)

  emetteur = voisin( 0,-1, 0)
  call MPI_Irecv(face3_r,ic_f34, MPI_DOUBLE_PRECISION,emetteur,F4 ,MPI_COMM_WORLD,Req(30),ierr)

  emetteur = voisin( 0, 0,+1)
  call MPI_Irecv(face6_r,ic_f56, MPI_DOUBLE_PRECISION,emetteur,F5 ,MPI_COMM_WORLD,Req(31),ierr)

  emetteur = voisin( 0, 0,-1)
  call MPI_Irecv(face5_r,ic_f56, MPI_DOUBLE_PRECISION,emetteur,F6 ,MPI_COMM_WORLD,Req(32),ierr)



  ! Edges
  emetteur = voisin(+1,+1,0)
  call MPI_Irecv(ar24_r,ic_arz, MPI_DOUBLE_PRECISION,emetteur,AR13   ,MPI_COMM_WORLD,Req(33),ierr)

  emetteur = voisin(+1,-1,0)
  call MPI_Irecv(ar23_r,ic_arz, MPI_DOUBLE_PRECISION,emetteur,AR14   ,MPI_COMM_WORLD,Req(34),ierr)

  emetteur = voisin(-1,+1,0)
  call MPI_Irecv(ar14_r,ic_arz, MPI_DOUBLE_PRECISION,emetteur,AR23   ,MPI_COMM_WORLD,Req(35),ierr)

  emetteur = voisin(-1,-1,0)
  call MPI_Irecv(ar13_r,ic_arz, MPI_DOUBLE_PRECISION,emetteur,AR24   ,MPI_COMM_WORLD,Req(36),ierr)


  emetteur = voisin(+1, 0,+1)
  call MPI_Irecv(ar26_r,ic_ary, MPI_DOUBLE_PRECISION,emetteur,AR15   ,MPI_COMM_WORLD,Req(37),ierr)

  emetteur = voisin(+1, 0,-1)
  call MPI_Irecv(ar25_r,ic_ary, MPI_DOUBLE_PRECISION,emetteur,AR16   ,MPI_COMM_WORLD,Req(38),ierr)

  emetteur = voisin(-1, 0,+1)
  call MPI_Irecv(ar16_r,ic_ary, MPI_DOUBLE_PRECISION,emetteur,AR25   ,MPI_COMM_WORLD,Req(39),ierr)

  emetteur = voisin(-1, 0,-1)
  call MPI_Irecv(ar15_r,ic_ary, MPI_DOUBLE_PRECISION,emetteur,AR26   ,MPI_COMM_WORLD,Req(40),ierr)


  emetteur = voisin( 0,+1,+1)
  call MPI_Irecv(ar46_r,ic_arx, MPI_DOUBLE_PRECISION,emetteur,AR35   ,MPI_COMM_WORLD,Req(41),ierr)

  emetteur = voisin( 0,+1,-1)
  call MPI_Irecv(ar45_r,ic_arx, MPI_DOUBLE_PRECISION,emetteur,AR36   ,MPI_COMM_WORLD,Req(42),ierr)

  emetteur = voisin( 0,-1,+1)
  call MPI_Irecv(ar36_r,ic_arx, MPI_DOUBLE_PRECISION,emetteur,AR45   ,MPI_COMM_WORLD,Req(43),ierr)

  emetteur = voisin( 0,-1,-1)
  call MPI_Irecv(ar35_r,ic_arx, MPI_DOUBLE_PRECISION,emetteur,AR46   ,MPI_COMM_WORLD,Req(44),ierr)


  ! Corners
  emetteur = voisin( +1,+1,+1)
  call MPI_Irecv(coin246_r,ic_coin, MPI_DOUBLE_PRECISION,emetteur,C135   ,MPI_COMM_WORLD,Req(45),ierr)

  emetteur = voisin( +1,+1,-1)
  call MPI_Irecv(coin245_r,ic_coin, MPI_DOUBLE_PRECISION,emetteur,C136   ,MPI_COMM_WORLD,Req(46),ierr)

  emetteur = voisin( +1,-1,+1)
  call MPI_Irecv(coin236_r,ic_coin, MPI_DOUBLE_PRECISION,emetteur,C145   ,MPI_COMM_WORLD,Req(47),ierr)

  emetteur = voisin( +1,-1,-1)
  call MPI_Irecv(coin235_r,ic_coin, MPI_DOUBLE_PRECISION,emetteur,C146   ,MPI_COMM_WORLD,Req(48),ierr)

  emetteur = voisin( -1,+1,+1)
  call MPI_Irecv(coin146_r,ic_coin, MPI_DOUBLE_PRECISION,emetteur,C235   ,MPI_COMM_WORLD,Req(49),ierr)

  emetteur = voisin( -1,+1,-1)
  call MPI_Irecv(coin145_r,ic_coin, MPI_DOUBLE_PRECISION,emetteur,C236   ,MPI_COMM_WORLD,Req(50),ierr)

  emetteur = voisin( -1,-1,+1)
  call MPI_Irecv(coin136_r,ic_coin, MPI_DOUBLE_PRECISION,emetteur,C245   ,MPI_COMM_WORLD,Req(51),ierr)

  emetteur = voisin( -1,-1,-1)
  call MPI_Irecv(coin135_r,ic_coin, MPI_DOUBLE_PRECISION,emetteur,C246   ,MPI_COMM_WORLD,Req(52),ierr)
   
  call MPI_Waitall(52,Req,Stat,ierr)

  call MPI_BARRIER(MPI_COMM_WORLD,ierr)


  ! 6 faces
  do j=1,nny
     do k=1,nnz
        xin((0-imin+(j-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+1) = face1_r(1,j,k,1) 
        xin((0-imin+(j-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+2) = face1_r(1,j,k,2)  
        xin((0-imin+(j-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+3) = face1_r(1,j,k,3) 
        xin((0-imin+(j-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+4) = face1_r(1,j,k,4)
        xin((0-imin+(j-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+5) = face1_r(1,j,k,5) 
        
        xin((nnx+1-imin+(j-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+1) = face2_r(1,j,k,1) 
        xin((nnx+1-imin+(j-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+2) = face2_r(1,j,k,2) 
        xin((nnx+1-imin+(j-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+3) = face2_r(1,j,k,3) 
        xin((nnx+1-imin+(j-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+4) = face2_r(1,j,k,4) 
        xin((nnx+1-imin+(j-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+5) = face2_r(1,j,k,5) 
     enddo
  enddo
  do i=1,nnx
     do k=1,nnz
        xin((i-imin+(0-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+1) = face3_r(i,1,k,1)  
        xin((i-imin+(0-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+2) = face3_r(i,1,k,2)  
        xin((i-imin+(0-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+3) = face3_r(i,1,k,3)  
        xin((i-imin+(0-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+4) = face3_r(i,1,k,4)  
        xin((i-imin+(0-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+5) = face3_r(i,1,k,5)  

        xin((i-imin+(nny+1-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+1) = face4_r(i,1,k,1) 
        xin((i-imin+(nny+1-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+2) = face4_r(i,1,k,2) 
        xin((i-imin+(nny+1-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+3) = face4_r(i,1,k,3) 
        xin((i-imin+(nny+1-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+4) = face4_r(i,1,k,4) 
        xin((i-imin+(nny+1-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+5) = face4_r(i,1,k,5) 
     enddo
  enddo
  do i=1,nnx
     do j=1,nny
        xin((i-imin+(j-jmin)*(imax-imin+1)+(0-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+1) = face5_r(i,j,1,1) 
        xin((i-imin+(j-jmin)*(imax-imin+1)+(0-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+2) = face5_r(i,j,1,2) 
        xin((i-imin+(j-jmin)*(imax-imin+1)+(0-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+3) = face5_r(i,j,1,3) 
        xin((i-imin+(j-jmin)*(imax-imin+1)+(0-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+4) = face5_r(i,j,1,4) 
        xin((i-imin+(j-jmin)*(imax-imin+1)+(0-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+5) = face5_r(i,j,1,5) 

        xin((i-imin+(j-jmin)*(imax-imin+1)+(nnz+1-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+1) = face6_r(i,j,1,1) 
        xin((i-imin+(j-jmin)*(imax-imin+1)+(nnz+1-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+2) = face6_r(i,j,1,2) 
        xin((i-imin+(j-jmin)*(imax-imin+1)+(nnz+1-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+3) = face6_r(i,j,1,3) 
        xin((i-imin+(j-jmin)*(imax-imin+1)+(nnz+1-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+4) = face6_r(i,j,1,4) 
        xin((i-imin+(j-jmin)*(imax-imin+1)+(nnz+1-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+5) = face6_r(i,j,1,5) 
     enddo
  enddo

  ! 12 edges
  do k=1,nnz
     xin((0-imin+(0-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+1) = ar13_r(1, 1, k, 1  )
     xin((0-imin+(0-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+2) = ar13_r(1, 1, k, 2  )
     xin((0-imin+(0-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+3) = ar13_r(1, 1, k, 3  )
     xin((0-imin+(0-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+4) = ar13_r(1, 1, k, 4  )
     xin((0-imin+(0-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+5) = ar13_r(1, 1, k, 5  )

     xin((0-imin+(nny+1-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+1) = ar14_r(1, 1, k, 1  )
     xin((0-imin+(nny+1-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+2) = ar14_r(1, 1, k, 2  )
     xin((0-imin+(nny+1-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+3) = ar14_r(1, 1, k, 3  )
     xin((0-imin+(nny+1-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+4) = ar14_r(1, 1, k, 4  )
     xin((0-imin+(nny+1-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+5) = ar14_r(1, 1, k, 5  )
     
     xin((nnx+1-imin+(0-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+1) = ar23_r(1, 1, k, 1  )
     xin((nnx+1-imin+(0-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+2) = ar23_r(1, 1, k, 2  )
     xin((nnx+1-imin+(0-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+3) = ar23_r(1, 1, k, 3  )
     xin((nnx+1-imin+(0-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+4) = ar23_r(1, 1, k, 4  )
     xin((nnx+1-imin+(0-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+5) = ar23_r(1, 1, k, 5  )

     xin((nnx+1-imin+(nny+1-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+1) = ar24_r(1, 1, k, 1  )
     xin((nnx+1-imin+(nny+1-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+2) = ar24_r(1, 1, k, 2  )
     xin((nnx+1-imin+(nny+1-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+3) = ar24_r(1, 1, k, 3  )
     xin((nnx+1-imin+(nny+1-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+4) = ar24_r(1, 1, k, 4  )
     xin((nnx+1-imin+(nny+1-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+5) = ar24_r(1, 1, k, 5  )
  enddo
  do j=1,nny
     xin((0-imin+(j-jmin)*(imax-imin+1)+(0-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+1) = ar15_r(1, j, 1, 1  ) 
     xin((0-imin+(j-jmin)*(imax-imin+1)+(0-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+2) = ar15_r(1, j, 1, 2  ) 
     xin((0-imin+(j-jmin)*(imax-imin+1)+(0-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+3) = ar15_r(1, j, 1, 3  ) 
     xin((0-imin+(j-jmin)*(imax-imin+1)+(0-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+4) = ar15_r(1, j, 1, 4  ) 
     xin((0-imin+(j-jmin)*(imax-imin+1)+(0-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+5) = ar15_r(1, j, 1, 5  ) 

     xin((nnx+1-imin+(j-jmin)*(imax-imin+1)+(0-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+1) = ar25_r(1, j, 1, 1  )
     xin((nnx+1-imin+(j-jmin)*(imax-imin+1)+(0-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+2) = ar25_r(1, j, 1, 2  )
     xin((nnx+1-imin+(j-jmin)*(imax-imin+1)+(0-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+3) = ar25_r(1, j, 1, 3  )
     xin((nnx+1-imin+(j-jmin)*(imax-imin+1)+(0-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+4) = ar25_r(1, j, 1, 4  )
     xin((nnx+1-imin+(j-jmin)*(imax-imin+1)+(0-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+5) = ar25_r(1, j, 1, 5  )

     xin((0-imin+(j-jmin)*(imax-imin+1)+(nnz+1-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+1) = ar16_r(1, j, 1, 1  )
     xin((0-imin+(j-jmin)*(imax-imin+1)+(nnz+1-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+2) = ar16_r(1, j, 1, 2  )
     xin((0-imin+(j-jmin)*(imax-imin+1)+(nnz+1-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+3) = ar16_r(1, j, 1, 3  )
     xin((0-imin+(j-jmin)*(imax-imin+1)+(nnz+1-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+4) = ar16_r(1, j, 1, 4  )
     xin((0-imin+(j-jmin)*(imax-imin+1)+(nnz+1-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+5) = ar16_r(1, j, 1, 5  )

     xin((nnx+1-imin+(j-jmin)*(imax-imin+1)+(nnz+1-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+1) = ar26_r(1, j, 1, 1  )
     xin((nnx+1-imin+(j-jmin)*(imax-imin+1)+(nnz+1-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+2) = ar26_r(1, j, 1, 2  )
     xin((nnx+1-imin+(j-jmin)*(imax-imin+1)+(nnz+1-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+3) = ar26_r(1, j, 1, 3  )
     xin((nnx+1-imin+(j-jmin)*(imax-imin+1)+(nnz+1-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+4) = ar26_r(1, j, 1, 4  )
     xin((nnx+1-imin+(j-jmin)*(imax-imin+1)+(nnz+1-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+5) = ar26_r(1, j, 1, 5  )
  enddo
  do i=1,nnx
     xin((i-imin+(0-jmin)*(imax-imin+1)+(0-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+1) = ar35_r(i, 1, 1, 1  )
     xin((i-imin+(0-jmin)*(imax-imin+1)+(0-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+2) = ar35_r(i, 1, 1, 2  )
     xin((i-imin+(0-jmin)*(imax-imin+1)+(0-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+3) = ar35_r(i, 1, 1, 3  )
     xin((i-imin+(0-jmin)*(imax-imin+1)+(0-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+4) = ar35_r(i, 1, 1, 4  )
     xin((i-imin+(0-jmin)*(imax-imin+1)+(0-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+5) = ar35_r(i, 1, 1, 5  )
     
     xin((i-imin+(nny+1-jmin)*(imax-imin+1)+(0-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+1) = ar45_r(i, 1, 1, 1  ) 
     xin((i-imin+(nny+1-jmin)*(imax-imin+1)+(0-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+2) = ar45_r(i, 1, 1, 2  ) 
     xin((i-imin+(nny+1-jmin)*(imax-imin+1)+(0-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+3) = ar45_r(i, 1, 1, 3  ) 
     xin((i-imin+(nny+1-jmin)*(imax-imin+1)+(0-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+4) = ar45_r(i, 1, 1, 4  ) 
     xin((i-imin+(nny+1-jmin)*(imax-imin+1)+(0-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+5) = ar45_r(i, 1, 1, 5  ) 
     
     xin((i-imin+(0-jmin)*(imax-imin+1)+(nnz+1-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+1) = ar36_r(i, 1, 1, 1  ) 
     xin((i-imin+(0-jmin)*(imax-imin+1)+(nnz+1-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+2) = ar36_r(i, 1, 1, 2  ) 
     xin((i-imin+(0-jmin)*(imax-imin+1)+(nnz+1-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+3) = ar36_r(i, 1, 1, 3  ) 
     xin((i-imin+(0-jmin)*(imax-imin+1)+(nnz+1-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+4) = ar36_r(i, 1, 1, 4  ) 
     xin((i-imin+(0-jmin)*(imax-imin+1)+(nnz+1-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+5) = ar36_r(i, 1, 1, 5  ) 

     xin((i-imin+(nny+1-jmin)*(imax-imin+1)+(nnz+1-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+1) = ar46_r(i, 1, 1, 1  )
     xin((i-imin+(nny+1-jmin)*(imax-imin+1)+(nnz+1-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+2) = ar46_r(i, 1, 1, 2  )
     xin((i-imin+(nny+1-jmin)*(imax-imin+1)+(nnz+1-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+3) = ar46_r(i, 1, 1, 3  )
     xin((i-imin+(nny+1-jmin)*(imax-imin+1)+(nnz+1-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+4) = ar46_r(i, 1, 1, 4  )
     xin((i-imin+(nny+1-jmin)*(imax-imin+1)+(nnz+1-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+5) = ar46_r(i, 1, 1, 5  )
  enddo
 
  ! 8 corners
  xin((0-imin+(0-jmin)*(imax-imin+1)+(0-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+1) = coin135_r(1, 1, 1, 1)  
  xin((0-imin+(0-jmin)*(imax-imin+1)+(0-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+2) = coin135_r(1, 1, 1, 2)  
  xin((0-imin+(0-jmin)*(imax-imin+1)+(0-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+3) = coin135_r(1, 1, 1, 3)  
  xin((0-imin+(0-jmin)*(imax-imin+1)+(0-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+4) = coin135_r(1, 1, 1, 4)  
  xin((0-imin+(0-jmin)*(imax-imin+1)+(0-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+5) = coin135_r(1, 1, 1, 5)  

  xin((0-imin+(0-jmin)*(imax-imin+1)+(nnz+1-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+1) = coin136_r(1, 1, 1, 1)
  xin((0-imin+(0-jmin)*(imax-imin+1)+(nnz+1-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+2) = coin136_r(1, 1, 1, 2)
  xin((0-imin+(0-jmin)*(imax-imin+1)+(nnz+1-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+3) = coin136_r(1, 1, 1, 3)
  xin((0-imin+(0-jmin)*(imax-imin+1)+(nnz+1-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+4) = coin136_r(1, 1, 1, 4)
  xin((0-imin+(0-jmin)*(imax-imin+1)+(nnz+1-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+5) = coin136_r(1, 1, 1, 5)

  xin((0-imin+(nny+1-jmin)*(imax-imin+1)+(0-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+1) = coin145_r(1, 1, 1, 1)  
  xin((0-imin+(nny+1-jmin)*(imax-imin+1)+(0-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+2) = coin145_r(1, 1, 1, 2)  
  xin((0-imin+(nny+1-jmin)*(imax-imin+1)+(0-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+3) = coin145_r(1, 1, 1, 3)  
  xin((0-imin+(nny+1-jmin)*(imax-imin+1)+(0-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+4) = coin145_r(1, 1, 1, 4)  
  xin((0-imin+(nny+1-jmin)*(imax-imin+1)+(0-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+5) = coin145_r(1, 1, 1, 5)  

  xin((0-imin+(nny+1-jmin)*(imax-imin+1)+(nnz+1-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+1) = coin146_r(1, 1, 1, 1) 
  xin((0-imin+(nny+1-jmin)*(imax-imin+1)+(nnz+1-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+2) = coin146_r(1, 1, 1, 2) 
  xin((0-imin+(nny+1-jmin)*(imax-imin+1)+(nnz+1-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+3) = coin146_r(1, 1, 1, 3) 
  xin((0-imin+(nny+1-jmin)*(imax-imin+1)+(nnz+1-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+4) = coin146_r(1, 1, 1, 4) 
  xin((0-imin+(nny+1-jmin)*(imax-imin+1)+(nnz+1-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+5) = coin146_r(1, 1, 1, 5) 


  xin((nnx+1-imin+(0-jmin)*(imax-imin+1)+(0-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+1) = coin235_r(1, 1, 1, 1) 
  xin((nnx+1-imin+(0-jmin)*(imax-imin+1)+(0-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+2) = coin235_r(1, 1, 1, 2) 
  xin((nnx+1-imin+(0-jmin)*(imax-imin+1)+(0-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+3) = coin235_r(1, 1, 1, 3) 
  xin((nnx+1-imin+(0-jmin)*(imax-imin+1)+(0-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+4) = coin235_r(1, 1, 1, 4) 
  xin((nnx+1-imin+(0-jmin)*(imax-imin+1)+(0-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+5) = coin235_r(1, 1, 1, 5) 

  xin((nnx+1-imin+(0-jmin)*(imax-imin+1)+(nnz+1-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+1) = coin236_r(1, 1, 1, 1) 
  xin((nnx+1-imin+(0-jmin)*(imax-imin+1)+(nnz+1-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+2) = coin236_r(1, 1, 1, 2) 
  xin((nnx+1-imin+(0-jmin)*(imax-imin+1)+(nnz+1-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+3) = coin236_r(1, 1, 1, 3) 
  xin((nnx+1-imin+(0-jmin)*(imax-imin+1)+(nnz+1-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+4) = coin236_r(1, 1, 1, 4) 
  xin((nnx+1-imin+(0-jmin)*(imax-imin+1)+(nnz+1-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+5) = coin236_r(1, 1, 1, 5)  

  xin((nnx+1-imin+(nny+1-jmin)*(imax-imin+1)+(0-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+1) = coin245_r(1, 1, 1, 1) 
  xin((nnx+1-imin+(nny+1-jmin)*(imax-imin+1)+(0-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+2) = coin245_r(1, 1, 1, 2) 
  xin((nnx+1-imin+(nny+1-jmin)*(imax-imin+1)+(0-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+3) = coin245_r(1, 1, 1, 3) 
  xin((nnx+1-imin+(nny+1-jmin)*(imax-imin+1)+(0-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+4) = coin245_r(1, 1, 1, 4) 
  xin((nnx+1-imin+(nny+1-jmin)*(imax-imin+1)+(0-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+5) = coin245_r(1, 1, 1, 5) 

  xin((nnx+1-imin+(nny+1-jmin)*(imax-imin+1)+(nnz+1-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+1) = coin246_r(1, 1, 1, 1)
  xin((nnx+1-imin+(nny+1-jmin)*(imax-imin+1)+(nnz+1-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+2) = coin246_r(1, 1, 1, 2)
  xin((nnx+1-imin+(nny+1-jmin)*(imax-imin+1)+(nnz+1-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+3) = coin246_r(1, 1, 1, 3)
  xin((nnx+1-imin+(nny+1-jmin)*(imax-imin+1)+(nnz+1-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+4) = coin246_r(1, 1, 1, 4)
  xin((nnx+1-imin+(nny+1-jmin)*(imax-imin+1)+(nnz+1-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+5) = coin246_r(1, 1, 1, 5)
  
  return
  
end subroutine com_3D_xin

#endif

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine LIMITES_XIN
!
!> Boundary conditions for rthe xin vector.
!<
subroutine limites_xin(xin)
  
  use parameters, only : ndim,nx,nvar_ray
  use unites    , only : c
  use divers    , only : verbose
  use cl_ray
  use tabgmres  , only : imin,imax,jmin,jmax,kmin,kmax

  implicit none

  real, dimension(-nvar_ray+1:(imax-imin+(jmax-jmin)*(imax-imin+1)+(kmax-kmin)*(imax-imin+1)*(jmax-jmin+1))*nvar_ray) :: xin

  integer, dimension(1:6) :: face
  integer                 :: ivar,iface,i,j,k,idim
  
  ! ICL = 1 : FREE-FLOW
  ! ICL = 2 : PERIODIC       (already done in the communications)
  ! ICL = 3 : REFLEXIVE
  ! ICL = 4 : IMPOSED VALUES (inside the ghost cells)
  ! ICL = 5 : ANALYTICAL

  if(verbose) write(*,*) 'Debut limites_xin'

  ! Order in which the faces are treated
  face(1) = 1
  face(2) = 2
  face(3) = 3
  face(4) = 4
  face(5) = 5
  face(6) = 6

  do iface = 1,2*ndim

    ! 1) ------------------

    if (face(iface) == 1) then
    
    if   (ICL_RAY(1) == 1 .or. ICL_RAY(1) == 4) then ! flux impose ou gradient nul
       do ivar=1,nvar_ray
          do j=jmin,jmax
             do k=kmin,kmax
                xin((0-imin+(j-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+ivar) = xin((1-imin+(j-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+ivar)
             enddo
          enddo
       enddo
#if PARA==0
       else if(ICL_RAY(1) == 2) then ! periodique
       do ivar=1,nvar_ray
          do j=jmin,jmax
             do k=kmin,kmax
                xin((0-imin+(j-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+ivar) = xin((nx(1)-imin+(j-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+ivar)
             enddo
          enddo
       enddo
#endif
    else if (ICL_RAY(1) == 3) then ! refexif
       do ivar=1,nvar_ray
          do j=jmin,jmax
             do k=kmin,kmax
                xin((0-imin+(j-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+ivar) = xin((1-imin+(j-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+ivar)
             enddo
          enddo
       enddo         
       do j=jmin,jmax
          do k=kmin,kmax
             xin((0-imin+(j-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+3) = -xin((1-imin+(j-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+3)
          enddo
       enddo
    else if (ICL_RAY(1) == 5) then
       do j=jmin,jmax
          do k=kmin,kmax
             xin((0-imin+(j-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+1) = 1.
             xin((0-imin+(j-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+2) = 1.
             do idim=1,ndim
                xin((0-imin+(j-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+2+idim) = Fr_l(idim,1)/(c*Er_l(1))
             enddo
          enddo
       enddo
    endif

    endif

    !2)----------------------------------------

    if (face(iface) == 2) then

    if   (ICL_RAY(2) == 1 .or. ICL_RAY(2) == 4) then ! flux impose ou gradient nul
       do ivar=1,nvar_ray
          do j=jmin,jmax
             do k=kmin,kmax
                xin((nx(1)+1-imin+(j-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+ivar) = xin((nx(1)-imin+(j-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+ivar)
             enddo
          enddo
       enddo
#if PARA==0
    else if (ICL_RAY(2) == 2) then ! periodique
       do ivar=1,nvar_ray
          do j=jmin,jmax
             do k=kmin,kmax
                xin((nx(1)+1-imin+(j-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+ivar) = xin((1-imin+(j-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+ivar)
             enddo
          enddo
       enddo
#endif
    else if (ICL_RAY(2) == 3) then ! reflexif
       do ivar=1,nvar_ray
          do j=jmin,jmax
             do k=kmin,kmax
                xin((nx(1)+1-imin+(j-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+ivar) = xin((nx(1)-imin+(j-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+ivar)
             enddo
          enddo
       enddo
       do j=jmin,jmax
          do k=kmin,kmax
             xin((nx(1)+1-imin+(j-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+3) = -xin((nx(1)-imin+(j-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+3)
          enddo
       enddo
    else if (ICL_RAY(2) == 5) then
       do j=jmin,jmax
          do k=kmin,kmax
             xin((nx(1)+1-imin+(j-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+1) = 1.
             xin((nx(1)+1-imin+(j-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+2) = 1.
             do idim=1,ndim
                xin((nx(1)+1-imin+(j-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+2+idim) = Fr_l(idim,2)/(c*Er_l(2))
             enddo
          enddo
       enddo
    endif

    endif

    !3)----------------------------------------

    if (face(iface) == 3) then

    if   (ICL_RAY(3) == 1 .or. ICL_RAY(3) == 4) then ! flux impose ou gradient nul
       do ivar=1,nvar_ray
          do i=imin,imax
             do k=kmin,kmax
                xin((i-imin+(0-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+ivar) = xin((i-imin+(1-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+ivar)
             enddo
          enddo
       enddo
#if PARA==0
    else if (ICL_RAY(3) == 2) then ! periodique
       do ivar=1,nvar_ray
          do i=imin,imax
             do k=kmin,kmax
                xin((i-imin+(0-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+ivar) = xin((i-imin+(nx(2)-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+ivar)
             enddo
          enddo
       enddo
#endif
    else if (ICL_RAY(3) == 3) then ! reflexif
       do ivar=1,nvar_ray
          do i=imin,imax
             do k=kmin,kmax
                xin((i-imin+(0-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+ivar) = xin((i-imin+(1-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+ivar)
             enddo
          enddo
       enddo
       do i=imin,imax
          do k=kmin,kmax
             xin((i-imin+(0-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+4) = -xin((i-imin+(1-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+4)
          enddo
       enddo
    else if (ICL_RAY(3) == 5) then
       do i=imin,imax
          do k=kmin,kmax
             xin((i-imin+(0-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+1) = 1.
             xin((i-imin+(0-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+2) = 1.
             do idim=1,ndim
                xin((i-imin+(0-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+2+idim) = Fr_l(idim,3)/(c*Er_l(3))
             enddo
          enddo
       enddo
    endif

    endif

    !4)----------------------------------------

    if (face(iface) == 4) then

    if   (ICL_RAY(4) == 1 .or. ICL_RAY(4) == 4) then ! flux impose ou gradient nul
       do ivar=1,nvar_ray
          do i=imin,imax
             do k=kmin,kmax
                xin((i-imin+(nx(2)+1-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+ivar) = xin((i-imin+(nx(2)-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+ivar)
             enddo
          enddo
       enddo
#if PARA==0
    else if(ICL_RAY(4) == 2) then ! periodique
       do ivar=1,nvar_ray
          do i=imin,imax
             do k=kmin,kmax
                xin((i-imin+(nx(2)+1-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+ivar) = xin((i-imin+(1-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+ivar)
             enddo
          enddo
       enddo
#endif
    else if (ICL_RAY(4) == 3) then ! reflexif
       do ivar=1,nvar_ray
          do i=imin,imax
             do k=kmin,kmax
                xin((i-imin+(nx(2)+1-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+ivar) = xin((i-imin+(nx(2)-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+ivar)
             enddo
          enddo
       enddo
       do i=imin,imax
          do k=kmin,kmax
             xin((i-imin+(nx(2)+1-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+4) = -xin((i-imin+(nx(2)-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+4)
          enddo
       enddo
    else if (ICL_RAY(4) == 5) then
       do i=imin,imax
          do k=kmin,kmax
             xin((i-imin+(nx(2)+1-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+1) = 1.
             xin((i-imin+(nx(2)+1-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+2) = 1.
             do idim=1,ndim
                xin((i-imin+(nx(2)+1-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+2+idim) = Fr_l(idim,4)/(c*Er_l(4))
             enddo
          enddo
       enddo
    endif

    endif

    !5)----------------------------------------

    if (face(iface) == 5) then

    if   (ICL_RAY(5) == 1 .or. ICL_RAY(5) == 4) then ! flux impose ou gradient nul
       do ivar=1,nvar_ray
          do i=imin,imax
             do j=jmin,jmax
                xin((i-imin+(j-jmin)*(imax-imin+1)+(0-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+ivar) = xin((i-imin+(j-jmin)*(imax-imin+1)+(1-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+ivar)
             enddo
          enddo
       enddo
#if PARA==0
    else if (ICL_RAY(5) == 2) then ! periodique
       do ivar=1,nvar_ray
          do i=imin,imax
             do j=jmin,jmax
                xin((i-imin+(j-jmin)*(imax-imin+1)+(0-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+ivar) = xin((i-imin+(j-jmin)*(imax-imin+1)+(nx(3)-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+ivar)
             enddo
          enddo
       enddo
#endif
    else if (ICL_RAY(5) == 3) then ! reflexif
       do ivar=1,nvar_ray
          do i=imin,imax
             do j=jmin,jmax
                xin((i-imin+(j-jmin)*(imax-imin+1)+(0-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+ivar) = xin((i-imin+(j-jmin)*(imax-imin+1)+(1-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+ivar)
             enddo
          enddo
       enddo
       do i=imin,imax
          do j=jmin,jmax
             xin((i-imin+(j-jmin)*(imax-imin+1)+(0-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+5) = -xin((i-imin+(j-jmin)*(imax-imin+1)+(1-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+5)
          enddo
       enddo
    else if (ICL_RAY(5) == 5) then
          do i=imin,imax
             do j=jmin,jmax
                xin((i-imin+(j-jmin)*(imax-imin+1)+(0-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+1) = 1.
                xin((i-imin+(j-jmin)*(imax-imin+1)+(0-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+2) = 1.
                do idim=1,ndim
                   xin((i-imin+(j-jmin)*(imax-imin+1)+(0-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+2+idim) = Fr_l(idim,5)/(c*Er_l(5))
                enddo
             enddo
          enddo
    endif

    endif

    !6)----------------------------------------

    if (face(iface) == 6) then

    if   (ICL_RAY(6) == 1 .or. ICL_RAY(6) == 4) then ! flux impose ou gradient nul
       do ivar=1,nvar_ray
          do i=imin,imax
             do j=jmin,jmax
                xin((i-imin+(j-jmin)*(imax-imin+1)+(nx(3)+1-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+ivar) = xin((i-imin+(j-jmin)*(imax-imin+1)+(nx(3)-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+ivar)
             enddo
          enddo
       enddo
#if PARA==0
    else if (ICL_RAY(6) == 2) then ! periodique
       do ivar=1,nvar_ray
          do i=imin,imax
             do j=jmin,jmax
                xin((i-imin+(j-jmin)*(imax-imin+1)+(nx(3)+1-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+ivar) = xin((i-imin+(j-jmin)*(imax-imin+1)+(1-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+ivar)
             enddo
          enddo
       enddo
#endif
    else if (ICL_RAY(6) == 3) then ! reflexif
       do ivar=1,nvar_ray
          do i=imin,imax
             do j=jmin,jmax
                xin((i-imin+(j-jmin)*(imax-imin+1)+(nx(3)+1-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+ivar) = xin((i-imin+(j-jmin)*(imax-imin+1)+(nx(3)-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+ivar)
             enddo
          enddo
       enddo
       do i=imin,imax
          do j=jmin,jmax
             xin((i-imin+(j-jmin)*(imax-imin+1)+(nx(3)+1-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+5) = -xin((i-imin+(j-jmin)*(imax-imin+1)+(nx(3)-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+5)
          enddo
       enddo
    else if (ICL_RAY(6) == 5) then
          do i=imin,imax
             do j=jmin,jmax
                xin((i-imin+(j-jmin)*(imax-imin+1)+(nx(3)+1-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+1) = 1.
                xin((i-imin+(j-jmin)*(imax-imin+1)+(nx(3)+1-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+2) = 1.
                do idim=1,ndim
                   xin((i-imin+(j-jmin)*(imax-imin+1)+(nx(3)+1-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+2+idim) = Fr_l(idim,6)/(c*Er_l(6))
                enddo
             enddo
          enddo
    endif

    endif

    enddo

  if(verbose) write(*,*) 'Fin limites_xin'
  
  return

end subroutine limites_xin
