!> \file
!! Contains modules varray, divers_ray, cl_ray,
!! valeurs_propres, TabGmres, VarGmres and interfaces
!<

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Module VARRAY
!
!> Contains the M1 radiative transfer variables.
!<
module varray
  implicit none
  real, dimension(:,:,:  ), allocatable :: Eray             !< Radiative energy
  real, dimension(:,:,:  ), allocatable :: Erayold          !< Old radiative energy
  real, dimension(:,:,:,:), allocatable :: Fray             !< Radiative flux
  real, dimension(:,:,:,:), allocatable :: Frayold          !< Old radiative flux
  real, dimension(:,:,:  ), allocatable :: kappa_emi_ray    !< Emission opacity \f$ \kappa_{\mathrm{em}} \f$
  real, dimension(:,:,:  ), allocatable :: kappa_abs_ray    !< Absorption opacity \f$ \kappa_{\mathrm{abs}} \f$
  real, dimension(:,:,:  ), allocatable :: kappa_sca_ray    !< Scattering coefficient \f$ \kappa_{s} \f$
end module varray

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Module DIVERS_RAY
!
!> Contains various variables for the M1 radiative transfer.
!<
module divers_ray
  implicit none
  integer, dimension(6) :: ordre_des_faces_pour_le_ray        !< Order in which the boundary faces are filled for radiative transfer
  integer               :: slope_type_ray                     !< Type of slope limiter for radiative transfer
  integer               :: nitetot_ray                        !< Total number of iterations in the implicit resolution
  integer               :: nitemax_ray                        !< Maximum number of iterations per timestep allowed in the implicit resolution
  integer               :: ngrp_ray                           !< Number of frequency groups
  real                  :: varrel_ray                         !< Maximum radiative quantity variation allowed in an implicit timestep
  real                  :: varrel_gaz                         !< Maximum gas temperature variation allowed in an implicit timestep
  real                  :: alpha                              !< \f$ \alpha \f$
  real                  :: alpha_explicite                    !< \f$ \alpha_{exp} \f$
  real                  :: dt_ray_imp                         !< Implicit radiative transfer timestep
  real                  :: dt_ray_exp                         !< Explicit radiative transfer timestep
  real                  :: dt_ray_old                         !< Old radiative transfer timestep
  real                  :: dt_ray                             !< Radiative transfer timestep
  real                  :: dt_ray_glob                        !< Global radiative transfer timestep
  real                  :: rho1                               !< Density used to compute opacity \f$ \kappa \f$
  real                  :: T1                                 !< Temperature used to compute opacity \f$ \kappa \f$
  real                  :: alpha_var                          !< \f$ \alpha_{var} \f$
  real                  :: valp_min                           !< Minimum allowed value for eigenvalues
  logical               :: IMP_RAY                            !< Solves radiative transfer implicitely if .true.
  logical               :: EXP_RAY                            !< Solves radiative transfer explicitely if .true.
  logical               :: precond_ray                        !< Performs matrix preconditionning if .true.
  logical               :: cal_valp                           !< Computes M1 eigenvalues if .true.
  logical               :: deja_converge                      !< .true. if already converged
  character (len=10)    :: matrix_inv_solver                  !< Type of implicit solver ('GS', 'GMRES', 'GS-GMRES', 'GMRES-GS' or 'yoyo')
  character (len=2)     :: rad_trans_model                    !< Type of radiative transfer model: 'M1' or 'P1'
end module divers_ray

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Module CL_RAY
!
!> Contains the boundary conditions variables for the M1
!! radiative transfer.
!<
module cl_ray
  implicit none
  integer, dimension(6  )              :: icl_ray             !< Type of radiative boundary conditions
  integer, dimension(6  )              :: icl_ray_glob        !< Type of global radiative boundary conditions
  real   , dimension(:  ), allocatable :: Er_l                !< Radiative energy imposed in the ghost cells
  real   , dimension(:  ), allocatable :: Tr_l                !< Gas temperature imposed in the ghost cells
  real   , dimension(:,:), allocatable :: Fr_l                !< Radiative flux imposed in the ghost cells
end module cl_ray

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Module VALEURS_PROPRES
!
!> Contains the array to hold the tabulated M1 eignevalues.
!<
module valeurs_propres
  implicit none
  integer                               :: n_points           !< Number of points in the tabulated eigenvalues curve
  real, dimension(:,:,:,:), allocatable :: valp               !< Array to hold the tabulated eigenvalues as a function of \f$\theta\f$ and \f$\epsilon\f$
end module valeurs_propres

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Module TABGMRES
!
!> Contains workspace arrays for GMRES.
!<
module TabGmres
  implicit none
  integer                         :: imin                     !< imin
  integer                         :: imax                     !< imax
  integer                         :: jmin                     !< jmin
  integer                         :: jmax                     !< jmax
  integer                         :: kmin                     !< kmin
  integer                         :: kmax                     !< kmax
  real, dimension(:), allocatable :: work                     !< work
  real, dimension(:), allocatable :: workred                  !< workred
  real, dimension(:), allocatable :: xin                      !< xin
  real, dimension(:), allocatable :: xout                     !< xout
  real, dimension(:), allocatable :: b                        !< b
end module TabGmres

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Module VARGMRES
!
!> Contains the parameters for GMRES.
!<
module VarGmres
  implicit none
  integer, dimension(7) :: icntl                              !< icntl
  integer               :: restart                            !< restart
  integer               :: itermax                            !< itermax
  integer               :: lwork                              !< lwork
  real   , dimension(5) :: cntl                               !< cntl
end module VarGmres
 
!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Module INTERFACES
!
!> GMRES interfaces.
!<
module interfaces
  implicit none
  interface
    subroutine gmres(n,m,nloc,work,nx,ny,nz,nvar_ray,imin,imax,jmin,jmax,kmin,kmax,workred,lwork,cntl,icntl,info)
      integer, intent(in)       :: n                          !< n
      integer, intent(in)       :: m                          !< m
      integer, intent(in)       :: nloc                       !< nloc
      integer, intent(in)       :: lwork                      !< lwork
      integer, intent(in)       :: nx                         !< nx
      integer, intent(in)       :: ny                         !< ny
      integer, intent(in)       :: nz                         !< nz
      integer, intent(in)       :: nvar_ray                   !< nvar_ray
      integer, intent(in)       :: imin                       !< imin
      integer, intent(in)       :: imax                       !< imax
      integer, intent(in)       :: jmin                       !< jmin
      integer, intent(in)       :: jmax                       !< jmax
      integer, intent(in)       :: kmin                       !< kmin
      integer, intent(in)       :: kmax                       !< kmax
      integer, dimension(7)     :: icntl                      !< icntl
      integer, dimension(3)     :: info                       !< info
      real   , dimension(-nvar_ray+1:(imax-imin+(jmax-jmin)*(imax-imin+1)+(kmax-kmin)*(imax-imin+1)*(jmax-jmin+1))*nvar_ray):: workred !< workred
      real   , dimension(lwork) :: work                       !< work
      real   , dimension(5)     :: cntl                       !< cntl
    end subroutine gmres
  end interface
end module interfaces
