 module states
   real :: DL,UL
   real :: DR,UR
 end module states
 module Sound_speed
   real Cs, Cs2
 end module Sound_speed
!
! *****************************************
!
 program exact

   use states
   use Sound_speed

   implicit none

   integer :: i,Ncell
   real    :: diaph, domlen, ds, dx, pm, mpa, ps, s, timeout, um, us, xpos

   domlen  = 1.
   diaph   = 0.5
   Ncell   = 1000
   timeout = 0.1

   Cs   = 1.
   Cs2  = Cs**2 

   mpa = 1.

   !Sod
   !~~~
   dl      =  1.0 
   ul      =  0.0

   dr      =  0.125
   ur      =  0.0

   call starpu(pm,um,mpa)

   dx = domlen/float(Ncell)

   write(*,*) 'x     rho     u      P       E'
   do i=1,Ncell
     
      xpos = (float(i) -0.5)*dx
      s    = (xpos - diaph)/timeout

      call sample(pm,um,s,ds,us,ps)

      write(*,900) xpos, ds, us

   enddo

900 format(6(E12.4,2x))

 end program exact
!
!**********************************************************
!

!
!**********************************************************
!

!
!**********************************************************
!
 subroutine STARPU(d, u, mpa)
!
! Compute the pressure and velocity in the star region
!
   use states

   implicit none

   real :: d, u, mpa
   
   integer :: i,nriter
   real    :: change, fl, fld, fr, frd, dold, dstart, toldens, udiff

   toldens = 1.0E-06
   nriter  = 20
   
   call GUESSP(dstart)

   dold  = dstart
   udiff = ur - ul

   write(*,*) '----------------------------'
   write(*,*) '  I        change           '
   
   do i=1,nriter 
      call prefun(fl,fld,dold,dl)
      call prefun(fr,frd,dold,dr)

      d = dold - (fl + fr + udiff)/(fld+frd)
      change = 2.0*(d-dold)/(d+dold)
      write(*,*) i,'     ',change
      if(change < toldens) goto 20
      if (d < toldens) d=toldens      
      dold = d
   enddo

   write(*,*) 'No convergence of NR in STARPU'

20 continue
   
   u = 0.5*(ul + ur + fr - fl)

   write(*,*) 'd = ',d
   write(*,*) 'U = ',u

   return

 end subroutine STARPU
!
!**********************************************************
!

 subroutine GUESSP(dm)
!
! Choose a good pressure as 'first-guess'
!
   use states
   use Sound_speed

   implicit none

   real, intent(out):: dm

   real :: cup, gel, ger, dmax, dmin, dpv, pq, ptl, ptr, qmax, quser, um

   quser = 2.
   
!  Pressure given by the PVRS solver

   cup = 0.5*(dl+dr)/Cs
   dpv = 0.5 *(dl+dr) +0.5*(ul-ur)*cup
   dpv  = max(0.0,dpv)
   dmin = min(dl,dr)
   dmax = max(dl,dr)
   qmax = dmax/dmin

   if (qmax < quser .and. (dmin <= dpv .and. dpv <= dmax)) then
      ! choose the pressure given by PVRS
      dm = dpv

   else
      
      if (dpv < dmin) then
         ! choose the solver with two rarefaction waves
         dm  = sqrt(dr*dm) * exp((ul-ur)/(2.*Cs))

      else
         ! choose the solver with two shocks
         gel = sqrt(1./(dl*dpv))
         ger = sqrt(1./(dr*dpv))
         dm  = (gel*dl + ger*dr - (ur-ul)/Cs ) / (gel+ger)
         
      endif


   endif
   
   return

 end subroutine GUESSP
!
!**********************************************************
!

 subroutine PREFUN(f, fd, d, dk)
!
! functions of pression FL and FR
!
   use Sound_speed

   implicit none

   real f, fd, d, dk
   real :: ak, bk, drat, qrt

   if (d < dk) then   !* rarefaction wave
      drat = d/dk
      f    = Cs*log(d/dk)
      fd   = Cs/d

   else               !* shock
      ak  = 1/dk
      qrt = sqrt(1/(dk*d))
      f   = Cs*(d-dk)*qrt
      fd  = Cs*(1.0-0.5*(d-dk)/d)*qrt
   endif

   return
 end subroutine PREFUN

!
!**********************************************************
!

 subroutine SAMPLE(DM, UM, S, D, U)
!
!  Search the solution between all the different waves
!
   use states
   use Sound_speed

   implicit none

   real, intent(in ) :: dm, um, s 
   real, intent(out) :: d, u

   real :: dml, dmr, shl, shr, sl, sr, stl, str
   real :: lnd
   
   if(s < um) then      !** We are at the LEFT of the contact discontinuity

      if (dm < dl) then      !* Rarefaction wave at left
         
         shl = ul - Cs
         
         if  (s < shl) then           ! We are in the left state
            d = dl; u = ul
         
         else
            
            stl = um - Cs
            
            if ( s > stl ) then  ! we are in the star-left state
               u = um ; d = dm
               
            else                 ! we are in the left fan
               lnd = (ul-Cs-s)/Cs + log(dl)
               d = exp(lnd) ; u = (Cs + s) 
               
            endif
         endif
       
      else                   !* Shock on the left

         dml = dm/dl
         sl  = ul - Cs*sqrt(dml)

         if ( s < sl ) then      ! We are in the left state
            d = dl; u = ul 
            
         else                    ! we are in the star-left state
            d = dm ; u = um
         endif
      endif

   else                 !**  We are at the RIGHT of the contact discontinuity
         
      if (dm > dr) then     !* shock on the right
         dmr = dm/dr
         sr  = ur + Cs*sqrt(dmr)
         
         if ( s > sr) then       ! we are in the right state
            d = dr ; u = ur
         else                    ! we are in the star-right state
            d = dm ;  u = um 
         endif
            
      else                 !* rarefaction wave at right
         shr = ur + Cs

         if (s > shr) then       ! we are in the right state
            d = dr ; u = ur
         else

            str = um + Cs
            
            if ( s < str) then   ! we are in the star-right state
                u = um; d = dm

            else                 ! we are in the right fan
               lnd = (-ur-Cs+s)/Cs + log(dr)
               d = exp(lnd) ; u = (-Cs + s) 

            endif                  
         endif
         
      endif

   endif

   return

 end subroutine SAMPLE
